/******************************************************************************\
 * File: cbuffer.c                                                            *
 * Author: James Greensky                                                     *
 * Notes: Implementation of a circular block buffer.                          *
\******************************************************************************/

#include "cbuffer.h"
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

struct CBufferRecord {
  short ItemSize;
  int Size, Capacity, Head, Tail;
  char* Array;
};

inline int CBufferSize(CBuffer buffer) {
  assert(buffer != NULL);

  return buffer->Size / buffer->ItemSize;
}

inline int CBufferCapacity(CBuffer buffer) {
  assert(buffer != NULL);
  
  return buffer->Capacity / buffer->ItemSize;
}

CBuffer CBufferCreate(int capacity, int itemSize) {
  assert(itemSize > 0);

  if (!capacity) 
    capacity = 4096;
  else if (capacity % 4096 != 0) {
    capacity /= 4096;
    capacity++;
    capacity *= 4096;
  }

  CBuffer buffer = malloc(sizeof(struct CBufferRecord));
  if (buffer == NULL) return buffer;
  buffer->ItemSize = itemSize;
  buffer->Capacity = capacity * itemSize;
  buffer->Array = malloc(buffer->Capacity);
  if (buffer->Array == NULL) {
    free(buffer);
    return NULL;
  }
  buffer->Size = buffer->Head = buffer->Tail = 0;
  return buffer;
}

void CBufferClear(CBuffer buffer) {
  assert(buffer != NULL);

  buffer->Size = buffer->Head = buffer->Tail = 0;
}

void CBufferDispose(CBuffer buffer) {
  assert(buffer != NULL);

  free(buffer->Array);
  free(buffer);
}

static void* ResizeCBuffer(CBuffer buffer, int size) {
  assert(buffer != NULL);
  assert(size > -1);

  if (!size)
    size = 4096;
  if (size % 4096 != 0) {
    size /= 4096;
    size++;
    size *= 4096;
  }

  int capacity = buffer->Capacity + (size * buffer->ItemSize);
  char* ptr = malloc(sizeof(char) * capacity);
  if (ptr == NULL) return ptr;
  if (buffer->Size != 0) {
    if (buffer->Tail > buffer->Head)
      memcpy(ptr, &buffer->Array[buffer->Head], buffer->Size);
    else {
      int length = buffer->Capacity - buffer->Head;
      memcpy(ptr, &buffer->Array[buffer->Head], length);
      memcpy(&ptr[length], buffer->Array, buffer->Tail);
    }
  }
  buffer->Head = 0;
  buffer->Tail = buffer->Size;
  free(buffer->Array);
  buffer->Array = ptr;
  buffer->Capacity = capacity;
  return &buffer->Array[buffer->Tail];
}

void* CBufferGetWriteBlock(CBuffer buffer, int* count) {
  assert(buffer != NULL);
  assert(count != NULL);
  assert(*count > -1);

  int length = 0;
  if (buffer->Capacity != buffer->Size)
    length = (((buffer->Tail < buffer->Head) ? buffer->Head : buffer->Capacity)  - 
	    buffer->Tail) / buffer->ItemSize;

  if ((!length) || (*count > length && (((buffer->Tail + length) % buffer->Capacity) == 
			  (buffer->Head % buffer->Capacity)))) {
    if (ResizeCBuffer(buffer, *count - length) == NULL)
      return NULL;
    else
      length = (*count) ? *count : buffer->Capacity - buffer->Size;
  }

  if ((!*count) || *count > length)
    *count = length;

  return &buffer->Array[buffer->Tail];
}

void* CBufferGetWriteBlockAbs(CBuffer buffer, int count) {
  assert(buffer != NULL);
  assert(count > -1);

  int length = 0;
  if (buffer->Capacity != buffer->Size)
    length = (((buffer->Tail < buffer->Head) ? buffer->Head : buffer->Capacity) -
		buffer->Tail) / buffer->ItemSize;

  if (length < count) {
    if (buffer->Capacity - buffer->Size < count * buffer->ItemSize) {
      if (ResizeCBuffer(buffer, count - length) == NULL) 
	return NULL;
    }
    else {
      memmove(buffer->Array, &buffer->Array[buffer->Head], buffer->Size);
      buffer->Head = 0;
      buffer->Tail = buffer->Size;
    }
  }
  
  return &buffer->Array[buffer->Tail];
}

int CBufferCommitBlock(CBuffer buffer, int count) {
  assert(buffer != NULL);
  assert(count > -1);
  assert(count + buffer->Size <= buffer->Capacity);

  int length = 0;
  if (buffer->Capacity != buffer->Size)
    length = (((buffer->Tail < buffer->Head) ? buffer->Head : buffer->Capacity) -
	      buffer->Tail) / buffer->ItemSize;

  if ((!length) || (count > length && (((buffer->Tail + length) % buffer->Capacity) == 
				       (buffer->Head % buffer->Capacity)))) {
    if (ResizeCBuffer(buffer, count - length) == NULL)
      return -1;
    else
      length = (count) ? count : buffer->Capacity - buffer->Size;
  }

  if (length > count)
    length = count;

  buffer->Tail = (buffer->Tail + length) % buffer->Capacity;
  buffer->Size += length;

  return length;
}

void* CBufferGetReadBlock(CBuffer buffer, int* count) {
  assert(buffer != NULL);
  assert(count != NULL);
  assert(*count > -1);
  
  if (buffer->Size == 0) 
    *count = 0;
  else {
    int length = (((buffer->Head < buffer->Tail) ? buffer->Tail : buffer->Capacity) - 
		  buffer->Head) / buffer->ItemSize;
    if ((!*count) || *count > length) 
      *count = length;
  }

  return &buffer->Array[buffer->Head];
}

void* CBufferGetReadBlockAt(CBuffer buffer, int offset, int* count) {
  assert(buffer != NULL);
  assert(offset > -1);
  assert(count != NULL);
  assert(*count > -1);

  int head = buffer->Head;

  if (buffer->Size == 0)
    *count = 0;
  else {
    head = (buffer->Head + (offset * buffer->ItemSize)) % buffer->Capacity;
    int length = (((head < buffer->Tail) ? buffer->Tail : buffer->Capacity) -
		  head) / buffer->ItemSize;
    if ((!*count) || *count > length) 
      *count = length;
  }

  return &buffer->Array[head];
}

int CBufferInsertBuffer(CBuffer dest, CBuffer src) {
  assert(dest != NULL);
  assert(src != NULL);
  assert(dest->ItemSize == src->ItemSize);

  if (src->Size) {
    int length = (((src->Head < src->Tail) ? src->Tail : src->Capacity) -
		  src->Head) / src->ItemSize;
    if ((src->Size - length)) {
      if (CBufferInsert(dest, src->Array, src->Tail) < 0)
	return -1;
    }

    if (CBufferInsert(dest, &src->Array[src->Head], length) < 0) 
      return -1;
  }
  
  return src->Size / src->ItemSize;
}

int CBufferDecommitBlock(CBuffer buffer, int count) {
  assert(buffer != NULL);
  assert(count > -1);

  int length = (((buffer->Head < buffer->Tail) ? buffer->Tail : buffer->Capacity) -
		buffer->Head) / buffer->ItemSize;
  if (count > length)
    count = length;

  buffer->Head = (buffer->Head + count) % buffer->Capacity;
  buffer->Size -= count;

  return count;
}

int CBufferPeek(CBuffer buffer, void* buf, int count) {
  assert(buffer != NULL);
  assert(buf != NULL);
  assert(count > -1);
  
  int length = buffer->Size / buffer->ItemSize;
  if (count > length)
    count = length;

  if (count) {
    length = (((buffer->Head < buffer->Tail) ? buffer->Tail : buffer->Capacity) -
	      buffer->Head) / buffer->ItemSize;
    if (length > count)
      length = count;
    memcpy(buf, &buffer->Array[buffer->Head], length * buffer->ItemSize);
    
    count -= length;
    if (count != 0)
      memcpy(&((char*)buf)[length * buffer->ItemSize], buffer->Array, count * buffer->ItemSize);
  }

  return count;
}

int CBufferPeekAt(CBuffer buffer, void* buf, int offset, int count) {
  assert(buffer != NULL);
  assert(buf != NULL);
  assert(count > -1);
  assert(offset > -1);

  int length = buffer->Size / buffer->ItemSize;
  if (count > length - offset)
    count = length - offset;

  if (count) {
    int head = (buffer->Head + (offset * buffer->ItemSize)) % buffer->Capacity;
    length = (((head < buffer->Tail) ? buffer->Tail : buffer->Capacity) -
	      head) / buffer->ItemSize;
    if (length > count)
      length = count;
    memcpy(buf, &buffer->Array[head], length * buffer->ItemSize);

    count -= length;
    if (count != 0)
      memcpy(&((char*)buf)[length * buffer->ItemSize], buffer->Array, count * buffer->ItemSize);
  }

  return count;
}

int CBufferWrite(CBuffer buffer, void* buf, int count) {
  assert(buffer != NULL);
  assert(buf != NULL);
  assert(count > -1);

  int length = (buffer->Capacity - buffer->Size) / buffer->ItemSize;
  if (length < count) {
    if (ResizeCBuffer(buffer, count - length) == NULL) 
      return -1;
  }

  length = (((buffer->Tail < buffer->Head) ? buffer->Head : buffer->Capacity) 
	    - buffer->Tail) / buffer->ItemSize;
  if (length > count) 
    length = count;
  memcpy(&buffer->Array[buffer->Tail], buf, length * buffer->ItemSize);

  count -= length;
  if (count)
    memcpy(buffer->Array, &((char*)buf)[length * buffer->ItemSize], count * buffer->ItemSize);

  buffer->Tail = (buffer->Tail + ((length + count) * buffer->ItemSize)) % buffer->Capacity;
  buffer->Size += (length + count) * buffer->ItemSize;

  return length + count;
}

int CBufferWriteBuffer(CBuffer dest, CBuffer src) {
  assert(dest != NULL);
  assert(src != NULL);
  assert(dest->ItemSize == src->ItemSize);

  if (src->Size) {
    int length = (((src->Tail < src->Head) ? src->Head : src->Capacity)
		  - src->Tail) / src->ItemSize;
    if (CBufferWrite(dest, &src->Array[src->Head], length) < 0) 
      return -1;
    if ((src->Size - length)) {
      if (CBufferWrite(dest, src->Array, src->Tail) < 0) {
	CBufferRemoveAt(dest, dest->Size - length, length);
	return -1;
      }
    }
  }

  return src->Size;
}

int CBufferRead(CBuffer buffer, void* buf, int count) {
  assert(buffer != NULL);
  assert(buf != NULL);
  assert(count > -1);

  int length = buffer->Size / buffer->ItemSize;
  if (count > length)
    count = length;

  CBufferPeek(buffer, buf, count);

  buffer->Head = (buffer->Head + (count * buffer->ItemSize)) % buffer->Capacity;
  buffer->Size -= (count * buffer->ItemSize);

  return count;
}

int CBufferInsert(CBuffer buffer, void* buf, int count) {
  assert(buffer != NULL);
  assert(buf != NULL);
  assert(count > -1);

  int length = (buffer->Capacity - buffer->Size) / buffer->ItemSize;
  if (length < count) {
    if (ResizeCBuffer(buffer, count - length) == NULL) 
      return -1;
  }

  length = ((buffer->Head == 0) ? buffer->Capacity : buffer->Head) 
    - ((buffer->Head == 0) ? buffer->Tail : ((buffer->Tail < buffer->Head) ? buffer->Tail : 0)) 
    / buffer->ItemSize;

  if (length > count)
    length = count;

  memcpy(&buffer->Array[((buffer->Head == 0) ? buffer->Capacity : buffer->Head) - (length * buffer->ItemSize)], 
	 &((char*)buf)[(count - length) * buffer->ItemSize], length * buffer->ItemSize);

  count -= length;
  if (count)
    memcpy(&buffer->Array[buffer->Capacity - (count * buffer->ItemSize)],
	   buf, count * buffer->ItemSize);
  
  buffer->Head -= ((length + count) * buffer->ItemSize);
  if (buffer->Head < 0)
    buffer->Head = buffer->Capacity + buffer->Head;
  buffer->Size += ((length + count) * buffer->ItemSize);

  return length + count;
}

int CBufferRemove(CBuffer buffer, int count) {
  assert(buffer != NULL);
  assert(count > -1);

  int length = buffer->Size/ buffer->ItemSize;
  if (count > length)
    count = length;

  if (count) {
    buffer->Head = (buffer->Head + (count * buffer->ItemSize)) % buffer->Capacity;
    buffer->Size -= count * buffer->ItemSize;
  }

  return count;
}

int CBufferRemoveAt(CBuffer buffer, int offset, int count) {
  assert(buffer != NULL);
  assert(offset > -1);
  assert(count > -1);

  int length = buffer->Size / buffer->ItemSize;
  if (count > length - offset)
    count = length - offset;

  if (count) {
    int rem = length - offset - count;
    if (rem) {
      char buf[rem];
      CBufferPeekAt(buffer, buf, offset + count, rem);
      buffer->Tail = (buffer->Head + (offset * buffer->ItemSize)) % buffer->Capacity;
      buffer->Size = offset * buffer->ItemSize;
      CBufferWrite(buffer, buf, rem);
    }
    else {
      buffer->Tail = (buffer->Head + (offset * buffer->ItemSize)) % buffer->Capacity;
      buffer->Size = offset * buffer->ItemSize;
    }
  }

  return count;
}
