/******************************************************************************\
 * Name: file_monitor.c                                                       *
 * Author: James Greensky                                                     *
 * Notes: Implementation of a file monitor                                    *
\******************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/mman.h>

typedef struct FDRecord {
  int FD;
  char* Path;
  struct MonitorRecord* Next;
}* FD;

static const struct option LongOptions[] = {
  { "help", 0, NULL, 'h' },
  { NULL, 0, NULL, 0 }
};

static const char* ShortOptions = "s:l:h";

static const char* const Usage =
"Usage: %s [ options ] [ file ]\n"
"  -h, --help                  Print this information.\n";

static int EventFD = -1, Closing = 0;

static void PrintUsage(int exitCode) {
  fprintf((exitCode != 0) ? stderr : stdout, Usage, program_invocation_short_name);
  exit(exitCode);
}

static void SIGTERM_Handler(int signal) {
  Closing = 1;
}

static void SIGRTMIN_Handler(int signal, siginfo_t* sigInfo, void* data) {
  EventFD = sigInfo->si_fd;
}

static void Cclose(int fd) {
  assert(fd > -1);

  while(close(fd) < 0) {
    int err = errno;
    if (err != EINTR) {
      fprintf(stderr, "%s: close failed on file descriptor.  %s.", program_invocation_short_name, strerror(err));
      return;
    }
  }
}

int main(int argc, char** argv) {
  int nextOption, retVal = 1, k, fd, dfd, length = 0, mLength;;
  struct sigaction sa;
  struct stat fInfo, dInfo;
  char *path, *fmap, *name, *fsave = NULL;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = &SIGTERM_Handler;
  while (sigaction(SIGTERM, &sa, NULL) == -1) {
    if (errno != EINTR) {
      retVal = errno;
      fprintf(stderr, "%s: sigaction failed for SIGTERM.  %s.\n\n", 
	      program_invocation_short_name, strerror(retVal));
      exit(2);
    }
  }  
  while (sigaction(SIGINT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      retVal = errno;
      fprintf(stderr, "%s: sigaction failed for SIGINT.  %s.\n\n", 
	      program_invocation_short_name, strerror(retVal));
      exit(2);
    }
  }  
  while (sigaction(SIGQUIT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      fprintf(stderr, "%s: sigaction failed for SIGQUIT.  %s.\n\n", 
	      program_invocation_short_name, strerror(retVal));
      exit(2);
    }
  }  
  while (sigaction(SIGHUP, &sa, NULL) == -1) {
    if (errno != EINTR) {
      fprintf(stderr, "%s: sigaction failed for SIGHUP.  %s.\n\n", 
	      program_invocation_short_name, strerror(retVal));
      exit(2);
    }
  }
  memset(&sa, 0, sizeof(sa));
  sa.sa_sigaction = SIGRTMIN_Handler;
  sa.sa_flags = SA_SIGINFO;
  while (sigaction(SIGRTMIN, &sa, NULL) == -1) {
    if (errno != EINTR) {
      fprintf(stderr, "%s: sigaction failed for SIGRTMIN.  %s.\n\n", program_invocation_short_name, strerror(retVal));
      exit(2);
    }
  }

  do {
    nextOption = getopt_long(argc, argv, ShortOptions, LongOptions, NULL);
    switch(nextOption) {
    case 'h':
      PrintUsage(0);
      break;
    case '?':
      fprintf(stderr, "%s: Invalid option specified.\n\n", program_invocation_short_name);
      PrintUsage(5);
      break;
    case -1:
      break;
    }
  } while (nextOption != -1);

  if (argv[optind] == NULL) {
    fprintf(stderr, "%s: No file specified.\n\n", program_invocation_short_name);
    PrintUsage(6);
  }
  length = strlen(argv[optind]);
  if (argv[optind][length - 1] == '/')
    argv[optind][length - 1] = '\0';
  fmap = strrchr(argv[optind], '/');
  if (fmap == NULL)
    fmap = argv[optind];
  if (fmap == argv[optind]) {
    if ((path = malloc(sizeof(char) * 2)) == NULL) {
      fprintf(stderr, "%s: malloc failed on path.\n\n", program_invocation_short_name);
      exit(1);
    }
    sprintf(path, ".");
  }
  else if ((path = strndup(argv[optind], fmap - argv[optind])) == NULL) {
    fprintf(stderr, "%s: strdup failed on path.\n\n", program_invocation_short_name);
    exit(1);
  }
  if (fmap[0] == '/')
    fmap++;
  if ((name = strndup(fmap, strlen(argv[optind]) - strlen(fmap))) == NULL) {
    fprintf(stderr, "%s: strdup failed on name.\n\n", program_invocation_short_name);
    free(path);
    exit(1);
  }
  
  if (argv[optind + 1] != NULL) {
    fprintf(stderr, "%s: To many  arguments specified.\n\n", program_invocation_short_name);
    free(name);
    free(path);
    PrintUsage(7);
  }
  
  if ((dfd = open(path, O_RDONLY)) == -1) {
    retVal = errno;
    fprintf(stderr, "%s: open failed on %s: %s.\n\n", program_invocation_short_name, path,
	    strerror(retVal));
    free(name);
    free(path);
    exit(8);
  }
  if (fstat(dfd, &dInfo) == -1) {
    retVal = errno;
    fprintf(stderr, "%s: fstat failed on %s: %s.\n\n", program_invocation_short_name, path,
	    strerror(retVal));
    Cclose(dfd);
    free(name);
    free(path);
    exit(9);
  }
  if (!S_ISDIR(dInfo.st_mode)) {
    fprintf(stderr, "%s: %s is not a valid directory.\n\n", program_invocation_short_name, path);
    Cclose(dfd);
    free(name);
    free(path);
    exit(10);
  }
  if (fcntl(dfd, F_SETSIG, SIGRTMIN) == -1) {
    retVal = errno;
    fprintf(stderr, "%s: fcntl failed on F_SETSIG.  %s.\n\n", program_invocation_short_name,
	    strerror(retVal));
    free(fsave);
    free(fmap);
    Cclose(dfd);
    free(name);
    free(path);
    exit(13);
  }
  if (fcntl(dfd, F_NOTIFY, DN_MODIFY | DN_MULTISHOT) == -1) {
    retVal = errno;
    fprintf(stderr, "%s: fcntl failed on F_NOTIFY.  %s.\n\n", program_invocation_short_name,
	    strerror(retVal));
    free(fsave);
    free(fmap);
    Cclose(dfd);
    free(name);
    free(path);
    exit(13);
  }
  for (errno = 0; (fd = open(argv[optind], O_RDONLY)) == -1 && errno == EINTR;);
  if (fd != -1) {
    do {
      if (fstat(fd, &fInfo) == -1) {
	retVal = errno;
	fprintf(stderr, "%s: fstat failed on %s:  %s.\n\n", program_invocation_short_name, path,
		strerror(retVal));
	Cclose(fd);
	Cclose(dfd);
	free(name);
	free(path);
	exit(9);
      }
    } while (!fInfo.st_size);
    if (!S_ISREG(fInfo.st_mode)) {
      fprintf(stderr, "%s:  %s is not a regular file.\n\n", program_invocation_short_name, path);
      Cclose(fd);
      Cclose(dfd);
      free(name);
      free(path);
      exit(12);
    }
    
    mLength = length = fInfo.st_size;
    if ((fmap = malloc(sizeof(char) * length)) == NULL) {
      fprintf(stderr, "%s: malloc failed on fmap.\n\n", program_invocation_short_name);
      Cclose(fd);
      Cclose(dfd);
      free(name);
      free(path);
      exit(1);
    }
    if ((fsave = malloc(sizeof(char) * length)) == NULL) {
      fprintf(stderr, "%s: malloc failed on fsave.\n\n", program_invocation_short_name);
      free(fmap);
      Cclose(fd);
      Cclose(dfd);
      free(name);
      free(path);
      exit(1);
    }
    for (k = 0; k < length;) {
      retVal = read(fd, &fsave[k], length - k);
      if (retVal == -1) {
	retVal = errno;
	if (retVal != EINTR) {
	  fprintf(stderr, "%s: read failed on fd.  %s.\n\n", program_invocation_short_name, strerror(retVal));
	  free(fsave);
	  free(fmap);
	  Cclose(fd);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(15);
	}
      }
      else
	k += retVal;
    }
    lseek(fd, 0, SEEK_SET);
    Cclose(fd);

    k = 0;
    while (k < 4) {
      retVal = write(STDOUT_FILENO, &((char*)&length)[k], 4 - k);
      if (retVal == -1) {
	retVal = errno;
	if (retVal != EINTR) {
	  fprintf(stderr, "%s: write failed on stdout.  %s.\n\n", program_invocation_short_name,
		  strerror(retVal));
	  free(fsave);
	  free(fmap);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(14);
	}
	else if (Closing) {
	  free(fsave);
	  free(fmap);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(0);
	}
      }
      else
	k += retVal;
    }	
    k = 0;
    while (k < length) {
      retVal = write(STDOUT_FILENO, &fsave[k], length - k);
      if (retVal == -1) {
	retVal = errno;
	if (retVal != EINTR) {
	  fprintf(stderr, "%s: write failed on stdout.  %s.\n\n", program_invocation_short_name,
		  strerror(retVal));
	  free(fsave);
	  free(fmap);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(14);
	}
	else if (Closing) {
	  free(fsave);
	  free(fmap);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(0);
	}
      }
      else
	k += retVal;
    }
  }
  else {
    if (errno == ENOENT) {
      mLength = length = 1;
      if ((fmap = malloc(sizeof(char) * length)) == NULL) {
	fprintf(stderr, "%s: malloc failed on fmap.\n\n", program_invocation_short_name);
	Cclose(fd);
	Cclose(dfd);
	free(name);
	free(path);
	exit(1);
      }
      if ((fsave = malloc(sizeof(char) * length)) == NULL) {
	fprintf(stderr, "%s: malloc failed on fsave.\n\n", program_invocation_short_name);
	free(fmap);
	Cclose(fd);
	Cclose(dfd);
	free(name);
	free(path);
	exit(1);
      }
    }
    else {
      retVal = errno;
      fprintf(stderr, "%s: open failed on %s.  %s.\n\n", program_invocation_short_name, argv[optind], strerror(retVal));
      Cclose(dfd);
      free(name);
      free(path);
      exit(11);
    }
  }
  
  while (!Closing) {
    while (EventFD != -1) {
      EventFD = -1;

      for (errno = 0; (fd == open(argv[optind], O_RDONLY)) == -1 && errno == EINTR;);
      if (fd != -1) {
	do {
	  if (fstat(fd, &fInfo) == -1) {
	    retVal = errno;
	    fprintf(stderr, "%s: fstat failed on %s:  %s.\n\n", program_invocation_short_name, path,
		    strerror(retVal));
	    Cclose(fd);
	    Cclose(dfd);
	    free(name);
	    free(path);
	    exit(9);
	  }
	} while (!fInfo.st_size);
	if (!S_ISREG(fInfo.st_mode)) {
	  fprintf(stderr, "%s:  %s is not a regular file.\n\n", program_invocation_short_name, path);
	  Cclose(fd);
	  Cclose(dfd);
	  free(name);
	  free(path);
	  exit(12);
	}
	
	nextOption = 0;
	if (length != fInfo.st_size) {
	  length = fInfo.st_size;
	  if (mLength < length) {
	    free(fmap);
	    mLength = length;
	    if ((fmap = malloc(sizeof(char) * length)) == NULL) {
	      fprintf(stderr, "%s: malloc failed on fmap.\n\n", program_invocation_short_name);
	      free(fsave);
	      Cclose(fd);
	      Cclose(dfd);
	      free(name);
	      free(path);
	      exit(1);
	    }
	    free(fsave);
	    if ((fsave = malloc(sizeof(char) * length)) == NULL) {
	      fprintf(stderr, "%s: malloc failed on fsave.\n\n", program_invocation_short_name);
	      free(fmap);
	      Cclose(fd);
	      Cclose(dfd);
	      free(name);
	      free(path);
	      exit(1);
	    }
	  }
	  for (k = 0; k < length;) {
	    retVal = read(fd, &fsave[k], length - k);
	    if (retVal == -1) {
	      retVal = errno;
	      if (retVal != EINTR) {
		fprintf(stderr, "%s: read failed on fd.  %s.\n\n", program_invocation_short_name, strerror(retVal));
		free(fsave);
		free(fmap);
		Cclose(fd);
		Cclose(dfd);
		free(name);
		free(path);
		exit(15);
	      }
	    }
	    else
	      k += retVal;
	  }
	  nextOption = 1;
	}
	else {
	  for (k = 0; k < length;) {
	    retVal = read(fd, &fmap[k], length - k);
	    if (retVal == -1) {
	      retVal = errno;
	      if (retVal != EINTR) {
		fprintf(stderr, "%s: read failed on fd.  %s.\n\n", program_invocation_short_name, strerror(retVal));
		free(fsave);
		free(fmap);
		Cclose(fd);
		Cclose(dfd);
		free(name);
		free(path);
		exit(15);
	      }
	    }
	    else
	      k += retVal;
	  }
	  if ((nextOption = memcmp(fmap, fsave, length)))
	    memmove(fsave, fmap, length);
	}
	lseek(fd, 0, SEEK_SET);
	Cclose(fd);

	if (nextOption) {
	  k = 0;
	  while (k < 4) {
	    retVal = write(STDOUT_FILENO, &((char*)&length)[k], 4 - k);
	    if (retVal == -1) {
	      retVal = errno;
	      if (retVal != EINTR) {
		fprintf(stderr, "%s: write failed on stdout.  %s.\n\n", program_invocation_short_name,
			strerror(retVal));
		free(fsave);
		free(fmap);
		Cclose(dfd);
		free(name);
		free(path);
		exit(14);
	      }
	      else if (Closing) {
		free(fsave);
		free(fmap);
		Cclose(dfd);
		free(name);
		free(path);
		exit(0);
	      }
	    }
	    else
	      k += retVal;
	  }
	  k = 0;
	  while (k < length) {
	    retVal = write(STDOUT_FILENO, &fsave[k], length - k);
	    if (retVal == -1) {
	      retVal = errno;
	      if (retVal != EINTR) {
		fprintf(stderr, "%s: write failed on stdout.  %s.\n\n", program_invocation_short_name,
			strerror(retVal));
		free(fsave);
		free(fmap);
		Cclose(dfd);
		free(name);
		free(path);
		exit(14);
	      }
	      else if (Closing) {
		free(fsave);
		free(fmap);
		Cclose(dfd);
		free(name);
		free(path);
		exit(0);
	      }
	    }
	    else
	      k += retVal;
	  }
	}
      }
      else if (errno != ENOENT) {
	retVal = errno;
	fprintf(stderr, "%s: open failed on %s.  %s.\n\n", program_invocation_short_name, argv[optind], strerror(retVal));
	free(fsave);
	free(fmap);
	Cclose(dfd);
	free(name);
	free(path);
	exit(11);
      }
    }

    pause();
  }

  free(fsave);
  free(fmap);
  Cclose(dfd);
  free(name);
  free(path);

  return 0;
}
