/******************************************************************************\
 * Name: fmon_handler.c                                                       *
 * Author: James Greensky                                                     *
 * Notes: Implementation of the File Monitor handler                          *
\******************************************************************************/

#define _GNU_SOURCE

#include "server.h"
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

typedef enum FmonResponse {
  FmonOK = 0,
  FmonEMemErr = 1,
  FmonSigHErr = 2,
  FmonOptErr = 5,
  FmonInputErr = 6,
  FmonBadInput = 7,
  FmonDOpenErr = 8,
  FmonStatErr = 9,
  FmonNDirErr = 10,
  FmonFOpenErr = 11,
  FmonNFileErr = 12,
  FmonFcntlErr = 13,
  FmonEWriteErr = 14,
  FmonMapErr = 15,
  FmonMAdviseErr = 16,
  FmonUMapErr = 17,
  FmonMemErr = 18,
  FmonPipeErr = 19,
  FmonIOErr = 20,
  FmonExecErr = 21,
  FmonForkErr = 22,
  FmonReadErr = 23,
  FmonWriteErr = 24,
  FmonError = 25,
} FmonResponse;

typedef struct FmonClientRecord {
  LcseExecutable Executable;
  pid_t ExecutableID;
  LcseClient Client;
  CBuffer Buffer;
  struct FmonClientRecord* Next;
}* FmonClient;

extern Log Logger;
static void* FmonHandle = NULL;
static FmonClient FmonClients = NULL;

void Initialize(void* handle) {
  assert(handle != NULL);

  FmonHandle = handle;
}

inline static int ClientOutput(FmonClient client, void* buffer, int count) {
  assert(client != NULL);
  assert(buffer != NULL);
  assert(count > -1);

  if (CBufferWrite(client->Client->OutBuffer, buffer, count) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on client->Client->OutBuffer.");
    CloseClient(client->Client);
    return 0;
  }
  return 1;
}

inline static int ClientReply(FmonClient client, FmonResponse response) {
  assert(client != NULL);

  int reply = htonl(response);
  return ClientOutput(client, &reply, 4);
}

static void AddClient(const LcseClient client) {
  assert(client != NULL);

  FmonClient fmonClient = malloc(sizeof(struct FmonClientRecord));
  if (fmonClient == NULL) {
    int response = htonl(FmonMemErr);
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on fmonClient.");
    CloseClient(client);
    return;
  }
  fmonClient->Buffer = CBufferCreate(0, 1);
  if (fmonClient->Buffer == NULL) {
    int response = htonl(FmonMemErr);
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create fmonClient->Buffer failed.");
    free(fmonClient);
    CloseClient(client);
    return;
  }

  fmonClient->Client = client;
  fmonClient->ExecutableID = -1;
  fmonClient->Executable = NULL;
  fmonClient->Next = FmonClients;
  FmonClients = fmonClient;
  ClientReply(fmonClient, FmonOK);
}

static void RemoveClient(const LcseClient client) {
  assert(client != NULL);

  FmonClient current = NULL, previous = NULL;
  for (previous = NULL, current = FmonClients; current != NULL && current->Client != client;
       previous = current, current = current->Next);
  if (current != NULL) {
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Removing fmon client.");

    if (previous == NULL)
      FmonClients = current->Next;
    else
      previous->Next = current->Next;
    
    CBufferDispose(current->Buffer);
    
    if (current->ExecutableID > 0)
      kill(current->ExecutableID, SIGTERM);
    else if (current->Executable != NULL)
      kill(current->Executable->ProcessID, SIGTERM);

    free(current);
  }
}

void Dispose() {
  for (;FmonClients != NULL;)
    RemoveClient(FmonClients->Client);
}

struct timeval* ClientHandler(const LcseClient client, LcseClientEvent event) {
  assert(client != NULL);

  switch(event) {
  case ClientAccepted:
    AddClient(client);
    break;
  case ClientDataReceived: {
    FmonClient current = NULL;
    for (current = FmonClients; current != NULL && current->Client != client; current = current->Next);
    if (current != NULL) {
      if (current->ExecutableID == -1) {
	char *buffer = NULL, *ptr = NULL;
	int count = 0;
	while (CBufferSize(client->InBuffer)) {
	  buffer = CBufferGetWriteBlock(current->Buffer, &count);
	  if (count > CBufferSize(client->InBuffer))
	    count = CBufferSize(client->InBuffer);
	  CBufferRead(client->InBuffer, buffer, count);
	  ptr = memrchr(buffer, '\0', count);
	  if (ptr != NULL) {
	    char* argv[3];
	    CBufferCommitBlock(current->Buffer, ptr - buffer + 1);
	    if (CBufferSize(client->InBuffer))
	      CBufferRemove(client->InBuffer, CBufferSize(client->InBuffer));
	    count = strlen(ModuleDirectory);
	    argv[0] = malloc(sizeof(char) * strlen(ModuleDirectory) + sizeof(char) * 18);
	    if (argv[0] == NULL) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "malloc failed on argv[0].");
	      ClientReply(current, FmonMemErr);
	      CloseClient(client);
	    }
	    
	    sprintf(argv[0], "%s/%s", ModuleDirectory, "file_monitor.exe");
	    argv[1] = malloc(sizeof(char) * CBufferSize(current->Buffer));
	    if (argv[1] == NULL) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "malloc failed on argv[1].");
	      ClientReply(current, FmonMemErr);
	      CloseClient(client);
	    }
	    CBufferRead(current->Buffer, argv[1], CBufferSize(current->Buffer));
	    argv[2] = NULL;
	    
	    current->ExecutableID = SpawnExecutable(FmonHandle, argv);
	    free(argv[1]);
	    free(argv[0]);
	  }
	}
      }
      else
	CBufferRemove(client->InBuffer, CBufferSize(client->InBuffer));
    }
    else
      CloseClient(client);
  }
    break;
  case ClientDisconnected:
    RemoveClient(client);
    break;
  }

  return NULL;
}

static void ProcessData(const LcseExecutable executable) {
  assert(executable != NULL);

  FmonClient current = NULL;
  for (current = FmonClients; current != NULL && current->Executable != executable &&
	 current->ExecutableID != executable->ProcessID; current = current->Next);
  if (current != NULL) {
    char* buffer = NULL;
    int length = 0, retVal = 1, count = 0, total = 0;;
    
    if (current->Executable == NULL) {
      int err = EINTR;
      length = 1;
      buffer = CBufferGetWriteBlock(current->Buffer, &length);
      while (read(executable->StdOut, buffer, length) < 1 && err == EINTR) {
	err = errno;
	if (err != EINTR) {
	  if (err != EAGAIN) {
	    if (LogIsError(Logger))
	      LogMessage(Logger, LogError, "read failed on executable->StdOut.  %s.", strerror(err));
	    kill(current->ExecutableID, SIGTERM);
	  }
	}
      }
      if (err == EINTR) {
	current->Executable = executable;
	current->ExecutableID = -1;
      }
    }
    if (current->ExecutableID == -1) {
      current->ExecutableID = 0;
      ClientReply(current, FmonOK);
    }
    retVal = 1;
    while (retVal > 0) {
      count = 0;
      buffer = CBufferGetWriteBlock(current->Buffer, &count);
      if (buffer == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "getwriteblock failed on current->Buffer.");
	ClientReply(current, FmonMemErr);
	CloseClient(current->Client);
	return;
      }
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "reading from current->Executable->StdOut to current->Buffer.");
      total += retVal = read(executable->StdOut, buffer, count);
      if (retVal < 0) {
	int err = errno;
	if (err == EINTR)
	  retVal = 1;
	else if (err != EAGAIN) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "read failed on current->Executable->StdOut.");
	  kill(executable->ProcessID, SIGTERM);
	  ClientReply(current, FmonReadErr);
	  CloseClient(current->Client);
	  return;
	}
      }
      else if (retVal > 0) {
	CBufferCommitBlock(current->Buffer, retVal);
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "read %d bytes from current->Executable->StdOut.", retVal);
      }
    }
    
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "read a total of %d bytes from current->Executable->StdOut.", total);
    
    retVal = 1;
    while (CBufferSize(current->Buffer) > 3 && retVal) {
      CBufferPeek(current->Buffer, &length, 4);
      if (CBufferSize(current->Buffer) >= length + 4) {
	length = htonl(length);
	if (!ClientOutput(current, &length, 4))
	  return;
	length = ntohl(length);
	CBufferRemove(current->Buffer, 4);
	for (total = 0; total < length;) {
	  count = length - total;
	  buffer = CBufferGetWriteBlock(current->Client->OutBuffer, &count);
	  if (buffer == NULL) {
	    if (LogIsError(Logger))
	      LogMessage(Logger, LogError, "getwriteblock failed current->Client->OutBuffer.");
	    ClientReply(current, FmonMemErr);
	    CloseClient(current->Client);
	    return;
	  }
	  CBufferRead(current->Buffer, buffer, count);
	  CBufferCommitBlock(current->Client->OutBuffer, count);
	  total += count;
	}
      }
      else
	retVal = 0;
    }
  }
  else
    kill(executable->ProcessID, SIGTERM);
}
  
struct timeval* ExecutableHandler(const LcseExecutable executable, LcseExecutableEvent event) {
  assert(executable != NULL);

  switch(event) {
  case ExecutableDataReady:
    ProcessData(executable);
    break;
  case ExecutableWriteReady:
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "ExecutableHandler called in fmon with ExecutableWriteReady.");
    assert(1);
    break;
  case ExecutableExited: {
    FmonClient current;
    for (current = FmonClients; current != NULL && current->Executable != executable &&
	   current->ExecutableID != executable->ProcessID; current = current->Next);
    if (current != NULL) {
      if (executable->Status && LogIsError(Logger))
	LogMessage(Logger, LogError, "executable exited with a status of %d.", executable->Status);
      ClientReply(current, FmonError);
      CloseClient(current->Client);
    }
  }
    break;
  }

  return NULL;
}

struct timeval* TimeoutHandler() {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "TimoutHandler called in fmon.");
  assert(1);

  return NULL;
}

struct timeval* IOHandler(int fd, LcseIOEvent event) {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "IOHandler called in fmon.");
  assert(1);

  return NULL;
}
