/******************************************************************************\
 * Name: hvr_handler.c                                                        *
 * Author: James Greensky                                                     *
 * Notes: Implementation of the Higherarchial Volume Renderer handler         *
\******************************************************************************/

#include "server.h"
#include "queue.h"
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>

typedef enum HvrCommands {
  HvrAddKey = 0,
  HvrAddFwd = 1,
  HvrRmvFwd = 2,
  HvrGetLog = 3
} HvrCommand;

typedef enum HvrResponses {
  HvrOK = 0,
  HvrXErr = 1,
  HvrVizErr = 2,
  HvrHvErr = 3,
  HvrFBOErr = 4,
  HvrCacheErr = 5,
  HvrCgErr = 6,
  HvrLogErr = 7,
  HvrGlewErr = 8,
  HvrIPrntErr = 9,
  HvrBadHvFile = 10,
  HvrSigTerm = 11,
  HvrClientErr = 12,
  HvrFwdIOFail = 13,
  HvrFwdMemFail = 14,
  HvrFwdIllForm = 15,
  HvrFwdLibNFnd = 16,
  HvrFwdSymNFnd = 17,
  HvrFwdOK = 18,
  HvrFwdEMemErr = 19,
  HvrFwdProcErr = 20,
  HvrFwdHNameErr = 21,
  HvrFwdLogErr = 22,
  HvrFwdXOpenErr = 23,
  HvrFwdXFatalErr = 24,
  HvrFwdEReadErr = 25,
  HvrFwdStdInClosd = 26,
  HvrFwdEWriteErr = 27,
  HvrFwdArgInvalid = 28,
  HvrFwdArgUnknown = 29,
  HvrFwdVizErr = 30,
  HvrFwdSigErr = 31,
  HvrFwdXContextErr = 32,
  HvrFwdXCurrentErr = 33,
  HvrFwdXConnErr = 34,
  HvrFwdShmErr = 35,
  HvrFwdShmAttErr = 36,
  HvrFwdMemErr = 37,
  HvrFwdPipeErr = 38,
  HvrFwdIOErr = 39,
  HvrFwdExecErr = 40,
  HvrFwdForkErr = 41,
  HvrFwdReadErr = 42,
  HvrFwdWriteErr = 43,
  HvrFwdInUse = 44,
  HvrFwdNFnd = 45,
  HvrFwdTimeErr = 46,
  HvrFwdErr = 47,
  HvrMemErr = 48,
  HvrPipeErr = 49,
  HvrIOErr = 50,
  HvrForkErr = 51,
  HvrExecErr = 52,
  HvrReadErr = 53,
  HvrWriteErr = 54,
  HvrSockErr = 55,
  HvrConnErr = 56,
  HvrConnClsd = 57,
  HvrNFnd = 58,
  HvrNConn = 59,
  HvrDup = 60,
  HvrSysErr = 61,
  HvrError = 62
} HvrResponse;

typedef struct HvrEndPointRecord {
  uint32_t Address;
  uint16_t Port;
}* HvrEndPoint;

typedef struct HvrForwardRecord {
  int SocketFD, Offset, Connected, Response, Key;
  CBuffer InBuffer;
  HvrEndPoint RemoteEP;
  struct HvrForwardRecord* Next;
}* HvrForward;

struct HvrAsyncIORecord;
struct HvrJobRecord;
typedef struct HvrClientRecord {
  LcseClient Client;
  HvrForward Forwards;
  struct HvrAsyncIORecord* AsyncIO;
  struct HvrJobRecord* Jobs;
  struct HvrClientRecord* Next;
}* HvrClient;

typedef struct HvrAsyncIORecord {
  HvrClient Client;
  int Fds[2];
  void* Data;
  HvrCommand Command;
  struct HvrAsyncIORecord* Next;
}* HvrAsyncIO;

typedef struct HvrJobRecord {
  HvrClient Client;
  char Transmit;
  char* KeyFrame;
  HvrEndPoint Forwards;
  struct timeval Submitted, Created, Finished;
  int Offset, Length, SequenceNumber, Status;
  int ImageX, ImageY, ImageWidth, ImageHeight, PanelWidth, PanelHeight;
  int SectionX, SectionY, SectionMaxX, SectionMaxY;
  struct HvrJobRecord* Next;
}* HvrJob;

extern Log Logger;
static void* HvrHandle = NULL;
static Queue HvrQ = NULL;
static HvrClient HvrClients = NULL;
static const char ImageViewerLibrary[] = "iv_handler.so";

static pid_t HvrID = 0;
static LcseExecutable HvrExecutable = NULL;
static CBuffer HvrBuffer = NULL;
static int HvrTotal = -1, ReadErrors = 0, WriteErrors = 0;

void Initialize(void* handle) {
  assert(handle != NULL);

  char* argv[2];
  argv[0] = malloc(sizeof(char) * strlen(ModuleDirectory) + 12);
  if (argv[0] == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on argv[0].");
    HvrID = -2;
    return;
  }
  sprintf(argv[0], "%s/%s", ModuleDirectory, "hvr_cl.exe");
  argv[1] = NULL;

  HvrBuffer = CBufferCreate(0, 1);
  if (HvrBuffer == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create hvr buffer failed.");
    free(argv[0]);
    HvrID = -2;
    return;
  }
  HvrQ = QueueCreate();
  if (HvrQ == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create hvr q failed.");
    CBufferDispose(HvrBuffer);
    free(argv[0]);
    HvrID = -2;
    return;
  }

  HvrHandle = handle;

  HvrID = SpawnExecutable(handle, argv);

  free(argv[0]);
}

static void DisposeForward(HvrForward forward) {
  assert(forward != NULL);

  if (IsSelectIOSet(forward->SocketFD, SelectWrite))
    RemoveSelectIO(HvrHandle, forward->SocketFD, SelectWrite);
  if (IsSelectIOSet(forward->SocketFD, SelectRead))
    RemoveSelectIO(HvrHandle, forward->SocketFD, SelectRead);

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Disposing forward record.");

  CBufferDispose(forward->InBuffer);
  free(forward->RemoteEP);
  Cclose(forward->SocketFD);
  free(forward);
}

static void DisposeJob(HvrJob job) {
  assert(job != NULL);
  
  free(job->KeyFrame);
  free(job->Forwards);
  free(job);
}

static void RemoveJob(HvrJob job) {
  assert(job != NULL);

  if (job->Client != NULL) {
    HvrJob previous = NULL, current = NULL;
    for (previous = NULL, current = job->Client->Jobs; current != NULL && current != job;
	 previous = current, current = current->Next);
    if (current != NULL) {
      if (previous == NULL)
	job->Client->Jobs = current->Next;
      else
	previous->Next = current->Next;
    }
  }
  DisposeJob(job);
}

static void RemoveClient(const LcseClient client) {
  assert(client != NULL);

  HvrClient current = NULL, previous = NULL;
  for (previous = NULL, current = HvrClients; current != NULL && current->Client != client;
       previous = current, current = current->Next);
  if (current != NULL) {
    HvrJob job = NULL;
    HvrForward forward = NULL;
    void* save;

    for (job = current->Jobs; job != NULL;) {
      job->Client = NULL;
      if (QueuePeek(HvrQ) != job) {
	save = job->Next;
	QueueRemove(HvrQ, job);
	DisposeJob(job);
	job = save;
      }
      else
	job = job->Next;
    }

    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Removing client.");

    for (forward = current->Forwards; forward != NULL;) {
      save = forward->Next;
      DisposeForward(forward);
      forward = save;
    }

    if (previous == NULL)
      HvrClients = current->Next;
    else
      previous->Next = current->Next;
    free(current);
  }
}

void Dispose() {
  void* save;
  HvrClient client = NULL;
  HvrJob job = NULL;
  HvrForward forward = NULL;

  if (HvrID > 0)
    kill(HvrID, SIGTERM);
  else if (HvrExecutable != NULL)
    kill(HvrExecutable->ProcessID, SIGTERM);

  if (QueueSize(HvrQ)) {
    job = QueuePeek(HvrQ);
    if (job->Client == NULL)
      DisposeJob(job);
  }

  for (client = HvrClients; client != NULL;) {
    for (job = client->Jobs; job != NULL;) {
      save = job->Next;
      DisposeJob(job);
      job = save;
    }
    
    for (forward = client->Forwards; forward != NULL;) {
      save = forward->Next;
      DisposeForward(forward);
      forward = save;
    }
    
    save = client->Next;
    free(client);
    client = save;
  }
  
  CBufferDispose(HvrBuffer);
  QueueDispose(HvrQ);
}

inline static int ClientOutput(HvrClient client, void* buffer, int count) {
  assert(client != NULL);
  assert(buffer != NULL);
  assert(count > -1);

  if (CBufferWrite(client->Client->OutBuffer, buffer, count) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on client->Client->OutBuffer.");
    CloseClient(client->Client);
    return 0;
  }
  return 1;
}

inline static int JobOutput(HvrJob job, void* buffer, int count) {
  assert(job != NULL);

  if (job->Client != NULL)
    return ClientOutput(job->Client, buffer, count);
  return 0;
}

inline static int ClientReply(HvrClient client, HvrResponse response) {
  assert(client != NULL);

  int reply = htonl(response);
  return ClientOutput(client, &reply, 4);
}

inline static int JobReply(HvrJob job, HvrResponse response) {
  assert(job != NULL);

  if (job->Client != NULL)
    return ClientReply(job->Client, response);
  return 0;
}

static void StartJob() {
  assert(HvrExecutable != NULL);
  assert(QueueSize(HvrQ));

  HvrJob job = QueuePeek(HvrQ);
  int length = job->Length - job->Offset, retVal, multi = job->Offset, total = 0;
  if (!multi)
    HvrTotal = -1;

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "starting new job.");

  while (length > 0) {
    total += retVal = write(HvrExecutable->StdIn, &job->KeyFrame[job->Offset], length);
    if (retVal > 0 && LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "wrote %d bytes to HvrExecutable->StdIn", retVal);
    if (retVal > -1) {
      if (retVal > 0)
	WriteErrors = 0;
      if (!(length -= retVal)) {
	if (gettimeofday(&job->Submitted, NULL) && LogIsError(Logger)) {
	  int err = errno;
	  LogMessage(Logger, LogError, "gettimeofday failed on job->Submitted.  %s.", strerror(err));
	}
	if (multi)
	  RemoveExecutableStdIn(HvrExecutable);
      }
      job->Offset += retVal;
    }
    else {
      retVal = errno;
      if (retVal != EINTR) {
	if (retVal == EAGAIN) {
	  if (!multi)
	    AddExecutableStdIn(HvrExecutable);
	  length = 0;
	}
	else {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "write failed on HvrExecutable->StdIn.  %s.", 
		       strerror(retVal));
	  if (multi)
	    RemoveExecutableStdIn(HvrExecutable);
	  WriteErrors++;
	  kill(HvrExecutable->ProcessID, SIGTERM);
	  QueueDequeue(HvrQ);
	  JobReply(job, HvrWriteErr);
	  RemoveJob(job);
	  length = -1;
	}
      }
    }
  }
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "wrote a total of %d bytes to HvrExecutable->StdIn, %d left.", total, job->Length - job->Offset);
}

struct timeval* TimeoutHandler() {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "TimeoutHandler called in hvr.");
  assert(1);
 
  return NULL;
}

inline static void JobCleanUp(int startJob) {
  HvrJob job = QueueDequeue(HvrQ);
  RemoveJob(job);
  CBufferClear(HvrBuffer);

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Job cleanup.");

  if (startJob && QueueSize(HvrQ))
    StartJob();
}

static int SocketSend(int socketFD, void* buffer, int length) {
  assert(socketFD > -1);
  assert(buffer != NULL);

  int retVal = 1, err = EINTR, count = 0, total = 0;

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "length = %d.", length);
  
  while (length > total && retVal > -1) {
    count = length - total;
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "count = %d, total = %d.", count, total);
    retVal = send(socketFD, buffer, count, 0);
    if (retVal < 0) {
      err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else {
      if (LogIsDebug(Logger)) {
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
	  LogMessage(Logger, LogDebug, "Sent %d bytes to %s:%u.", retVal,
		     inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	else
	  LogMessage(Logger, LogDebug, "Sent %d bytes.", retVal);
      }
      total += retVal;
    }
  }

  if (retVal < 0) {
    if (err != EAGAIN) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "send failed to socketFD,  %s.", strerror(err));
      return -2;
    }
  }

  if (LogIsDebug(Logger)) {
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);
    if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
      LogMessage(Logger, LogDebug, "Sent a total of %d bytes to %s:%u.", retVal,
		 inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
    else
      LogMessage(Logger, LogDebug, "Sent a total of %d bytes.", retVal);
  }

  return total;
}

static int SocketSendBuffer(int socketFD, CBuffer buffer, int offset) {
  assert(socketFD > -1);
  assert(buffer != NULL);

  int retVal = 1, err = EINTR, count = 0, total = 0, flags = 0;
  char* buff = NULL;
  
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "offset = %d, MSG_NOSIGNAL = %d, MSG_MORE = %d.", offset, MSG_NOSIGNAL, MSG_MORE);

  while (CBufferSize(buffer) > total + offset && retVal > -1) {
    count = 0;
    buff = CBufferGetReadBlockAt(buffer, total + offset, &count);
    flags = MSG_NOSIGNAL | ((CBufferSize(buffer) - total - offset - count > 0) ? MSG_MORE : 0);
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "count = %d, total = %d, bufferSize = %d, flags = %d.", 
      count, total, CBufferSize(buffer), flags);
    retVal = send(socketFD, buff, count, flags);
    if (retVal < 0) {
      err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else {
      /* if (LogIsDebug(Logger)) {
	char* strbuffer = malloc((retVal * 4) + (((retVal / 16) + 1) * 8) + 33); 
	int m, n;
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(socketFD, (struct sockaddr *) &addr, &len) == -1) {
	  sprintf(strbuffer, "Sent -");
	  n = 6;
	}
	else {
	  sprintf(strbuffer, "Sent to %s:%05u -", 
		  inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	  n = 16 + strlen(inet_ntoa(addr.sin_addr));
	}
	for (m = 0; m < retVal; m++) {
	  if (m == 0 || (((m + total + offset) % 16) == 0)) {
	    sprintf(&strbuffer[n], "\n%07d", m + total + offset);
	    n+=8;
	  }
	  sprintf(&strbuffer[n], "%4u,", (unsigned char)buff[m]);
	  n+=4;
	}
	sprintf(&strbuffer[n], ".");
	strbuffer[++n] = '\0';
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "%s", strbuffer);
	  free(strbuffer);
	  }*/
      if (LogIsDebug(Logger)) {
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
	  LogMessage(Logger, LogDebug, "Sent %d bytes to %s:%u.", retVal,
		     inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	else
	  LogMessage(Logger, LogDebug, "Sent %d bytes.", retVal);
      }
      total += retVal;
    }
  }

  if (retVal < 0) {
    if (err != EAGAIN) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "send failed to socketFD.  %s.", strerror(err));
      return -2;
    }
  }

  if (LogIsDebug(Logger)) {
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);
    if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
      LogMessage(Logger, LogDebug, "Sent %d bytes to %s:%u starting at %d for a total of %d.",
		 total, inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port), 
		 offset, total + offset);
    else
      LogMessage(Logger, LogDebug, "Sent %d bytes starting at %d for a total of %d.",
		 total, offset, total + offset);
  }
  
  return total;
}

static int SocketReceiveBuffer(int socketFD, CBuffer buffer) {
  assert(socketFD > -1);
  assert(buffer != NULL);
  
  int retVal = 1, err = EINTR, total = 0, count = 0;
  char* buff = NULL;
  while (retVal > 0) {
    count = 0;
    buff = CBufferGetWriteBlock(buffer, &count);
    if (buff == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "get write block failed on buffer.");
      return -3;
    }
    retVal = recv(socketFD, buff, count, 0);
    if (retVal < 0) {
      err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else {
      if (LogIsDebug(Logger)) {
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
	  LogMessage(Logger, LogDebug, "Received %d bytes from %s:%u.", retVal,
		     inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	else
	  LogMessage(Logger, LogDebug,"Received %d bytes.", retVal);
      }
      CBufferCommitBlock(buffer, retVal);
      total += retVal;
    }
  }
  
  if (retVal == 0) {
    if (LogIsInfo(Logger))
      LogMessage(Logger, LogInfo, "socketFD connection closed.");
    return -1;
  }
  else if (err != EAGAIN) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "recv failed from socketFD.  %s.", strerror(err));
    return -2;
  }
  
  if (LogIsDebug(Logger)) {
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);
    if (getpeername(socketFD, (struct sockaddr *) &addr, &len) != -1)
      LogMessage(Logger, LogDebug, "Received a total of %d bytes from %s:%u.", total,
		 inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
    else
      LogMessage(Logger, LogDebug, "Received a total of %d bytes.", total);
  }
  return total;
}

static void FinishJob() {
  assert(QueueSize(HvrQ));

  int k = 0, result = 0, span = 0;
  struct timeval now;
  HvrForward forward = NULL, previous = NULL;
  HvrJob job = QueuePeek(HvrQ);
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Job finished.");
  if (!JobReply(job, HvrOK)) {
    JobCleanUp(1);
    return;
  }
	      
  for (k = 0; job->Forwards[k].Address; k++);
  k = htonl(k);
  if (!ClientOutput(job->Client, &k, 4)) {
    JobCleanUp(1);
    return;
  }
  k = ntohl(k);

  if ((job->Submitted.tv_sec || job->Submitted.tv_usec) && (job->Finished.tv_sec || job->Finished.tv_usec))
    span = (((job->Finished.tv_sec - job->Submitted.tv_sec) * 1000000) + 
      (job->Finished.tv_usec - job->Submitted.tv_usec)) / 1000;
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "render span = %d.", span);
  span = htonl(span);
  if (!ClientOutput(job->Client, &span, 4)) {
    JobCleanUp(1);
    return;
  }
  span = 0;
  if (job->Created.tv_sec || job->Created.tv_usec) {
    if (gettimeofday(&now, NULL)) {
      if (LogIsError(Logger)) {
	int err = errno;
	LogMessage(Logger, LogError, "gettimeofday failed on now.  %s.", strerror(err));
      }
    }
    else {
      span = (((now.tv_sec - job->Created.tv_sec) * 1000000) +
	(now.tv_usec - job->Created.tv_usec)) / 1000;
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "server span = %d.", span);
    }
  }
  span = htonl(span);
  if (!ClientOutput(job->Client, &span, 4)) {
    JobCleanUp(1);
    return;
  }

  for (k--; k > -1; k--) {
    for (previous = NULL, forward = job->Client->Forwards; forward != NULL &&
	   !(forward->RemoteEP->Address == job->Forwards[k].Address &&
	   forward->RemoteEP->Port == job->Forwards[k].Port); 
	 previous = forward, forward = forward->Next);
    if (!ClientOutput(job->Client, &job->Forwards[k], 6)) {
      JobCleanUp(1);
      return;
    }
    if (forward == NULL) {
      if (!ClientReply(job->Client, HvrNFnd)) {
	JobCleanUp(1);
	return;
      }
    }
    else {
      HvrResponse response = HvrOK;
      if (!forward->Connected)
	response = HvrNConn;
      else if (forward->Connected < 0) {
	if (forward->Connected == -1)
	  response = HvrConnClsd;
	else if (forward->Connected == -2)
	  response = HvrConnErr;
	else if (forward->Connected == -3)
	  response = HvrMemErr;
	if (previous == NULL)
	  job->Client->Forwards = forward->Next;
	else
	  previous->Next = forward->Next;
	DisposeForward(forward);
      }
      else {
	CBufferRead(forward->InBuffer, &result, 4);
	result = ntohl(result);
	if (result) {
	  if (result < 128)
	    response = HvrFwdOK + result;
	  else
	    response = result;
	}
      }
      if (!ClientReply(job->Client, response)) {
	JobCleanUp(1);
	return;
      }
    }
  }

  if (job->Transmit) {
    /*if (LogIsDebug(Logger)) {
      char* buffer = malloc(800);
      char* strbuffer = malloc(3200 + (201 * 8) + 14); 
      int m, n;
      CBufferPeek(HvrBuffer, buffer, 800);
      sprintf(strbuffer, "HvrBuffer =");
      for (m = 0, n = 11; m < 800; m++) {
	if (m == 0 || ((m % 16) == 0)) {
	  sprintf(&strbuffer[n], "\n%07d", m);
	  n+=8;
	}
	sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	n+=4;
      }
      free(buffer);
      sprintf(&strbuffer[n], ".");
      strbuffer[++n] = '\0';
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "%s", strbuffer);
      free(strbuffer);
      }*/
    CBuffer saveBuffer = HvrBuffer;
    if (job->Forwards[0].Address) {
      CBufferRemove(HvrBuffer, 20);
      /* if (LogIsDebug(Logger)) {
	char* buffer = malloc(800);
	char* strbuffer = malloc(3200 + (201 * 8) + 14); 
	int m, n;
	CBufferPeek(HvrBuffer, buffer, 800);
	sprintf(strbuffer, "HvrBuffer =");
	for (m = 0, n = 11; m < 800; m++) {
	  if (m == 0 || ((m % 16) == 0)) {
	    sprintf(&strbuffer[n], "\n%07d", m);
	    n+=8;
	  }
	  sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	  n+=4;
	}
	free(buffer);
	sprintf(&strbuffer[n], ".");
	strbuffer[++n] = '\0';
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "%s", strbuffer);
	free(strbuffer);
	}*/
    }
    if (CBufferSize(job->Client->Client->OutBuffer)) {
      if (CBufferInsertBuffer(HvrBuffer, job->Client->Client->OutBuffer) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "insert buffer failed on Hvrbuffer.");
	ClientReply(job->Client, HvrMemErr);
	JobCleanUp(1);
	return;
      }
      /*if (LogIsDebug(Logger)) {
	char* buffer;
	char* strbuffer;
	int m, n, size = 800;
	if (CBufferSize(job->Client->Client->OutBuffer) > 800)
	  size = CBufferSize(job->Client->Client->OutBuffer);
	buffer = malloc(size);
	strbuffer = malloc((size * 4) + (((size / 16) + 1) * 8) + 14);
	if (size != CBufferSize(job->Client->Client->OutBuffer))
	  size = CBufferSize(job->Client->Client->OutBuffer);
	CBufferPeek(job->Client->Client->OutBuffer, buffer, size);
	sprintf(strbuffer, "OutBuffer =");
	for (m = 0, n = 11; m < size; m++) {
	  if (m == 0 || ((m % 16) == 0)) {
	    sprintf(&strbuffer[n], "\n%07d", m);
	    n+=8;
	  }
	  sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	  n+=4;
	}
	sprintf(&strbuffer[n], ".");
	strbuffer[++n] = '\0';
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "%s", strbuffer);
	size = 800;
        CBufferPeek(HvrBuffer, buffer, size);
	sprintf(strbuffer, "HvrBuffer =");
	for (m = 0, n = 11; m < size; m++) {
	  if (m == 0 || ((m % 16) == 0)) {
	    sprintf(&strbuffer[n], "\n%07d", m);
	    n+=8;
	  }
	  sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	  n+=4;
	}
	free(buffer);
	sprintf(&strbuffer[n], ".");
	strbuffer[++n] = '\0';
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "%s", strbuffer);
	free(strbuffer);
	}*/
    }
    HvrBuffer = job->Client->Client->OutBuffer;
    job->Client->Client->OutBuffer = saveBuffer;
  }

  JobCleanUp(1);
}

static void ConnectForward(HvrClient client, HvrForward current, HvrForward previous) {
  assert(client != NULL);
  assert(current != NULL);
  assert(current->Offset > -1);

  int length = strlen(ImageViewerLibrary) + 1, retVal;
  HvrResponse response;

  if (current->Offset < length) {
    retVal = SocketSend(current->SocketFD, (void*)&ImageViewerLibrary[current->Offset], 
			length - current->Offset);
    if (retVal < 0) {
      HvrResponse response = HvrError;
      switch(retVal) {
      case -1: response = HvrConnClsd; break;
      case -2: response = HvrConnErr; break;
      case -3: response = HvrMemErr; break;
      }
      if (previous == NULL)
	client->Forwards = current->Next;
      else
	previous->Next = current->Next;
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
      DisposeForward(current);
      ClientReply(client, response);
      return;
    }
    else {
      current->Offset += retVal;
      if (current->Offset == retVal)
	AddSelectIO(HvrHandle, current->SocketFD, SelectRead);
      if (current->Offset == length)
	RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
    }
  }
  
  if (current->Offset >= length && current->Offset < length + 4) {
    retVal = SocketReceiveBuffer(current->SocketFD, current->InBuffer);
    if (retVal < 0) {
      HvrResponse response = HvrError;
      switch(retVal) {
      case -1: response = HvrConnClsd; break;
      case -2: response = HvrConnErr; break;
      case -3: response = HvrMemErr; break;
      }
      if (previous == NULL)
	client->Forwards = current->Next;
      else
	previous->Next = current->Next;
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
      DisposeForward(current);
      ClientReply(client, response);
      return;
    }
    else {
      current->Offset += retVal;
      if (current->Offset == length + 4) {
	CBufferRead(current->InBuffer, &response, 4);
	if (response != HvrOK) {
	  response += (HvrFwdEMemErr - 1);
	  RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
	  DisposeForward(current);
	  ClientReply(client, response);
	  return;
	}
      }
    }
  }
  
  if (current->Offset >= length + 4 && current->Offset < length + 8) {
    int offset = current->Offset - length - 4;
    retVal = SocketSend(current->SocketFD, &((char*)&current->Key)[offset], 4 - offset);
    if (retVal < 0) {
      switch(retVal) {
      case -1: response = HvrConnClsd; break;
      case -2: response = HvrConnErr; break;
      case -3: response = HvrMemErr; break;
      }
      if (previous == NULL)
	client->Forwards = current->Next;
      else
	previous->Next = current->Next;
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
      DisposeForward(current);
      ClientReply(client, response);
    }
    else {
      current->Offset += retVal;
      if (current->Offset < length + 8) {
	if (!offset)
	  AddSelectIO(HvrHandle, current->SocketFD, SelectWrite);
      }
      else if (offset)
	RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
    }
  }
  
  if (current->Offset >= length + 8) {
    retVal = SocketReceiveBuffer(current->SocketFD, current->InBuffer);
    if (retVal < 0) {
      switch(retVal) {
      case -1: response = HvrConnClsd; break;
      case -2: response = HvrConnErr; break;
      case -3: response = HvrMemErr; break;
      }
      if (previous == NULL)
	client->Forwards = current->Next;
      else
	previous->Next = current->Next;
      RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
      DisposeForward(current);
      ClientReply(client, response);
    }
    else {
      current->Offset += retVal;
      if (current->Offset == length + 12) {
	CBufferRead(current->InBuffer, &response, 4);
	if (response != HvrOK) {
	  response += (HvrFwdEMemErr - 1);
	  RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
	  DisposeForward(current);
	}
	else {
	  current->Offset = 0;
	  current->Connected = 1;
	}
	ClientReply(client, response);
      }
    }
  }
}


struct timeval* IOHandler(int fd, LcseIOEvent event) {
  assert(fd > -1);

  HvrForward current = NULL, previous = NULL;
  HvrAsyncIO currentIO = NULL, previousIO = NULL;
  HvrClient client;
  for (client = HvrClients; client != NULL && current == NULL && currentIO == NULL;) {
    for (previous = NULL, current = client->Forwards; current != NULL && current->SocketFD != fd; 
	 previous = current, current = current->Next);
    if (current == NULL) {
      for (previousIO = NULL, currentIO = client->AsyncIO; currentIO != NULL && currentIO->Fds[0] != fd;
	   previousIO = currentIO, currentIO = currentIO->Next);
    }
    client = (current == NULL && currentIO == NULL) ? client->Next : client;
  } 

  if (current == NULL && currentIO == NULL) {
    if (LogIsWarn(Logger))
      LogMessage(Logger, LogWarn, "current and currentIO == NULL");
    return NULL;
  }

  switch(event) {
  case IOWriteReady:
    if (current) {
      if (current->Connected == 0) {
	if (current->Offset < 0) {
	  int result = 0;
	  socklen_t resultLen = sizeof(result);
	  if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &result, &resultLen) < 0 || (result)) {
	    int err = errno;
	    if (LogIsError(Logger)) {
	      if (result)
		LogMessage(Logger, LogError, "connect failed on forwarding socket.  %s.", 
			   strerror(result));
	      else
		LogMessage(Logger, LogError, "getsockopt failed on forwarding socket.  %s.", 
			   strerror(err));
	    }
	    if (previous == NULL)
	      client->Forwards = current->Next;
	    else
	      previous->Next = current->Next;
	    DisposeForward(current);
	    ClientReply(client, (result) ? HvrConnErr : HvrIOErr);
	    return NULL;
	  }
	  else
	    current->Offset++;
	}
	ConnectForward(client, current, previous);
      }
      else if (QueueSize(HvrQ)) {
	HvrJob job = QueuePeek(HvrQ);
	if (job->Offset != job->Length)
	  RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
	else {
	  int retVal = SocketSendBuffer(current->SocketFD, HvrBuffer, current->Offset);
	  if (retVal < 0)
	    current->Connected = retVal;
	  else
	    current->Offset += retVal;
	  if (current->Offset == CBufferSize(HvrBuffer) || retVal < 0)
	    RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
	}
      }
      else
	RemoveSelectIO(HvrHandle, current->SocketFD, SelectWrite);
    } else {
      switch (currentIO->Command) {
      default:
	RemoveSelectIO(HvrHandle, currentIO->Fds[0], SelectWrite);
	break;
      }
    }
    break;
  case IOReadReady:
    if (current) {
      if (current->Connected == 0)
	ConnectForward(client, current, previous);
      else {
	int retVal = SocketReceiveBuffer(current->SocketFD, current->InBuffer);
	if (retVal < 0) {
	  current->Connected = retVal;
	  RemoveSelectIO(HvrHandle, current->SocketFD, SelectRead);
	}
	if (!QueueSize(HvrQ))
	  CBufferClear(current->InBuffer);
	else {
	  HvrJob job = QueuePeek(HvrQ);
	  if (job->Offset != job->Length)
	    CBufferClear(current->InBuffer);
	  else if (CBufferSize(current->InBuffer) > 3 || current->Connected < 0) {
	    HvrForward forward = NULL;
	    int complete = 1, k = 0;
	    for (k = 0; job->Forwards[k].Address && complete; k++) {
	      if (job->Forwards[k].Address != current->RemoteEP->Address ||
		  job->Forwards[k].Port != current->RemoteEP->Port) {
		for (forward = job->Client->Forwards; forward != NULL &&
		       !(forward->RemoteEP->Address == job->Forwards[k].Address &&
			 forward->RemoteEP->Port == job->Forwards[k].Port); forward = forward->Next);
		if (forward != NULL && forward->Connected > 0) {
		  if (forward->Offset < CBufferSize(HvrBuffer) || CBufferSize(forward->InBuffer) < 4)
		    complete = 0;
		}
	      }
	    }
	    
	    if (complete) {
	      /*if (LogIsDebug(Logger))
		LogMessage(Logger, LogDebug, "Finishing job from IOReadReady.");*/
	      FinishJob();
	    }
	  }
	}
      }
    }
    else {
      switch (currentIO->Command) {
      case HvrGetLog: {
	int rc;
	CBuffer cbuffer;
	void* save;

	RemoveSelectIO(HvrHandle, currentIO->Fds[0], SelectRead);
	memcpy(&rc, currentIO->Data, sizeof(int));

	if ((rc = pthread_join(rc, (void*)&cbuffer))) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "pthread_join failed on currentIO->Data.  %s.", strerror(rc));
	  ClientReply(client, HvrSysErr);
	  for (errno = 0; close(currentIO->Fds[1]) == -1 && errno == EINTR; errno = 0);
	  for (errno = 0; close(currentIO->Fds[0]) == -1 && errno == EINTR; errno = 0);
	  free(currentIO->Data);
	  save = currentIO->Next;
	  free(currentIO);
	  if (previousIO)
	    previousIO->Next = save;
	  else
	    client->AsyncIO = save;
	  return NULL;
	}

	if (cbuffer) {
	  if (CBufferSize(client->Client->OutBuffer)) {
	    if (CBufferInsertBuffer(cbuffer, client->Client->OutBuffer) == -1) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "CBufferInsertBuffer failed for cbuffer.");
	      ClientReply(client, HvrMemErr);
	      CBufferDispose(cbuffer);
	      cbuffer = NULL;
	    }
	  }
	  if (cbuffer) {
	    CBufferDispose(client->Client->OutBuffer);
	    client->Client->OutBuffer = cbuffer;
	  }
	}
	else
	  ClientReply(client, HvrMemErr);
	
	for (errno = 0; close(currentIO->Fds[1]) == -1 && errno == EINTR; errno = 0);
	for (errno = 0; close(currentIO->Fds[0]) == -1 && errno == EINTR; errno = 0);
	free(currentIO->Data);
	save = currentIO->Next;
	free(currentIO);
	if (previousIO)
	  previousIO->Next = save;
	else
	  client->AsyncIO = save;
      }
	break;
      default:
	RemoveSelectIO(HvrHandle, currentIO->Fds[0], SelectRead);
	break;
      }
    }
    break;
  }       
  
  return NULL;
}

static void ProcessOutput() {
  assert(QueueSize(HvrQ));

  HvrJob job = QueuePeek(HvrQ);
  int complete = 1;
  if (job->Forwards[0].Address) {
    HvrForward forward = NULL;
    int k = 0, retVal = 0;
    
    /*k = htonl(job->SequenceNumber);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
      }*/
    k = htonl(job->ImageY);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
    }
    k = htonl(job->ImageX);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
    }
    k = htonl(job->ImageHeight);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
    }
    k = htonl(job->ImageWidth);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
    }
    k = htonl(0);
    if (CBufferInsert(HvrBuffer, &k, 4) < 0) {
      JobReply(job, HvrMemErr);
      JobCleanUp(1);
      return;
    }

    /*if (LogIsDebug(Logger)) {
      char* buffer = malloc(CBufferSize(HvrBuffer));
      char* strbuffer = malloc((CBufferSize(HvrBuffer) * 4) + 
                              (((CBufferSize(HvrBuffer) / 16) + 1) * 8) + 14); 
      int m, n;
      CBufferPeek(HvrBuffer, buffer, CBufferSize(HvrBuffer));
      sprintf(strbuffer, "HvrBuffer =");
      for (m = 0, n = 11; m < CBufferSize(HvrBuffer); m++) {
	if (m == 0 || ((m % 16) == 0)) {
	  sprintf(&strbuffer[n], "\n%07d", m);
	  n+=8;
	}
	sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	n+=4;
      }
      free(buffer);
      sprintf(&strbuffer[n], ".");
      strbuffer[++n] = '\0';
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "%s", strbuffer);
      free(strbuffer);
    }*/
    
    for (k = 0; job->Forwards[k].Address; k++) {
      for (forward = job->Client->Forwards; forward != NULL && 
	     !(forward->RemoteEP->Address == job->Forwards[k].Address &&
	       forward->RemoteEP->Port == job->Forwards[k].Port); forward = forward->Next);
      if (forward != NULL && forward->Connected > 0) {
	retVal = SocketSendBuffer(forward->SocketFD, HvrBuffer, 0);
	if (retVal < 0)
	  forward->Connected = retVal;
	else {
	  forward->Offset = retVal;
	  if (retVal < CBufferSize(HvrBuffer)) {
	    complete = 0;
	    AddSelectIO(HvrHandle, forward->SocketFD, SelectWrite);
	  }
	}
      }
    }
    
    if (complete) {
      for (k = 0; job->Forwards[k].Address; k++) {
	for (forward = job->Client->Forwards; forward != NULL &&
	       !(forward->RemoteEP->Address == job->Forwards[k].Address &&
		 forward->RemoteEP->Port == job->Forwards[k].Port); forward = forward->Next);
	if (forward != NULL && forward->Connected > 0) {
	  retVal = SocketReceiveBuffer(forward->SocketFD, forward->InBuffer);
	  if (retVal < 4) {
	    if (retVal < 0)
	      forward->Connected = retVal;
	    else
	      complete = 0;
	  }
	}
      }
    }
  }
  
  if (complete) {
    /*if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Finishing job from ProcessOutput.");*/
    FinishJob();
  }
}

static void ProcessData(LcseExecutable executable) {
  assert(executable != NULL);
  
  char* buffer = NULL;
  int length = 0, retVal = 1;
  
  if (HvrExecutable == NULL) {
    int err = EINTR;
    length = 1;
    buffer = CBufferGetWriteBlock(HvrBuffer, &length);
    while (read(executable->StdOut, buffer, length) < 1 && err == EINTR) {
      err = errno;
      if (err != EINTR) {
	if (err != EAGAIN) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "read failed on executable->StdOut.  %s.", strerror(err));
	  ReadErrors++;
	  kill(HvrID, SIGTERM);
	}
      }
    }
    if (err == EINTR) {
      HvrExecutable = executable;
      HvrID = 0;
      ReadErrors = 0;
      if (QueueSize(HvrQ))
	StartJob();
    }
    return;
  }
  
  if (QueueSize(HvrQ)) {
    HvrJob job = QueuePeek(HvrQ);
    if (job->Offset == job->Length) {
      int count = 0, total = 0;
      if (!CBufferSize(HvrBuffer)) {
	if (gettimeofday(&job->Finished, NULL) && LogIsError(Logger)) {
	  int err = errno;
	  LogMessage(Logger, LogError, "gettimeofday failed on job->Finished.  %s.", strerror(err));
	}
      }
      retVal = 1;
      length = (job->ImageWidth * job->ImageHeight * 3) + ((job->Status == INT_MIN) ? 4 : 0);
      while (CBufferSize(HvrBuffer) != length && HvrTotal != length && retVal > 0) {	
	if (HvrTotal > -1) {
	  if (LogIsDebug(Logger))
	    LogMessage(Logger, LogDebug, "reading from HvrExecutable->StdOut to HvrBuffer.");
	  count = length - HvrTotal;
	  buffer = CBufferGetWriteBlock(HvrBuffer, &count);
	  total += retVal = read(HvrExecutable->StdOut, buffer, count);
	  if (retVal < 0) {
	    int err = errno;
	    if (err == EINTR)
	      retVal = 1;
	    else if (err != EAGAIN) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "read failed on HvrExecutable->StdOut.");
	      ReadErrors++;
	      kill(HvrExecutable->ProcessID, SIGTERM);
	      JobReply(job, HvrReadErr);
	      JobCleanUp(0);
	      HvrTotal = -1;
	      return;
	    }
	    else if (retVal > 0) {
	      if (LogIsDebug(Logger)) {
		LogMessage(Logger, LogDebug, "read %d bytes from HvrExecutable->StdOut.", retVal);
		LogMessage(Logger, LogDebug, "Discarding bytes from HvrExecutable->StdOut.");
	      }
	      CBufferClear(HvrBuffer);
	    }
	  }
	}
	else {
	  count = length - CBufferSize(HvrBuffer);
	  buffer = CBufferGetWriteBlock(HvrBuffer, &count);
	  if (buffer == NULL) {
	    HvrTotal = CBufferSize(HvrBuffer);
	    CBufferClear(HvrBuffer);
	    if (LogIsError(Logger))
	      LogMessage(Logger, LogError, "getwriteblock failed on HvrBuffer.");
	  }
	  else {
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "reading from HvrExecutable->StdOut to HvrBuffer.");
	    total += retVal = read(HvrExecutable->StdOut, buffer, count);
	    if (retVal < 0) {
	      int err = errno;
	      if (err == EINTR)
		retVal = 1;
	      else if (err != EAGAIN) {
		if (LogIsError(Logger))
		  LogMessage(Logger, LogError, "read failed on HvrExecutable->StdOut.");
		ReadErrors++;
		kill(HvrExecutable->ProcessID, SIGTERM);
		JobReply(job, HvrReadErr);
		JobCleanUp(0);
		return;
	      }
	    }
	    else if (retVal > 0) {
	      if (LogIsDebug(Logger))
		LogMessage(Logger, LogDebug, "read %d bytes from HvrExecutable->StdOut.", retVal);
	      CBufferCommitBlock(HvrBuffer, retVal);
	      /*if (LogIsDebug(Logger)) {
	      	char* buffer = malloc(800);
	      	char* strbuffer = malloc(3200 + (201 * 8) + 14);
	      	int m, n;
	      	CBufferPeek(HvrBuffer, buffer, 800);
	      	sprintf(strbuffer, "HvrBuffer =");
	      	for (m = 0, n = 11; m < 800; m++) {
	      	  if (m == 0 || ((m % 16) == 0)) {
	      	    sprintf(&strbuffer[n], "\n%07d", m);
	      	    n+=8;
	      	  }
	      	  sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
	      	  n+=4;
	      	}
	      	free(buffer);
		sprintf(&strbuffer[n], ".");
		strbuffer[++n] = '\0';
		if (LogIsDebug(Logger))
		  LogMessage(Logger, LogDebug, "%s", strbuffer);
		free(strbuffer);
		}*/
	      if (job->Status == INT_MIN && CBufferSize(HvrBuffer) > 3) {
		CBufferRead(HvrBuffer, &job->Status, 4);
		length -= 4;
		/*if (LogIsDebug(Logger)) {
		  char* buffer = malloc(800);
		  char* strbuffer = malloc(3200 + (201 * 8) + 14); 
		  int m, n;
		  CBufferPeek(HvrBuffer, buffer, 800);
		  sprintf(strbuffer, "HvrBuffer =");
		  for (m = 0, n = 11; m < 800; m++) {
		    if (m == 0 || ((m % 16) == 0)) {
		      sprintf(&strbuffer[n], "\n%07d", m);
		      n+=8;
		    }
		    sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
		    n+=4;
		  }
		  free(buffer);
		  sprintf(&strbuffer[n], ".");
		  strbuffer[++n] = '\0';
		  if (LogIsDebug(Logger))
		    LogMessage(Logger, LogDebug, "%s", strbuffer);
		  free(strbuffer);
		  }*/
		if (job->Status) {
		  if (LogIsError(Logger))
		    LogMessage(Logger, LogError, "hvr_cl failed with status code %d.", job->Status);
		  length = htonl(job->Status);
		  JobOutput(job, &length, 4);
		  JobCleanUp(0);
		  return;
		}
	      }
	    }
	  }
	}
      }
	
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "read a total of %d bytes from HvrExecutable->StdOut.", total);
      
      if (CBufferSize(HvrBuffer) == length || HvrTotal == length) {
	if (HvrTotal != length) {
	  if (job->Client != NULL)
	    ProcessOutput();
	  else
	    JobCleanUp(1);
	}
	else {
	  JobReply(job, HvrMemErr);
	  HvrTotal = -1;
	  JobCleanUp(1);
	}
      }
      return;
    }
  }

  for (CBufferClear(HvrBuffer), length = 0, buffer = CBufferGetWriteBlock(HvrBuffer, &length); 
       retVal > 0; 
       CBufferClear(HvrBuffer), length = 0, buffer = CBufferGetWriteBlock(HvrBuffer, &length))
    retVal = read(HvrExecutable->StdOut, buffer, length);
}
  
struct timeval* ExecutableHandler(const LcseExecutable executable, LcseExecutableEvent event) {
  assert(executable != NULL);
  
  switch(event) {
  case ExecutableDataReady:
    ProcessData(executable);
    break;
  case ExecutableWriteReady:
    StartJob();
    break;
  case ExecutableExited:
    if (QueueSize(HvrQ)) {
      HvrJob job = QueuePeek(HvrQ);
      if (job->Offset > 0) {
	if (!executable->Status)
	  JobReply(job, HvrError);
	else {
	  int k = htonl(executable->Status);
	  JobOutput(job, &k, 4);
	}
	QueueDequeue(HvrQ);
	RemoveJob(job);
      }
    }
    CBufferClear(HvrBuffer);
    if (executable->Status == 127)
      HvrID = -5;
    else if (ReadErrors > 4)
      HvrID = -6;
    else if (WriteErrors > 4)
      HvrID = -7;
    else
      HvrID = SpawnExecutable(HvrHandle, executable->Argv);
    if (HvrID < 0 && QueueSize(HvrQ)) {
      HvrJob job = NULL;
      int response = htonl(HvrError);
      switch (HvrID) {
      case -1: response = htonl(HvrForkErr); break;
      case -2: response = htonl(HvrMemErr); break;
      case -3: response = htonl(HvrPipeErr); break;
      case -4: response = htonl(HvrIOErr); break;
      case -5: response = htonl(HvrExecErr); break;
      case -6: response = htonl(HvrReadErr); break;
      case -7: response = htonl(HvrWriteErr); break;
      }
      for (job = QueueDequeue(HvrQ); QueueSize(HvrQ); job = QueueDequeue(HvrQ)) {
	if (JobReply(job, response)) {
	  job->Client->Jobs = NULL;
	  CloseClient(job->Client->Client);
	}
	DisposeJob(job);
      }
    }
    HvrExecutable = NULL;
    break;
  }
  return NULL;
}

static int AddJob(HvrClient client) {
  assert(client != NULL);
  
  if (CBufferSize(client->Client->InBuffer) < 57)
    return 0;
  else {
    int length = 0, forwards = 0, k, total;
    CBufferPeekAt(client->Client->InBuffer, &length, 4, 4);
    length = ntohl(length);
    CBufferPeekAt(client->Client->InBuffer, &forwards, 8, 4);
    forwards = ntohl(forwards);
    total = length + 57 + (forwards * 6);
    if (CBufferSize(client->Client->InBuffer) < total)
      return 0;
    
    HvrJob job = malloc(sizeof(struct HvrJobRecord));
    if (job == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job.");
      CBufferRemove(client->Client->InBuffer, total);
      ClientReply(client, HvrMemErr);
      return 1;
    }
    
    job->KeyFrame = malloc(sizeof(char) * (length + 8));
    if (job->KeyFrame == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job->KeyFrame.");
      free(job);
      CBufferRemove(client->Client->InBuffer, total);
      ClientReply(client, HvrMemErr);
      return 1;
    }
    
    job->Forwards = malloc(sizeof(struct HvrEndPointRecord) * (forwards + 1));
    if (job->Forwards == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job->Forwards.");
      free(job->KeyFrame);
      free(job);
      CBufferRemove(client->Client->InBuffer, total);
      ClientReply(client, HvrMemErr);
      return 1;
    }
    
    job->Offset = 0;
    job->Length = length + 7;
    job->Status = INT_MIN;
    CBufferRemove(client->Client->InBuffer, 12);
    CBufferRead(client->Client->InBuffer, &job->ImageX, 4);
    job->ImageX = ntohl(job->ImageX);
    CBufferRead(client->Client->InBuffer, &job->ImageY, 4);
    job->ImageY = ntohl(job->ImageY);
    CBufferRead(client->Client->InBuffer, &job->ImageWidth, 4);
    job->ImageWidth = ntohl(job->ImageWidth);
    CBufferRead(client->Client->InBuffer, &job->ImageHeight, 4);
    job->ImageHeight = ntohl(job->ImageHeight);
    CBufferRead(client->Client->InBuffer, &job->PanelWidth, 4);
    job->PanelWidth = ntohl(job->PanelWidth);
    CBufferRead(client->Client->InBuffer, &job->PanelHeight, 4);
    job->PanelHeight = ntohl(job->PanelHeight);
    CBufferRead(client->Client->InBuffer, &job->SectionX, 4);
    job->SectionX = ntohl(job->SectionX);
    CBufferRead(client->Client->InBuffer, &job->SectionY, 4);
    job->SectionY = ntohl(job->SectionY);
    CBufferRead(client->Client->InBuffer, &job->SectionMaxX, 4);
    job->SectionMaxX = ntohl(job->SectionMaxX);
    CBufferRead(client->Client->InBuffer, &job->SectionMaxY, 4);
    job->SectionMaxY = ntohl(job->SectionMaxY);
    CBufferRead(client->Client->InBuffer, &job->SequenceNumber, 4);
    job->SequenceNumber = ntohl(job->SequenceNumber);
    CBufferRead(client->Client->InBuffer, &job->Transmit, 1);
    for (k = 0; k < forwards; k++) {
      CBufferRead(client->Client->InBuffer, &job->Forwards[k].Address, 4);
      CBufferRead(client->Client->InBuffer, &job->Forwards[k].Port, 2);
    }
    job->Forwards[k].Address = job->Forwards[k].Port = 0;
    CBufferRead(client->Client->InBuffer, job->KeyFrame, length);
    sprintf(&job->KeyFrame[length], "%s", "render\n");
    job->Client = client;
    job->Created.tv_sec = job->Created.tv_usec = job->Submitted.tv_sec = job->Submitted.tv_usec = 
      job->Finished.tv_sec = job->Finished.tv_usec = 0;
    if (gettimeofday(&job->Created, NULL) && LogIsError(Logger)) {
      int err = errno;
      LogMessage(Logger, LogError, "gettimeofday failed on job->Created.  %s.", strerror(err));
    }
    job->Next = client->Jobs;
    client->Jobs = job;
    
    if (LogIsDebug(Logger)) {
      char keyFrame[length + 8];
      memcpy(keyFrame, job->KeyFrame, length + 7);
      keyFrame[length + 7] = '\0';
      LogMessage(Logger, LogDebug, "Job received: Length = %d, ImageX = %d, ImageY = %d, \
ImageWidth = %d, ImageHeight = %d, PanelWidth = %d, PanelHeight = %d, SectionX = %d, \
SectionY = %d, SectionMaxX = %d, SectionMaxY = %d, SequenceNumber = %d, Transmit = %d, \
KeyFrame :\n%s", job->Length, job->ImageX, job->ImageY, job->ImageWidth, job->ImageHeight, 
		 job->PanelWidth, job->PanelHeight, job->SectionX, job->SectionY, job->SectionMaxX, 
		 job->SectionMaxY, job->SequenceNumber, (int)job->Transmit, keyFrame);
    }

    if (QueueEnqueue(HvrQ, job) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "enqueue failed on hvr q.");
      free(job->KeyFrame);
      free(job->Forwards);
      free(job);
      ClientReply(client, HvrMemErr);
      return 1;
    }

    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "QueueSize(HvrQ) = %d, HvrExecutable != NULL = %s", QueueSize(HvrQ), 
		 (HvrExecutable != NULL) ? "true" : "false");
    
    if (QueueSize(HvrQ) == 1 && HvrExecutable != NULL)
      StartJob();
  }
  
  return 1;
}

static int AddForward(HvrClient client) {
  assert(client != NULL);
  
  if (CBufferSize(client->Client->InBuffer) < 16)
    return 0;
  else {
    int socketFD = -1, retVal = 0, on = 1, size;
    struct sockaddr_in destination;
    HvrForward forward;
    
    CBufferRemove(client->Client->InBuffer, 4);
    CBufferRead(client->Client->InBuffer, &destination.sin_addr.s_addr, 4);
    CBufferRead(client->Client->InBuffer, &destination.sin_port, 2);
    CBufferRead(client->Client->InBuffer, &destination.sin_family, 2);
    destination.sin_family = ntohs(destination.sin_family);
    memset(&(destination.sin_zero), 0, 8);

    for (retVal = 1, forward = client->Forwards; forward != NULL && retVal; forward = forward->Next) {
      if (forward->RemoteEP->Address == destination.sin_addr.s_addr && 
	  forward->RemoteEP->Port == destination.sin_port)
	retVal = 0;
    }

    if (!retVal) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "Forward end point %s:%u already established.",
		   inet_ntoa(destination.sin_addr), (int)ntohs(destination.sin_port));
      CBufferRemove(client->Client->InBuffer, 4);
      ClientReply(client, HvrDup);
      return 1;
    }
    
    socketFD = socket(destination.sin_family, SOCK_STREAM, 0);
    if (socketFD < 0) {
      int err = errno;
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "Could not create forwarding socket.  %s.", strerror(err));
      CBufferRemove(client->Client->InBuffer, 4);
      ClientReply(client, HvrSockErr);
      return 1;
    }
    
    retVal = SetNonBlockingIO(socketFD);
    if (retVal < 0) {
      int err = errno;
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "SetNonBlockingIO failed on forwarding socket.  %s.", 
		   strerror(err));
      Cclose(socketFD);
      CBufferRemove(client->Client->InBuffer, 4);
      ClientReply(client, HvrIOErr);
      return 1;
    }
    
    retVal = SetCloseOnExecIO(socketFD);
    if (retVal < 0) {
      int err = errno;
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "SetCloseOnExecIO failed on forwarding socket.  %s.", 
		   strerror(err));
      Cclose(socketFD);
      CBufferRemove(client->Client->InBuffer, 4);
      ClientReply(client, HvrIOErr);
      return 1;
    }

    if (destination.sin_family != 26) {
      retVal = setsockopt(socketFD, SOL_TCP, TCP_NODELAY, &on, sizeof(int));
      if (retVal < 0) {
	int err = errno;
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "setsockopt failed on forwarding socket.  %s.",
		     strerror(err));
	Cclose(socketFD);
	CBufferRemove(client->Client->InBuffer, 4);
	ClientReply(client, HvrIOErr);
	return 1;
      }
      
      size = BUFFER_SIZE;
      retVal = setsockopt(socketFD, SOL_SOCKET, SO_SNDBUF, &size, sizeof(int));
      if (retVal < 0) {
	int err = errno;
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "setsockopt failed on forwarding socket.  %s.",
		     strerror(err));
	Cclose(socketFD);
	CBufferRemove(client->Client->InBuffer, 4);
	ClientReply(client, HvrIOErr);
	return 1;
      }
    }
    
    if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "Attempting to connect to %s:%u.", 
		   inet_ntoa(destination.sin_addr), (int)ntohs(destination.sin_port));
    errno = 0;
    retVal = connect(socketFD, (struct sockaddr *)&destination, sizeof(destination));
    if (retVal >= 0 || errno == EINPROGRESS || errno == EINTR) {
      forward = malloc(sizeof(struct HvrForwardRecord));
      if (forward == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "malloc failed on forward.");
	Cclose(socketFD);
	CBufferRemove(client->Client->InBuffer, 4);
	ClientReply(client, HvrMemErr);
	return 1;
      }

      forward->RemoteEP = malloc(sizeof(struct HvrEndPointRecord));
      if (forward == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "malloc failed on forward->RemoteEP.");
	Cclose(socketFD);
	CBufferRemove(client->Client->InBuffer, 4);
	ClientReply(client, HvrMemErr);
	return 1;
      }

      forward->InBuffer = CBufferCreate(0, 1);
      if (forward->InBuffer == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "create failed on forward->InBuffer.");
	Cclose(socketFD);
	CBufferRemove(client->Client->InBuffer, 4);
	ClientReply(client, HvrMemErr);
	return 1;
      }
      
      forward->SocketFD = socketFD;
      forward->Offset = -1;
      forward->Connected = 0;
      forward->RemoteEP->Address = destination.sin_addr.s_addr;
      forward->RemoteEP->Port = destination.sin_port;
      CBufferRead(client->Client->InBuffer, &forward->Key, 4);
      forward->Next = client->Forwards;
      client->Forwards = forward;
      
      if (retVal >= 0) {
	forward->Offset = 0;
	ConnectForward(client, forward, NULL);
      }
      else
	AddSelectIO(HvrHandle, socketFD, SelectWrite);
    }
    else {
      int err = errno;
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "connect failed on forwarding socket.  %s.", strerror(err));
      Cclose(socketFD);
      CBufferRemove(client->Client->InBuffer, 4);
      ClientReply(client, HvrConnErr);
    }
  }

  return 1;
}

static void AddClient(const LcseClient client) {
  assert(client != NULL);
  
  if (HvrID < 0) {
    int response = htonl(HvrError);
    switch (HvrID) {
    case -1: response = htonl(HvrForkErr); break;
    case -2: response = htonl(HvrMemErr); break;
    case -3: response = htonl(HvrPipeErr); break;
    case -4: response = htonl(HvrIOErr); break;
    case -5: response = htonl(HvrExecErr); break;
    case -6: response = htonl(HvrReadErr); break;
    case -7: response = htonl(HvrWriteErr); break;
    }
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    CloseClient(client);
    return;
  }

  HvrClient hvrClient = malloc(sizeof(struct HvrClientRecord));
  if (hvrClient == NULL) {
    int response = htonl(HvrMemErr);
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    CloseClient(client);
    return;
  }
  
  hvrClient->Client = client;
  hvrClient->Forwards = NULL;
  hvrClient->AsyncIO = NULL;
  hvrClient->Jobs = NULL;
  hvrClient->Next = HvrClients;
  HvrClients = hvrClient;
  ClientReply(hvrClient, HvrOK);
}

static void* GetLog(void* arg) {
  assert(arg != NULL);

  HvrAsyncIO asyncIO = (HvrAsyncIO)arg;
  int fd, cnt, ret, total, total0 = 0;
  char* buffer;
  CBuffer cbuffer;

  LogFlush(Logger);

  if ((cbuffer = CBufferCreate(0, 1)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create cbuffer failed.");
    for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
    if (errno && LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
    return NULL;
  } 

  if ((fd = open(LogGetPath(Logger), O_RDONLY)) == -1) {
    if (errno == ENOENT)
      ret = htonl(HvrNFnd);
    else
      ret = htonl(HvrIOErr);
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "open failed for fd.  %s.", strerror(errno));
    if (CBufferWrite(cbuffer, &ret, 4) == -1) {
      CBufferDispose(cbuffer);
      cbuffer = NULL;
    }
    for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
    if (errno && LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
    return cbuffer;
  }

  for (total = 0, cnt = 0, ret = 1; ret > 0; cnt = 0) {
    buffer = CBufferGetWriteBlock(cbuffer, &cnt);
    ret = read(fd, buffer, cnt);
    if (ret == -1) {
      if (errno != EINTR) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "read failed on fd.  %s", strerror(errno));
	ret = htonl(HvrIOErr);
	CBufferClear(cbuffer);
	if (CBufferWrite(cbuffer, &ret, 4) == -1) {
	  CBufferDispose(cbuffer);
	  cbuffer = NULL;
	}
	for (errno = 0; close(fd) == -1 && errno == EINTR; errno = 0);
	for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
	if (errno && LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
	return cbuffer;
      }
    }
    else {
      total += ret;
      CBufferCommitBlock(cbuffer, ret);
    }
  }
  for (errno = 0; close(fd) == -1 && errno == EINTR; errno = 0);
 
  if ((buffer = malloc(sizeof(char) * (strlen(LogGetPath(Logger)) + 3))) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on buffer.", strerror(errno));
    CBufferDispose(cbuffer);
    for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
    if (errno && LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
    return NULL;
  }
  sprintf(buffer, "%s.0", LogGetPath(Logger));
  fd = open(buffer, O_RDONLY);
  free(buffer);
  if (fd != -1) {
    for (total0 = 0, cnt = 0, ret = 1; ret > 0; cnt = 0) {
      buffer = CBufferGetWriteBlock(cbuffer, &cnt);
      ret = read(fd, buffer, cnt);
      if (ret == -1) {
	if (errno != EINTR) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "read failed on fd.  %s.", strerror(errno));
	  ret = htonl(HvrIOErr);
	  CBufferClear(cbuffer);
	  if (CBufferWrite(cbuffer, &ret, 4) == -1) {
	    CBufferDispose(cbuffer);
	    cbuffer = NULL;
	  }
	  for (errno = 0; close(fd) == -1 && errno == EINTR; errno = 0);
	  for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
	  if (errno && LogIsError(Logger))
	    LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
	  return cbuffer;
	}
      }
      else {
	total0 += ret;
	CBufferCommitBlock(cbuffer, ret);
      }
    }
    for (errno = 0; close(fd) == -1 && errno == EINTR; errno = 0);
  }

  total0 = htonl(total0);
  if (CBufferInsert(cbuffer, &total0, 4) == -1) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "CBufferInsert failed on cbuffer.");
    CBufferDispose(cbuffer);
    for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
    if (errno && LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
    return NULL;
  }
  total0 = ntohl(total0);

  total = htonl(total);
  if (CBufferInsert(cbuffer, &total, 4) == -1) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "CBufferInsert failed on cbuffer.");
    CBufferDispose(cbuffer);
    for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
    if (errno && LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
    return NULL;
  }
  total = ntohl(total);

  ret = htonl(HvrOK);
  if (CBufferInsert(cbuffer, &ret, 4) == -1) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "CBufferInsert failed on cbuffer.");
    CBufferDispose(cbuffer);
    cbuffer = NULL;
  }

  for (errno = 0; write(asyncIO->Fds[1], &fd, 1) == -1 && errno == EINTR; errno = 0);
  if (errno && LogIsError(Logger))
    LogMessage(Logger, LogError, "write failed on asyncIO->FD.  %s.", strerror(errno));
  
  return cbuffer;
}

struct timeval* ClientHandler(const LcseClient client, LcseClientEvent event) {
  assert(client != NULL);
  
  switch(event) {
  case ClientAccepted:
    AddClient(client);
    break;
  case ClientDataReceived: {
    int cmd = -1;
    HvrClient current = NULL;
    for (current = HvrClients; current != NULL && current->Client != client; current = current->Next);
    if (current != NULL) {
      if (CBufferSize(client->InBuffer) > 3) {
	CBufferPeek(client->InBuffer, &cmd, 4);
	cmd = ntohl(cmd);
      }
      while (cmd != -1) {
	int retVal = 0;
	switch(cmd) {
	case HvrAddFwd:
	  retVal = AddForward(current);
	  break;
	case HvrRmvFwd:
	  if (CBufferSize(client->InBuffer) < 10)
	    retVal = 0;
	  else {
	    HvrForward forward = NULL, previous = NULL;
	    struct HvrEndPointRecord ep;
	    CBufferRemove(client->InBuffer, 4);
	    CBufferRead(client->InBuffer, &ep.Address, 4);
	    CBufferRead(client->InBuffer, &ep.Port, 2);
	    for (previous = NULL, forward = current->Forwards; forward != NULL && 
		   !(forward->RemoteEP->Address == ep.Address && forward->RemoteEP->Port == ep.Port); 
		 previous = forward, forward = forward->Next); 
	    if (forward == NULL)
	      ClientReply(current, HvrNFnd);
	    else {
	      if (previous == NULL)
		current->Forwards = forward->Next;
	      else
		previous->Next = forward->Next;
	      DisposeForward(forward);
	      ClientReply(current, HvrOK);
	    }
	    retVal = 1;
	  }
	  break;
	case HvrAddKey:
	  retVal = AddJob(current);
	  break;
	case HvrGetLog: { 
	  int fds[2];
	  
	  retVal = 1;
	  CBufferRemove(client->InBuffer, 4);
	  if (pipe(fds)) {
	    if (LogIsError(Logger))
	      LogMessage(Logger, LogError, "pipe failed for fds.  %s.", strerror(errno));
	    ClientReply(current, HvrPipeErr);
	  }
	  else {
	    int ret;
	    HvrAsyncIO asyncIO;
	    for (asyncIO = current->AsyncIO; asyncIO != NULL && asyncIO->Next != NULL; asyncIO = asyncIO->Next);

	    if (asyncIO) {
	      if ((asyncIO->Next = malloc(sizeof(struct HvrAsyncIORecord))) == NULL) {
		if (LogIsError(Logger))
		  LogMessage(Logger, LogError, "malloc failed for asyncIO->Next.");
		ClientReply(current, HvrMemErr);
		for (errno = 0; close(fds[1]) == -1 && errno == EINTR; errno = 0);
		for (errno = 0; close(fds[0]) == -1 && errno == EINTR; errno = 0);
		break;
	      }
	      asyncIO = asyncIO->Next;
	    }
	    else {
	      if ((current->AsyncIO = malloc(sizeof(struct HvrAsyncIORecord))) == NULL) {
		if (LogIsError(Logger))
		  LogMessage(Logger, LogError, "malloc failed for client->AsyncIO.");
		ClientReply(current, HvrMemErr);
		for (errno = 0; close(fds[1]) == -1 && errno == EINTR; errno = 0);
		for (errno = 0; close(fds[0]) == -1 && errno == EINTR; errno = 0);
		break;
	      }
	      asyncIO = current->AsyncIO;
	    }
	    if ((asyncIO->Data = malloc(sizeof(int))) == NULL) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "malloc failed for asyncIO->Data.");
	      ClientReply(current, HvrMemErr);
	      for (errno = 0; close(fds[1]) == -1 && errno == EINTR; errno = 0);
	      for (errno = 0; close(fds[0]) == -1 && errno == EINTR; errno = 0);
	      for (asyncIO = current->AsyncIO; 
		   asyncIO->Next != NULL && asyncIO->Next->Next != NULL; 
		   asyncIO = asyncIO->Next);
	      if (asyncIO->Next == NULL) {
		free(asyncIO);
		current->AsyncIO = NULL;
	      }
	      else {
		free(asyncIO->Next);
		asyncIO->Next = NULL;
	      }
	      break;
	    }
	    asyncIO->Client = current;
	    asyncIO->Fds[0] = fds[0];
	    asyncIO->Fds[1] = fds[1];
	    asyncIO->Command = cmd;
	    asyncIO->Next = NULL;
	    
	    if ((ret = pthread_create(asyncIO->Data, NULL, GetLog, asyncIO))) {
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "pthread_create failed for GetLog.  %s.", strerror(ret));
	      ClientReply(current, HvrSysErr);
	      free(asyncIO->Data);
	      for (errno = 0; close(fds[1]) == -1 && errno == EINTR; errno = 0);
	      for (errno = 0; close(fds[0]) == -1 && errno == EINTR; errno = 0);
	      for (asyncIO = current->AsyncIO;
		   asyncIO->Next != NULL && asyncIO->Next->Next != NULL;
		   asyncIO = asyncIO->Next);
	      if (asyncIO->Next == NULL) {
		free(asyncIO);
		current->AsyncIO = NULL;
	      }
	      else {
		free(asyncIO->Next);
		asyncIO->Next = NULL;
	      }
	      break;
	    }
	    AddSelectIO(HvrHandle, fds[0], SelectRead);
	  }
	}
	  break;
	}
	if (retVal && CBufferSize(client->InBuffer) > 3) {
	  CBufferPeek(client->InBuffer, &cmd, 4);
	  cmd = ntohl(cmd);
	}
	else
	  cmd = -1;
      }
    }
    else
      CloseClient(client);
  }
    break;
  case ClientDisconnected:
    RemoveClient(client);
    break;
  }
  
  return NULL;
}
