/******************************************************************************\
 * Name: imageviewer.c                                                        *
 * Author: James Greensky                                                     *
 * Notes: implementation of non 3d accelerated image viewer                   *
\******************************************************************************/


#define _GNU_SOURCE

#include "logger.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <getopt.h>
#include <netdb.h>

#ifdef __DRI_

#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/extensions/Xdbe.h>

#else

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xdbe.h>

#endif

enum IvReturnCodes {
  IvOK = 0,
  IvMemErr = 1,
  IvReadProcErr = 2,
  IvHostNameErr = 3,
  IvLogErr = 4,
  IvXOpenErr = 5,
  IvXFatalErr = 6,
  IvReadErr = 7,
  IvStdInClsd = 8,
  IvWriteErr = 9,
  IvArgInvalid = 10,
  IvArgUnknown = 11,
  IvXVizErr = 12,
  IvSigErr = 13,
  IvXContextErr = 14,
  IvXCurrErr = 15,
  IvXConnErr = 16,
  IvShmErr = 17,
  IvShmAttErr = 18
};

typedef struct Window_Record {
  Display* _Display;
  int Screen, Width, Height, Depth;
  Window _Window, Root;
  unsigned long Foreground, Background;
  XWindowAttributes Attributes;

#ifdef __DRI_

  XVisualInfo* _Visual;
  GLXContext Context;

#else

  Visual* _Visual;
  XImage* Image;
  XShmSegmentInfo ShmInfo;
  XdbeBackBuffer Buffer;
  XdbeSwapInfo SwapInfo;
  Pixmap _Pixmap;
  GC _GC;

#endif

}* Window_;

typedef struct {
  int X, Y, Width, Height;
} Rectangle;

typedef struct FDRecord {
  int FD;
  struct FDRecord* Next;
}* FD;

static const struct option LongOptions[] = {
  { "debug", 0, NULL, 'd' },
  { "verbose", 0, NULL, 'v' },
  { "logLevel", 1, NULL, 'l' },
  { "help", 0, NULL, 'h' },
  { NULL, 0, NULL, 0 }
};

static const char* ShortOptions = "dvl:h";

static const char* const Usage =
"Usage: %s [ options ] \n"
"  -h, --help                  Print this information.\n"
"  -d, --debug                 Set the level of the logger to DEBUG (by\n"
"                                default, logger is set to WARN).\n"
"  -v, --verbose               Set the level of the logger to INFO.\n"
"  -l, --logLevel LLEVEL       Level of fmessages printed by logger.\n"
"                                Valid entries include:  OFF(0), FATAL(1),\n"
"                                ERROR(2), WARN(3), INFO(4), DEBUG(5), ALL(5).\n\n";

static Window_ _Window;
static int Length = 0, Closing = 0, Mapped = 0, FDMax = 0;
static char* Buffer;
static FD FDs = NULL;
static fd_set FDSet;
static Log Logger;

#ifdef __DRI_

static GLuint Texture;

#endif

static void PrintUsage(int exitCode) {
  fprintf((exitCode != 0) ? stderr : stdout, Usage, program_invocation_short_name);
  exit(exitCode);
}

static void XCleanUp();
static void CleanUp();

static FD AddFD(int fd) {
  FD fdRecord = malloc(sizeof(struct FDRecord));
  if (fdRecord == NULL) {
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "malloc failed on fdRecord.");
    XCleanUp();
    CleanUp();
    exit(IvMemErr);
  }
  fdRecord->FD = fd;
  fdRecord->Next = FDs;
  FDs = fdRecord;
  FD_SET(fd, &FDSet);
  if (fd > FDMax)
    FDMax = fd;
  return fdRecord;
}

static void RemoveFD(FD fdRecord) {
  FD current, previous;
  for (previous = NULL, current = FDs; current != NULL && 
	   current != fdRecord; previous = current, current = current->Next);
  if (previous == NULL)
    FDs = current->Next;
  else
    previous->Next = current->Next;
  FD_CLR(current->FD, &FDSet);
  if (current->FD == FDMax) {
    FDMax = STDIN_FILENO;
    for (previous = FDs; previous != NULL; previous = previous->Next) {
      if (previous->FD > FDMax)
	FDMax = previous->FD;
    }
  }
  free(current);
}

static void XConnectionHandler(Display* display, XPointer client_data,
			       int fd, Bool opening, XPointer* watch_data) {
  if (opening)
    *watch_data = (XPointer)AddFD(fd);
  else
    RemoveFD((FD)*watch_data);
}

static void XCleanUp() {

#ifdef __DRI_

  glXMakeCurrent(_Window->_Display, None, NULL);
  glXDestroyContext(_Window->_Display, _Window->Context);

#else

  XFreeGC(_Window->_Display, _Window->_GC);
  XFreePixmap(_Window->_Display, _Window->_Pixmap);
  XdbeDeallocateBackBufferName(_Window->_Display, _Window->Buffer);
  XDestroyWindow(_Window->_Display, _Window->_Window);
  XShmDetach(_Window->_Display, &_Window->ShmInfo);
  shmdt(_Window->ShmInfo.shmaddr);
  shmctl(_Window->ShmInfo.shmid, IPC_RMID, 0);
  XDestroyImage(_Window->Image);

#endif

  XRemoveConnectionWatch(_Window->_Display, XConnectionHandler, NULL);
  XCloseDisplay(_Window->_Display);
}

static void CleanUp() {
  FD current, save;
  for (current = FDs; current != NULL;) {
    save = current->Next;
    free(current);
    current = save;
  }
  LogClose(Logger);
  LogDispose(Logger);
  free(_Window);
  free(Buffer);
}

static void SIGTERM_Handler(int signal) {
  Closing = 1;
}

static int XFatalErrorHandler(Display* display) {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "XFatalErrorHandler called.");

#ifndef __DRI_

  shmdt(_Window->ShmInfo.shmaddr);
  shmctl(_Window->ShmInfo.shmid, IPC_RMID, 0);

#endif

  CleanUp();
  exit(IvXFatalErr);
}

static int XProtocolErrorHandler(Display* display, XErrorEvent* event) {
  char* buffer = malloc(sizeof(char) * 4096);
  if (buffer == NULL) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "malloc failed on buffer.");
    XCleanUp();
    CleanUp();
    exit(IvMemErr);
  }
  XGetErrorText(display, event->error_code, buffer, 4096);
  if (LogIsError(Logger))
    LogMessage(Logger, LogError, "XErrorHandler called.  %s.", buffer);
  free(buffer);
  return 0;
}

static void ProcessData() {
  int k = 0, j, retVal, width, height, ok = 0;

#ifndef __DRI_

  char *src, *dest;

#endif

  Rectangle rect;
  while (k < 16) {
    retVal = read(STDIN_FILENO, &((char*)&rect.X)[k], 16 - k);
    if (retVal < 0) {
      retVal = errno;
      if (retVal != EINTR) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "read failed.  %s.", strerror(retVal));
	XCleanUp();
	CleanUp();
	exit(IvReadErr);
      }
      else if (Closing) {
	XCleanUp();
	CleanUp();
	exit(0);
      }
    }
    else if (!retVal) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "stdin closed.");
      XCleanUp();
      CleanUp();
      exit(IvStdInClsd);
    }
    else
      k += retVal;
  }
  if (rect.X > -1) {
    k = rect.Width * rect.Height * 3;
    if (Length < k) {
      if (Length)
	free(Buffer);
      Length = 0;
      Buffer = malloc(sizeof(char) * k);
      if (Buffer == NULL) {
	if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "malloc failed on Buffer.");
	XCleanUp();
	CleanUp();
	exit(IvMemErr);
      }
      Length = k;
    }
    j = 0;
    while (j < k) {
      retVal = read(STDIN_FILENO, &Buffer[j], k - j);
      if (retVal < 0) {      
	retVal = errno;
	if (retVal != EINTR) {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "read failed.  %s.", strerror(retVal));
	  XCleanUp();
	  CleanUp();
	  exit(IvReadErr);
	}
	else if (Closing) {
	  XCleanUp();
	  CleanUp();
	  exit(0);
	}
      }
      else if (!retVal) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "stdin closed.");
	XCleanUp();
	CleanUp();
	exit(IvStdInClsd);
      }
      else
	j += retVal;
    }
    
    if (rect.X < _Window->Width && rect.Y < _Window->Height) {

#ifdef __DRI_

      float x = (float)rect.X, y = (float)(_Window->Height - rect.Y);

#endif

      width = rect.X + rect.Width;
      if (width > _Window->Width)
	width = _Window->Width - rect.X;
      else
	width = rect.Width;
      height = rect.Y + rect.Height;
      if (height > _Window->Height)
	height = _Window->Height - rect.Y;
      else
	height = rect.Height;

#ifdef __DRI_

    /* Create Texture */
    glDeleteTextures(1, &Texture);
    glGenTextures(1, &Texture);
    glBindTexture(GL_TEXTURE_2D, Texture);
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, Buffer);      
    
    /* Render Texture */
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    glBegin(GL_QUADS);
    
      /* Bottom Left */
      glTexCoord2f(0.0, 0.0);
      glVertex2f(x, y - height);
      /* Top Left */
      glTexCoord2f(0.0, 1.0);
      glVertex2f(x, y);
      /* Top Right */
      glTexCoord2f(1.0, 1.0);
      glVertex2f(x + width, y);
      /* Bottom Right */
      glTexCoord2f(1.0, 0.0);
      glVertex2f(x + width, y - height);
      
      glEnd();

      glFlush();

#else
      
      for (k = 0; k < height; k++) {
	src = Buffer + rect.Width * (rect.Height - 1) * 3 - k * rect.Width * 3 + 2;
	dest = _Window->Image->data +
	  rect.X * (_Window->Image->bits_per_pixel>>3) +
	  (rect.Y + k) * _Window->Image->bytes_per_line;
	for (j = 0; j < width; j++) {
	  *(dest++) = *(src--);
	  *(dest++) = *(src--);
	  *dest = *src;
	  dest += 2;
	  src += 5;
	}
      }

      XCopyArea(_Window->_Display, _Window->_Pixmap, _Window->Buffer, _Window->_GC,
		rect.X, rect.Y, rect.Width, rect.Height, rect.X, rect.Y);

#endif

    }
    
    k = 0;
    while (k < 4) {
      retVal = write(STDOUT_FILENO, &((char*)&ok)[k], 4 - k);
      if (retVal < 0) {
	retVal == errno;
	if (retVal != EINTR) {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "write failed on stdout.  %s.", strerror(retVal));
	  XCleanUp();
	  CleanUp();
	  exit(IvWriteErr);
	}
	else if (Closing) {
	  XCleanUp();
	  CleanUp();
	  exit(0);
	}
      }
      else
	k += retVal;
    }
  }
  else if (rect.X < 0) {
    k = 0;
    while (k < 4) {
      retVal = write(STDOUT_FILENO, &((char*)&ok)[k], 4 - k);
      if (retVal < 0) {
	retVal = errno;
	if (retVal != EINTR) {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "write failed on stdout.  %s.", strerror(retVal));
	  XCleanUp();
	  CleanUp();
	  exit(IvWriteErr);
	}
	else if (Closing) {
	  XCleanUp();
	  CleanUp();
	  exit(IvWriteErr);
	}
      }
      else
	k += retVal;
    }
    
    if (Mapped) {
      if (rect.Y > 0)
	usleep((unsigned int) rect.Y);

#ifdef __DRI_

      glXSwapBuffers(_Window->_Display, _Window->_Window);
      glFlush();

#else

      XdbeSwapBuffers(_Window->_Display, &_Window->SwapInfo, 1);
      XFlush(_Window->_Display);

#endif

    }
  }
}

static char* GetExecutableDirectory() {
  int retVal;
  char* lastSlash;
  size_t resultLength;
  char* result;
  char* linkTarget = malloc(sizeof(char) * 1024);
  if (linkTarget == NULL) {
    fprintf(stderr, "malloc failed on linkTarget.  aborting.\n");
    exit(IvMemErr);
  }

  retVal = readlink("/proc/self/exe", linkTarget, 1024);
  if (retVal == -1) {
    fprintf(stderr, "error reading executable directory from /proc/self/exe. aborting\n");
    exit(IvReadProcErr);
  }
  
  linkTarget[retVal] = '\0';
  lastSlash = strrchr(linkTarget, '/');
  if (lastSlash == NULL || lastSlash == linkTarget) {
    fprintf(stderr, "error parsing executable directory from /proc/self/exe.  aborting.\n");
    exit(IvReadProcErr);
  }

  resultLength = lastSlash - linkTarget;
  result = (char *) malloc(resultLength + 1);
  if (result == NULL) {
    fprintf(stderr, "malloc failed on result.  aborting.\n");
    exit(IvMemErr);
  }
  strncpy(result, linkTarget, resultLength);
  result[resultLength] = '\0';
  free(linkTarget);
  return result;
}

int main(int argc, char** argv) {
  XEvent e;
  struct sigaction sa;
  fd_set fds;
  int retVal, nextOption;
  char *firstDot, *logFile, *shortHostName, *xDirectory;
  FD fd;

#ifdef __DRI_

  int attr[] = {GLX_DOUBLEBUFFER, GLX_RGBA, None};

#endif
  
  xDirectory = GetExecutableDirectory();

  Buffer = malloc(sizeof(char) * 100);
  if (Buffer == NULL) {
    fprintf(stderr, "malloc failed on Buffer.\n");
    free(xDirectory);
    exit(IvMemErr);
  }
  Length = 100;

  if (gethostname(Buffer, 100) < 0) {
    int err = errno;
    fprintf(stderr, "gethostname failed.  %s.\n", strerror(err));
    free(Buffer);
    free(xDirectory);
    exit(IvHostNameErr);
  }
  Buffer[99] = '\0';
  if ((firstDot = strchr(Buffer, '.')) != NULL) {
    int length = firstDot - Buffer;
    shortHostName = malloc(length + 1);
    if (shortHostName == NULL) {
      fprintf(stderr, "malloc failed on shortHostName.\n");
      free(Buffer);
      free(xDirectory);
      exit(IvMemErr);
    }
    strncpy(shortHostName, Buffer, length);
    shortHostName[length] = '\0';
  }
  else {
    shortHostName = strndup(Buffer, 100);
    if (shortHostName == NULL) {
      fprintf(stderr, "strndup failed on shortHostName.\n");
      free(Buffer);
      free(xDirectory);
      exit(IvMemErr);
    }
  }

  logFile = malloc(sizeof(char) * (strlen(xDirectory) + strlen(Buffer) +
				   strlen(program_invocation_short_name) + 7));
  if (logFile == NULL) {
    fprintf(stderr, "create failed on Logger.\n");
    free(logFile);
    free(shortHostName);
    free(Buffer);
    free(xDirectory);
    exit(IvMemErr);
  }
  sprintf(logFile, "%s/%s_%s.log", xDirectory, program_invocation_short_name, shortHostName);
  Logger = LogCreate(program_invocation_short_name, LogWarn);
  if (Logger == NULL) {
    fprintf(stderr, "create failed on Logger.\n");
    free(logFile);
    free(shortHostName);
    free(Buffer);
    free(xDirectory);
    exit(IvMemErr);
  }
  if (LogOpen(Logger, logFile) < 0) {
    fprintf(stderr, "open failed on Logger.\n");
    LogDispose(Logger);
    free(logFile);
    free(shortHostName);
    free(Buffer);
    free(xDirectory);
    exit(IvLogErr);
  }

  free(logFile);
  free(shortHostName);
  free(xDirectory);

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = &SIGTERM_Handler;
  while (sigaction(SIGTERM, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGTERM.");
      LogClose(Logger);
      LogDispose(Logger);
      free(Buffer);
      exit(IvSigErr);
    }
  }  
  while (sigaction(SIGINT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGINT.");
      LogClose(Logger);
      LogDispose(Logger);
      free(Buffer);
      exit(IvSigErr);
    }
  }  
  while (sigaction(SIGQUIT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGQUIT.");
      LogClose(Logger);
      LogDispose(Logger);
      free(Buffer);
      exit(IvSigErr);
    }
  }  
  while (sigaction(SIGHUP, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGHUP.");
      LogClose(Logger);
      LogDispose(Logger);
      free(Buffer);
      exit(IvSigErr);
    }
  }

  _Window = malloc(sizeof(struct Window_Record));
  if (_Window == NULL) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "malloc failed on _Window.");
    LogClose(Logger);
    LogDispose(Logger);
    free(Buffer);
    exit(IvMemErr);
  }

  do {
    nextOption = getopt_long(argc, argv, ShortOptions, LongOptions, NULL);
    switch(nextOption) {
    case 'h':
      PrintUsage(0);
      break;
    case 'd':
      LogSetLevel(Logger, LogDebug);
      break;
    case 'v':
      LogSetLevel(Logger, LogInfo);
      break;
    case 'l': {
      long value;
      char* end;

      value = strtol(optarg, &end, 10);
      if (end != '\0') {
	if (strcasecmp(optarg, "off") == 0)
	  value = LogOff;
	else if (strcasecmp(optarg, "fatal") == 0)
	  value = LogFatal;
	else if (strcasecmp(optarg, "error") == 0)
	  value = LogError;
	else if (strcasecmp(optarg, "warn") == 0)
	  value = LogWarn;
	else if (strcasecmp(optarg, "info") == 0)
	  value = LogInfo;
	else if (strcasecmp(optarg, "debug") == 0)
	  value = LogDebug;
	else if (strcasecmp(optarg, "all") == 0)
	  value = LogAll;
	else
	  value = -1;
      }
      if (value == -1) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "LogLevel %s is not valid.", optarg);
	CleanUp();
	fprintf(stderr, "LogLevel %s is not valid.\n\n", optarg);
	PrintUsage(IvArgInvalid);
      }
      else if (value < LogOff || value > LogAll) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "LogLevel %ld is outside the valid range.", value);
	CleanUp();
	fprintf(stderr, "LogLevel %ld is outside the valid range.\n\n", value);
	PrintUsage(IvArgInvalid);
      }
      else
	LogSetLevel(Logger, (LogLevel)value);
    }
      break;
    case '?':
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "Invalid option specified.");
      CleanUp();
      fprintf(stderr, "Invalid option specified.\n\n");
      PrintUsage(IvArgUnknown);
      break;
    case -1:
      break;
    }
  } while (nextOption != -1);
    
  FD_ZERO(&FDSet);
  FD_SET(STDIN_FILENO, &FDSet);
  FDMax = STDIN_FILENO;
  if ((_Window->_Display = XOpenDisplay(NULL)) == NULL) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not open display.");
    CleanUp();
    exit(IvXOpenErr);
  }
  AddFD(ConnectionNumber(_Window->_Display));
  if (XAddConnectionWatch(_Window->_Display, XConnectionHandler, NULL) == 0) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "XAddConnectionWatch failed.");
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvXConnErr);
  }
    
  _Window->Screen = DefaultScreen(_Window->_Display);
  _Window->Root = XDefaultRootWindow(_Window->_Display);
  _Window->Depth = XDefaultDepth(_Window->_Display, _Window->Screen);
  _Window->Width = DisplayWidth(_Window->_Display, _Window->Screen);
  _Window->Height = DisplayHeight(_Window->_Display, _Window->Screen);
  _Window->Foreground = WhitePixel(_Window->_Display, _Window->Screen);
  _Window->Background = BlackPixel(_Window->_Display, _Window->Screen);
  XGetWindowAttributes(_Window->_Display, _Window->Root, &(_Window->Attributes));
  _Window->_Window = XCreateSimpleWindow(_Window->_Display, _Window->Root, 0, 0,
					 _Window->Width, _Window->Height, 0,
					 _Window->Background, _Window->Background);

#ifdef __DRI_

  if ((_Window->_Visual = glXChooseVisual(_Window->_Display, _Window->Screen, attr)) == NULL) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not get visual.");
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvXVizErr);
  }
  if ((_Window->Context = glXCreateContext(_Window->_Display, _Window->_Visual, 0, GL_TRUE)) == NULL) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not create rendering context.");
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvXContextErr);
  }
  if (!glXMakeCurrent(_Window->_Display, _Window->_Window, _Window->Context)) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not make context current.");
    glXDestroyContext(_Window->_Display, _Window->Context);
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvXCurrErr);
  }
  glDrawBuffer(GL_BACK);
  glReadBuffer(GL_BACK);
  
  /* Set flat 2D viewport */
  glViewport(0, 0, (GLsizei)_Window->Width, (GLsizei)_Window->Height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0, (GLsizei)_Window->Width, 0, (GLsizei)_Window->Height);
  glClear(GL_COLOR_BUFFER_BIT);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glEnable(GL_TEXTURE_2D);

#else

  _Window->_Visual = XDefaultVisual(_Window->_Display, _Window->Screen);
  _Window->Image = XShmCreateImage(_Window->_Display, _Window->_Visual, _Window->Depth, ZPixmap,
				   NULL, &_Window->ShmInfo, _Window->Width, _Window->Height);
  _Window->ShmInfo.shmid = shmget(IPC_PRIVATE,
				  _Window->Image->bytes_per_line * _Window->Image->height, 
				  IPC_CREAT | 0777);
  if (_Window->ShmInfo.shmid == -1) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not create shared memory");
    XDestroyImage(_Window->Image);
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvShmErr);
  }
    
  _Window->ShmInfo.shmaddr = _Window->Image->data = shmat(_Window->ShmInfo.shmid, 0, 0);
  if (_Window->ShmInfo.shmaddr == (void*)-1) {
    if (LogIsFatal(Logger))
      LogMessage(Logger, LogFatal, "could not attach to shared memory");
    shmctl(_Window->ShmInfo.shmid, IPC_RMID, 0);
    XDestroyImage(_Window->Image);
    XCloseDisplay(_Window->_Display);
    CleanUp();
    exit(IvShmAttErr);
  }
  _Window->ShmInfo.readOnly = 0;
  XShmAttach(_Window->_Display, &_Window->ShmInfo);
  _Window->SwapInfo.swap_window = _Window->_Window;
  _Window->SwapInfo.swap_action = XdbeUndefined;
  _Window->Buffer = XdbeAllocateBackBufferName(_Window->_Display, _Window->_Window, 
					       _Window->SwapInfo.swap_action);
  
  _Window->_Pixmap = XShmCreatePixmap(_Window->_Display, _Window->_Window, 
				      _Window->Image->data, &_Window->ShmInfo, 
				      _Window->Width, _Window->Height, _Window->Depth);
  
  _Window->_GC = XCreateGC(_Window->_Display, _Window->_Window, 0, 0);
  
#endif

  XSelectInput(_Window->_Display, _Window->_Window, StructureNotifyMask | ExposureMask);
  XSetIOErrorHandler(XFatalErrorHandler);
  XSetErrorHandler(XProtocolErrorHandler);
  
  XMapRaised(_Window->_Display, _Window->_Window);
  
  while (!Closing) {
    fds = FDSet;
    FD_SET(STDIN_FILENO, &fds);
    XFlush(_Window->_Display);
    retVal = select(FDMax + 1, &fds, NULL, NULL, NULL);
    if (!Closing) {
      if (retVal < 0) {
	retVal = errno;
	if (retVal != EINTR && LogIsError(Logger))
	  LogMessage(Logger, LogError, "select failed.  %s.", strerror(retVal));
      }
      else {
	if (FD_ISSET(STDIN_FILENO, &fds))
	  ProcessData();

	for (fd = FDs; fd != NULL; fd = fd->Next) {
	  if (FD_ISSET(fd->FD, &fds))
	    XProcessInternalConnection(_Window->_Display, fd->FD);
	}

	retVal = XPending(_Window->_Display);
	while (retVal > 0) {
	  XNextEvent(_Window->_Display, &e);
	  switch(e.type) {
	  case Expose:
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "e.type == Expose.");

#ifndef __DRI_

	    XCopyArea(_Window->_Display, _Window->_Pixmap, _Window->_Window, _Window->_GC,
		      e.xexpose.x, e.xexpose.y, e.xexpose.width, e.xexpose.height, 
		      e.xexpose.x, e.xexpose.y);

#endif

	    break;
	  case MapNotify:
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "e.type = MapNotify.");
	    Mapped = 1;
	    break;
	  case UnmapNotify:
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "e.type = UnmapNotify.");
	    Mapped = 0;
	    break;
	  default:
	    if (LogIsDebug(Logger)) {
	      switch(e.type) {
	      case CirculateNotify: LogMessage(Logger, LogDebug, "e.type = CirculateNotify."); break;
	      case ConfigureNotify: LogMessage(Logger, LogDebug, "e.type = ConfigureNotify."); break;
	      case DestroyNotify: LogMessage(Logger, LogDebug, "e.type = DestroyNotify."); break;
	      case GravityNotify: LogMessage(Logger, LogDebug, "e.type = GravityNotify."); break;
	      case ReparentNotify: LogMessage(Logger, LogDebug, "e.type = ReparentNotify."); break;
	      case KeyPress: LogMessage(Logger, LogDebug, "e.type = KeyPress."); break;
	      case KeyRelease: LogMessage(Logger, LogDebug, "e.type = KeyRelease."); break;
	      case ButtonPress: LogMessage(Logger, LogDebug, "e.type = ButtonPress."); break;
	      case ButtonRelease: LogMessage(Logger, LogDebug, "e.type = ButtonRelease."); break;
	      case MotionNotify: LogMessage(Logger, LogDebug, "e.type = MotionNotify."); break;
	      case EnterNotify: LogMessage(Logger, LogDebug, "e.type = EnterNotify."); break;
	      case LeaveNotify: LogMessage(Logger, LogDebug, "e.type = LeaveNofity."); break;
	      case FocusIn: LogMessage(Logger, LogDebug, "e.type = FocusIn."); break;
	      case FocusOut: LogMessage(Logger, LogDebug, "e.type = FocusOut."); break;
	      case KeymapNotify: LogMessage(Logger, LogDebug, "e.type = KeymapNotify."); break;
	      case GraphicsExpose: LogMessage(Logger, LogDebug, "e.type = GraphicsExpose."); break;
	      case CirculateRequest: LogMessage(Logger, LogDebug, "e.type = CirculateRequest."); break;
	      case ConfigureRequest: LogMessage(Logger, LogDebug, "e.type = ConfigureRequest."); break;
	      case MapRequest: LogMessage(Logger, LogDebug, "e.type = MapRequest."); break;
	      case ResizeRequest: LogMessage(Logger, LogDebug, "e.type = ResizeRequest."); break;
	      case MappingNotify: LogMessage(Logger, LogDebug, "e.type = MappingNotify."); break;
	      case VisibilityNotify: LogMessage(Logger, LogDebug, "e.type = VisibilityNotify."); break;
	      case ColormapNotify: LogMessage(Logger, LogDebug, "e.type = ColormapNotfiy."); break;
	      case ClientMessage: LogMessage(Logger, LogDebug, "e.type = ClientMessage."); break;
	      case PropertyNotify: LogMessage(Logger, LogDebug, "e.type = PropertyNotify."); break;
	      case SelectionClear: LogMessage(Logger, LogDebug, "e.type = SelectionClear."); break;
	      case SelectionNotify: LogMessage(Logger, LogDebug, "e.type = SelectionNotify."); break;
	      case SelectionRequest: LogMessage(Logger, LogDebug, "e.type = SelectionRequest."); break;
	      case NoExpose: LogMessage(Logger, LogDebug, "e.type = NoExpose."); break;
	      default: LogMessage(Logger, LogDebug, "e.type = unknown.");
	      }
	    }
	  }
	  if (!--retVal)
	    retVal = XPending(_Window->_Display);
	}
      }
    }
  }
  
  XCleanUp();
  CleanUp();
  return 0;
}
