/******************************************************************************\
 * Name: iv_handler.c                                                         *
 * Author: James Greensky                                                     *
 * Notes: Implementation of the Image Viewer handler                          *
\******************************************************************************/

#include "server.h"
#include "queue.h"
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

typedef enum IvCommands {
  IvImage = 0,
  IvSwap = 1,
  IvTimeStamp = 2,
  IvSwapAbs = 3
} IvCommand;

typedef enum IvResponses {
  IvOK = 0,
  IvEMemErr = 1,
  IvProcErr = 2,
  IvHNameErr = 3,
  IvLogErr = 4,
  IvXOpenErr = 5,
  IvXFatalErr = 6,
  IvEReadErr = 7,
  IvStdInClsd = 8,
  IvEWriteErr = 9,
  IvArgInvalid = 10,
  IvArgUnknown = 11,
  IvVizErr = 12,
  IvSigErr = 13,
  IvXContextErr = 14,
  IvXCurrentErr = 15,
  IvXConnErr = 16,
  IvShmErr = 17,
  IvShmAttErr = 18,
  IvMemErr = 19,
  IvPipeErr = 20,
  IvIOErr = 21,
  IvExecErr = 22,
  IvForkErr = 23,
  IvReadErr = 24,
  IvWriteErr = 25,
  IvInUse = 26,
  IvNFnd = 27,
  IvTimeErr = 28,
  IvError = 29
} IvResponse;

struct IvJobRecord;
typedef struct IvClientRecord {
  LcseClient Client;
  int Key;
  struct IvJobRecord* Jobs;
  struct IvClientRecord* Next;
}* IvClient;

typedef struct IvJobRecord {
  IvClient Client;
  int X, Y, Width, Height, Offset, Length;
  char* Image;
  struct IvJobRecord* Next;
}* IvJob;

extern Log Logger;
static void* IvHandle = NULL;
static Queue IvQ = NULL;
static int IvRandomFD = -1;
static int IvKey = 0;

static pid_t IvID = 0, UnclutterID = 0;
static LcseExecutable IvExecutable = NULL;
static IvClient IvClients = NULL;
static CBuffer IvBuffer = NULL;
static int IvTotal = -1, ReadErrors = 0, WriteErrors = 0, UnclutterRetries = 5;

void Initialize(void* handle) {
  assert(handle != NULL);

  char* argv[6];
  int ret;
  argv[0] = malloc(sizeof(char) * strlen(ModuleDirectory) + sizeof(char) * 17);
  if (argv[0] == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on imageviewer argv[0].");
    IvID = -2;
    return;
  }
  sprintf(argv[0], "%s/%s", ModuleDirectory, "imageviewer.exe");
  if ((argv[1] = malloc(sizeof(char) * 3)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on imageviewer argv[1].");
    IvID = -1;
    free(argv[0]);
    return;
  }
  sprintf(argv[1], "-l");
  if ((argv[2] = malloc(sizeof(char) * 3)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on imageviewer argv[2].");
    IvID = -1;
    free(argv[1]);
    free(argv[0]);
    return;
  }
  sprintf(argv[2], "%d", LogGetLevel(Logger));
  if ((argv[3] = malloc(sizeof(char) * 3)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on imageviewer argv[3].");
    IvID = -1;
    free(argv[2]);
    free(argv[1]);
    free(argv[0]);
    return;
  }
  sprintf(argv[3], "-s");
  if ((argv[4] = malloc(sizeof(char) * 20)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on imageviewer argv[4].");
    IvID = -1;
    free(argv[3]);
    free(argv[2]);
    free(argv[1]);
    free(argv[0]);
    return;
  }
  if ((ret = snprintf(argv[4], 20, "%lld", (long long)LogGetMaxSize(Logger))) >= 20) {
    free(argv[4]);
    if ((argv[4] = malloc(sizeof(char) * (ret + 1))) == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on imageviewer argv[4].");
      IvID = -1;
      free(argv[3]);
      free(argv[2]);
      free(argv[1]);
      free(argv[0]);
      return;
    }
    sprintf(argv[4], "%lld", (long long)LogGetMaxSize(Logger));
  }
  argv[5] = NULL;

  IvHandle = handle;
  IvID = SpawnExecutable(handle, argv);
  free(argv[4]);
  free(argv[3]);
  free(argv[2]);
  free(argv[1]);
  free(argv[0]);
  
  if ((argv[0] = malloc(sizeof(char) * 10)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on unclutter argv[0].");
    free(argv[0]);
    IvID = -2;
    return;
  }
  sprintf(argv[0], "%s", "unclutter");
  if ((argv[1] = malloc(sizeof(char) * 6)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on unclutter argv[1].");
    free(argv[0]);
    IvID = -2;
    return;
  }
  sprintf(argv[1], "%s", "-root");
  if ((argv[2] = malloc(sizeof(char) * 11)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on unclutter argv[2].");
    free(argv[1]);
    free(argv[0]);
    IvID = -2;
    return;
  }
  sprintf(argv[2], "%s", "-onescreen");
  argv[3] = NULL;

  UnclutterID = SpawnExecutable(handle, argv);
  free(argv[2]);
  free(argv[1]);
  free(argv[0]);

  IvBuffer = CBufferCreate(0, 1);
  if (IvBuffer == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create iv buffer failed.");
    free(argv[0]);
    IvID = -2;
    return;
  }
  IvQ = QueueCreate();
  if (IvQ == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create iv q failed.");
    CBufferDispose(IvBuffer);
    free(argv[0]);
    IvID = -2;
    return;
  }

  srand(time(NULL) + getpid());
  IvRandomFD = open("/dev/urandom", O_RDONLY);

}

static int RandomNumber(int min, int max) {
  assert(max > min);

  int retVal = -1;
  unsigned int value;

  if (IvRandomFD > -1) {
    char* next;
    int cnt;

    next = (char*) &value;
    cnt = sizeof(value);
    do {
      retVal = read(IvRandomFD, next, cnt);
      if (retVal < 0) {
	if (errno == EINTR)
	  retVal = 0;
	else {
	  while (close(IvRandomFD) < 0 && errno == EINTR);
	  IvRandomFD = -1;
	}
      }
      else {
	cnt -= retVal;
	next += retVal;
      }
    } while (cnt > 0 && retVal > -1);
  }

  if (retVal < 0)
    value = rand();

  return min + value % (max - min + 1);
}

static void DisposeJob(IvJob job) {
  assert(job != NULL);
  
  free(job->Image);
  free(job);
}

static void RemoveClient(const LcseClient client) {
  assert(client != NULL);
  
  IvClient current = NULL, previous = NULL;
  for (previous = NULL, current = IvClients; current != NULL && current->Client != client;
       previous = current, current = current->Next);
  if (current != NULL) {
    IvJob job = NULL;
    void* save;
    
    for (job = current->Jobs; job != NULL;) {
      job->Client = NULL;
      if (QueuePeek(IvQ) != job) {
	save = job->Next;
	QueueRemove(IvQ, job);
	DisposeJob(job);
	job = save;
      }
      else
	job = job->Next;
    }

    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Removing iv client.");

    if (previous == NULL)
      IvClients = current->Next;
    else
      previous->Next = current->Next;
    free(current);
  }
}

void Dispose() {
  void* save;
  IvClient client = NULL;
  IvJob job = NULL;

  if (IvID > 0)
    kill(IvID, SIGTERM);
  else if (IvExecutable != NULL)
    kill(IvExecutable->ProcessID, SIGTERM);

  if (UnclutterID > 0)
    kill(UnclutterID, SIGTERM);

  if (QueueSize(IvQ)) {
    job = QueuePeek(IvQ);
    if (job->Client == NULL)
      DisposeJob(job);
  }

  for (client = IvClients; client != NULL;) {
    for (job = client->Jobs; job != NULL;) {
      save = job->Next;
      DisposeJob(job);
      job = save;
    }
  }

  CBufferDispose(IvBuffer);
  QueueDispose(IvQ);
}

inline static int ClientOutput(IvClient client, void* buffer, int count) {
  assert(client != NULL);
  assert(buffer != NULL);
  assert(count > -1);

  if (CBufferWrite(client->Client->OutBuffer, buffer, count) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "write failed on client->Client->OutBuffer.");
    CloseClient(client->Client);
    return 0;
  }
  return 1;
}

inline static int JobOutput(IvJob job, void* buffer, int count) {
  assert(job != NULL);
  
  if (job->Client != NULL)
    return ClientOutput(job->Client, buffer, count);
  return 0;
}

inline static int ClientReply(IvClient client, IvResponse response) {
  assert(client != NULL);

  int reply = htonl(response);
  return ClientOutput(client, &reply, 4);
}

inline static int JobReply(IvJob job, IvResponse response) {
  assert(job != NULL);

  if (job->Client != NULL)
    return ClientReply(job->Client, response);
  return 0;
}

static void RemoveJob(IvJob job) {
  assert(job != NULL);
  
  if (job->Client != NULL) {
    IvJob previous = NULL, current = NULL;
    for (previous = NULL, current = job->Client->Jobs; current != NULL && current != job;
	 previous = current, current = current->Next);
    if (current != NULL) {
      if (previous == NULL)
	job->Client->Jobs = current->Next;
      else
	previous->Next = current->Next;
    }
  }
  DisposeJob(job);
}

static void StartJob() {
  assert(IvExecutable != NULL);
  assert(QueueSize(IvQ));
  
  IvJob job = QueuePeek(IvQ);
  int length = job->Length - job->Offset, multi = job->Offset, total = 0, retVal;
  
  if (!multi)
    IvTotal = -1;
  
  while (length > 0) {
    total += retVal = write(IvExecutable->StdIn, &job->Image[job->Offset], length);
    if (retVal > 0 && LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "wrote %d bytes to IvExecutable->StdIn", retVal);
    if (retVal > -1) {
      if (retVal > 0)
	WriteErrors = 0;
      if (!(length -= retVal) && multi)
	RemoveExecutableStdIn(IvExecutable);
      
      job->Offset += retVal;
    }
    else {
      retVal = errno;
      if (retVal != EINTR) {
	if (retVal == EAGAIN) {
	  if (!multi)
	    AddExecutableStdIn(IvExecutable);
	  length = 0;
	}
	else {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "write failed on IvExecutable->StdIn.  %s.",
		       strerror(retVal));
	  if (multi)
	    RemoveExecutableStdIn(IvExecutable);
	  WriteErrors++;
	  kill(IvExecutable->ProcessID, SIGTERM);
	  QueueDequeue(IvQ);
	  JobReply(job, IvWriteErr);
	  RemoveJob(job);
	  length = -1;
	}
      }
    }
  }
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "wrote a total of %d bytes to IvExecutable->StdIn, %d left.", 
	       total, job->Length - job->Offset);
}

inline static void JobCleanUp(int startJob) {
  IvJob job = QueueDequeue(IvQ);
  RemoveJob(job);
  CBufferClear(IvBuffer);

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Job cleanup.");

  if (startJob && QueueSize(IvQ))
    StartJob();
} 

struct timeval* TimeoutHandler() {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "TimeoutHandler called in iv.");
  assert(1);

  return NULL;
}

struct timeval* IOHandler(int fd, LcseIOEvent event) {
  if (LogIsFatal(Logger))
    LogMessage(Logger, LogFatal, "IOHandler called in iv.");
  assert(1);

  return NULL;
}

static void ProcessData(const LcseExecutable executable) {
  assert(executable != NULL);

  char* buffer = NULL;
  int length = 0, retVal = 1;

  if (executable->ProcessID != UnclutterID) {
    if (IvExecutable == NULL) {
      int err = EINTR;
      length = 1;
      buffer = CBufferGetWriteBlock(IvBuffer, &length);
      while (read(executable->StdOut, buffer, length) < 1 && err == EINTR) {
	err = errno;
	if (err != EINTR && err != EAGAIN) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "read failed on executable->StdOut.  %s.", strerror(err));
	  ReadErrors++;
	  kill(IvID, SIGTERM);
	}
      }
      if (err == EINTR) {
	IvExecutable = executable;
	IvID = 0;
	ReadErrors = 0;
	if (QueueSize(IvQ))
	  StartJob();
      }
      return;
    }
    
    if (QueueSize(IvQ)) {
      IvJob job = QueuePeek(IvQ);
      if (job->Offset == job->Length) {
	int count = 0, total = 0;
	retVal = 1;
	while (CBufferSize(IvBuffer) != 4 && IvTotal != 4 && retVal > 0) {
	  if (IvTotal > -1) {
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "reading form IvExecutable->StdOut to IvBuffer.");
	    count = 4 - IvTotal;
	    buffer = CBufferGetWriteBlock(IvBuffer, &count);
	    total += retVal = read(IvExecutable->StdOut, buffer, count);
	    if (LogIsDebug(Logger))
	      LogMessage(Logger, LogDebug, "read %d bytes from IvExecutable->StdOut.", retVal);
	    if (retVal < 0) {
	      int err = errno;
	      if (err == EINTR)
		retVal = 1;
	      else if (err != EAGAIN) {
		if (LogIsError(Logger))
		  LogMessage(Logger, LogError, "read failed on IvExecutable->StdOut.");
		ReadErrors++;
		kill(IvExecutable->ProcessID, SIGTERM);
		JobReply(job, IvReadErr);
		JobCleanUp(0);
		IvTotal = -1;
		return;
	      }
	      if (LogIsDebug(Logger))
		LogMessage(Logger, LogDebug, "Discarding bytes from IvExecutable->StdOut.");
	      CBufferClear(IvBuffer);
	    }
	  }
	  else {
	    count = 4 - CBufferSize(IvBuffer);
	    buffer = CBufferGetWriteBlock(IvBuffer, &count);
	    if (buffer == NULL) {
	      IvTotal = CBufferSize(IvBuffer);
	      CBufferClear(IvBuffer);
	      if (LogIsError(Logger))
		LogMessage(Logger, LogError, "getwriteblock failed on IvBuffer.");
	    }
	    else {
	      if (LogIsDebug(Logger))
		LogMessage(Logger, LogDebug, "reading from IvExecutable->StdOut to IvBuffer.");
	      total += retVal = read(IvExecutable->StdOut, buffer, count);
	      if (retVal < 0) {
		int err = errno;
		if (err == EINTR)
		  retVal = 1;
		else if (err != EAGAIN) {
		  if (LogIsError(Logger))
		    LogMessage(Logger, LogError, "read failed on IvExecutable->StdOut.");
		  ReadErrors++;
		  kill(IvExecutable->ProcessID, SIGTERM);
		  JobReply(job, IvReadErr);
		  JobCleanUp(0);
		  return;
		}
	      }
	      else if (retVal > 0) {
		CBufferCommitBlock(IvBuffer, retVal);
		if (LogIsDebug(Logger))
		  LogMessage(Logger, LogDebug, "read %d bytes from IvExecutable->StdOut.", retVal);
	      }
	    }
	  }
	}
	
	if (LogIsDebug(Logger))
	  LogMessage(Logger, LogDebug, "read a total of %d bytes from IvExecutable->StdOut.", total);
	
	if (CBufferSize(IvBuffer) == 4 || IvTotal == 4) {
	  if (IvTotal != 4) {
	    if (job->Client != NULL) {
	      IvResponse response;
	      if (LogIsDebug(Logger))
		LogMessage(Logger, LogDebug, "Job finished.");
	      CBufferRead(IvBuffer, &response, 4);
	      JobReply(job, response);
	    }
	  }
	  else {
	    JobReply(job, IvMemErr);
	    IvTotal = -1;
	  }
	  JobCleanUp(1);
	}
	return;
      }
    }
  }
  else {
    if ((buffer = malloc(sizeof(char) * 4096)) == NULL) {
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "malloc failed on unclutter buffer.");
      return;
    }
    for (errno = 0; read(executable->StdOut, buffer, 4096) > -1 || errno == EINTR; errno = 0);
    if (errno != 0 && LogIsWarn(Logger)) {
      int err = errno;
      LogMessage(Logger, LogWarn, "read failed on unclutter->StdOut.  %s.", strerror(err));
    }
    free(buffer);
    return;
  }

  for (CBufferClear(IvBuffer), length = 0, buffer = CBufferGetWriteBlock(IvBuffer, &length);
       retVal > 0;
       CBufferClear(IvBuffer), length = 0, buffer = CBufferGetWriteBlock(IvBuffer, &length))
    retVal = read(IvExecutable->StdOut, buffer, length);
}	

struct timeval* ExecutableHandler(const LcseExecutable executable, LcseExecutableEvent event) {
  assert(executable != NULL);

  switch(event) {
  case ExecutableDataReady:
    ProcessData(executable);
    break;
  case ExecutableWriteReady:
    if (executable->ProcessID == UnclutterID) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "UnclutterID passed into ExecutableWriteReady.");
      break;
    }
    StartJob();
    break;
  case ExecutableExited:
    if (executable->ProcessID != UnclutterID) {
      if (QueueSize(IvQ)) {
	IvJob job = QueuePeek(IvQ);
	if (job->Offset > 0) {
	  if (!executable->Status)
	    JobReply(job, IvError);
	  else {
	    int k = htonl(executable->Status);
	    JobOutput(job, &k, 4);
	  }
	  QueueDequeue(IvQ);
	  RemoveJob(job);
	}
      }
      CBufferClear(IvBuffer);
      if (executable->Status == 127)
	IvID = -5;
      else if (ReadErrors > 4)
	IvID = -6;
      else if (WriteErrors > 4)
	IvID = -7;
      else
	IvID = SpawnExecutable(IvHandle, executable->Argv);
      if (IvID < 0 && QueueSize(IvQ)) {
	IvJob job = NULL;
	int response = htonl(IvError);
	switch (IvID) {
	case -1: response = htonl(IvForkErr); break;
	case -2: response = htonl(IvMemErr); break;
	case -3: response = htonl(IvPipeErr); break;
	case -4: response = htonl(IvIOErr); break;
	case -5: response = htonl(IvExecErr); break;
	case -6: response = htonl(IvReadErr); break;
	case -7: response = htonl(IvWriteErr); break;
	}
	for (job = QueueDequeue(IvQ); QueueSize(IvQ); job = QueueDequeue(IvQ)) {
	  if (JobReply(job, response)) {
	    job->Client->Jobs = NULL;
	    CloseClient(job->Client->Client);
	  }
	  DisposeJob(job);
	}
      }
      IvExecutable = NULL;
    }
    else {
      if (--UnclutterRetries > -1)
	UnclutterID = SpawnExecutable(IvHandle, executable->Argv);
    }
    break;
  }
  return NULL;
}

static int AddSwap(IvClient client) {
  assert(client != NULL);

  int cmd;
  CBufferPeek(client->Client->InBuffer, &cmd, 4);
  cmd = ntohl(cmd);
  if (cmd == IvSwap && CBufferSize(client->Client->InBuffer) < 8)
    return 0;
  else if (cmd == IvSwapAbs && CBufferSize(client->Client->InBuffer) < 12)
    return 0;
  else {
    IvJob job = malloc(sizeof(struct IvJobRecord));
    if (job == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job.");
      CBufferRemove(client->Client->InBuffer, 8);
      ClientReply(client, IvMemErr);
      return 1;
    }

    job->Image = malloc(sizeof(char) * 16);
    if (job->Image == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job.");
      CBufferRemove(client->Client->InBuffer, 8);
      ClientReply(client, IvMemErr);
      return 1;
    }

    if (cmd == IvSwap)
      job->X = -1;
    else
      job->X = -2;
    job->Width = job->Height = job->Offset = 0;
    job->Length = 16;
    CBufferRemove(client->Client->InBuffer, 4);
    CBufferRead(client->Client->InBuffer, &job->Y, 4);
    job->Y = ntohl(job->Y);
    if (cmd == IvSwapAbs) {
      CBufferRead(client->Client->InBuffer, &job->Width, 4);
      job->Width = ntohl(job->Width);
    }
    memcpy(job->Image, &job->X, 4);
    memcpy(&job->Image[4], &job->Y, 4);
    if (cmd == IvSwap)
      memset(&job->Image[8], '\0', 8);
    else {
      memcpy(&job->Image[8], &job->Width, 4);
      memset(&job->Image[12], '\0', 4);
    }
    job->Client = client;
    job->Next = client->Jobs;
    client->Jobs = job->Next;

    if (QueueEnqueue(IvQ, job) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "enqueue failed on iv q.");
      free(job->Image);
      free(job);
      ClientReply(client, IvMemErr);
      return 1;
    }

    if (QueueSize(IvQ) == 1 && IvExecutable != NULL)
      StartJob();
  }
  return 1;
}

static int AddJob(IvClient client) {
  assert(client != NULL);
  
  if (CBufferSize(client->Client->InBuffer) < 20)
    return 0;
  else {
    int width, height, length, total;
    CBufferPeekAt(client->Client->InBuffer, &width, 4, 4);
    width = ntohl(width);
    CBufferPeekAt(client->Client->InBuffer, &height, 8, 4);
    height = ntohl(height);
    length = width * height * 3;
    total = length + 16;
    if (CBufferSize(client->Client->InBuffer) < total + 4)
      return 0;
    
    IvJob job = malloc(sizeof(struct IvJobRecord));
    if (job == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job.");
      CBufferRemove(client->Client->InBuffer, total + 4);
      ClientReply(client, IvMemErr);
      return 1;
    }

    job->Image = malloc(sizeof(char) * total);
    if (job->Image == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on job->Image.");
      free(job);
      CBufferRemove(client->Client->InBuffer, total + 4);
      ClientReply(client, IvMemErr);
      return 1;
    }
    
    job->Width = width;
    job->Height = height;
    job->Length = total;
    job->Offset = 0;
    CBufferRemove(client->Client->InBuffer, 12);
    CBufferRead(client->Client->InBuffer, &job->X, 4);
    job->X = ntohl(job->X);
    CBufferRead(client->Client->InBuffer, &job->Y, 4);
    job->Y = ntohl(job->Y);
    memcpy(job->Image, &job->X, 4);
    memcpy(&job->Image[4], &job->Y, 4);
    memcpy(&job->Image[8], &job->Width, 4);
    memcpy(&job->Image[12], &job->Height, 4);
    CBufferRead(client->Client->InBuffer, &job->Image[16], length);
    job->Client = client;
    job->Next = client->Jobs;
    client->Jobs = job;

    if (QueueEnqueue(IvQ, job) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "enqueue failed on iv q.");
      free(job->Image);
      free(job);
      ClientReply(client, IvMemErr);
      return 1;
    }
    
    if (QueueSize(IvQ) == 1 && IvExecutable != NULL)
      StartJob();
  }
  
  return 1;
}

static void AddClient(const LcseClient client) {
  assert(client != NULL);
  
  if (IvID < 0) {
    int response = htonl(IvError);
    switch(IvID) {
    case -1: response = htonl(IvForkErr); break;
    case -2: response = htonl(IvMemErr); break;
    case -3: response = htonl(IvPipeErr); break;
    case -4: response = htonl(IvIOErr); break;
    case -5: response = htonl(IvExecErr); break;
    case -6: response = htonl(IvReadErr); break;
    case -7: response = htonl(IvWriteErr); break;
    }
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    CloseClient(client);
    return;
  }
  
  IvClient ivClient = malloc(sizeof(struct IvClientRecord));
  if (ivClient == NULL) {
    int response = htonl(IvMemErr);
    if (CBufferWrite(client->OutBuffer, &response, 4) < 0) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer.");
    }
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on ivClient.");
    CloseClient(client);
    return;
  }
  
  ivClient->Client = client;
  ivClient->Key = -1;
  ivClient->Jobs = NULL;
  ivClient->Next = IvClients;
  IvClients = ivClient;
  ClientReply(ivClient, IvOK);
}

static int DoTimeStamp(IvClient client) {
  assert(client != NULL);

  if (CBufferSize(client->Client->InBuffer) < 12)
    return 0;
  else {
    struct timeval t;
    time_t sec;
    suseconds_t usec;
    int k, cnt, ret;
    char* buffer;

    if (gettimeofday(&t, NULL)) {
      if (LogIsError(Logger)) 
	LogMessage(Logger, LogError, "gettimeofday failed on job->Created.  %s.", strerror(errno));
      CBufferRemove(client->Client->InBuffer, 12);		
      ClientReply(client, IvTimeErr);
      return 1;
    }

    CBufferRemove(client->Client->InBuffer, 4);
    if (!ClientReply(client, IvOK))
      return 1;
      
    for (k = 0; k < 8; 
	 k += ret = CBufferRead(client->Client->InBuffer, buffer, cnt), 
	   CBufferCommitBlock(client->Client->OutBuffer, ret)) {
      cnt = 8 - k;
      if ((buffer = CBufferGetWriteBlock(client->Client->OutBuffer, &cnt)) == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "getWriteBlock failed on client->OutBuffer.");
	CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - (5 + k), 4 + k);
	CBufferRemove(client->Client->InBuffer, 8 - k);
	ClientReply(client, IvMemErr);
	return 1;
      }
    }

    sec = htonl(t.tv_sec);
    if (CBufferWrite(client->Client->OutBuffer, &sec, 4) == -1) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer");
      CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - 13, 12);
      ClientReply(client, IvMemErr);
      return 1;
    }
    usec = htonl(t.tv_usec);
    if (CBufferWrite(client->Client->OutBuffer, &usec, 4) == -1) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer");
      CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - 17, 16);
      ClientReply(client, IvMemErr);
      return 1;
    }

    if (gettimeofday(&t, NULL)) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "gettimeofday failed on t.  %s.", strerror(errno));
      CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - 21, 20);
      ClientReply(client, IvTimeErr);
      return 1;
    }

    sec = htonl(t.tv_sec);
    if (CBufferWrite(client->Client->OutBuffer, &sec, 4) == -1) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer");
      CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - 21, 20);
      ClientReply(client, IvMemErr);
      return -1;
    }
    usec = htonl(t.tv_usec);
    if (CBufferWrite(client->Client->OutBuffer, &usec, 4) == -1) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "write failed on client->OutBuffer");
      CBufferRemoveAt(client->Client->OutBuffer, CBufferSize(client->Client->OutBuffer) - 25, 24);
      ClientReply(client, IvMemErr);
    }

    return 1;
  }
}

struct timeval* ClientHandler(const LcseClient client, LcseClientEvent event) {
  assert(client != NULL);
  
  switch(event) {
  case ClientAccepted:
    AddClient(client);
    break;
  case ClientDataReceived: {
    int cmd = -1;
    IvClient current = NULL;
    for (current = IvClients; current != NULL && current->Client != client; current = current->Next);
    if (current != NULL) {
      if (current->Key > -1) {
	if (CBufferSize(client->InBuffer) > 3) {
	  CBufferPeek(client->InBuffer, &cmd, 4);
	  cmd = ntohl(cmd);
	}
	while (cmd != -1) {
	  int retVal = 0;
	  switch(cmd) {
	  case IvImage:
	    retVal = AddJob(current);
	    break;
	  case IvSwap:
	  case IvSwapAbs:
	    retVal = AddSwap(current);
	    break;
	  case IvTimeStamp:
	    retVal = DoTimeStamp(current);
	    break;
	  }
	  if (retVal && CBufferSize(client->InBuffer) > 3) {
	    CBufferPeek(client->InBuffer, &cmd, 4);
	    cmd = ntohl(cmd);
	  }
	  else
	    cmd = -1;
	}
      }
      else {
	if (CBufferSize(client->InBuffer) > 3) {
	  CBufferRead(client->InBuffer, &current->Key, 4);
	  current->Key = ntohl(current->Key);
	  if (IvKey) {
	    if (current->Key == IvKey)
	      ClientReply(current, IvOK);
	    else {
	      current->Key = -1;
	      ClientReply(current, IvInUse);
	      CloseClient(current->Client);
	    }
	  } 
	  else {
	    if (current->Key) {
	      ClientReply(current, IvNFnd);
	      CloseClient(current->Client);
	    }
	    else {
	      IvKey = RandomNumber(1, INT_MAX);
	      ClientReply(current, IvOK);
	      int key = htonl(IvKey);
	      ClientOutput(current, &key, 4);
	    }
	  }
	}
      }
    }
    else
      CloseClient(client);
  }
    break;
  case ClientDisconnected: {
    IvClient current;
    for (current = IvClients; current != NULL && current->Client != client; current = current->Next);
    if (current != NULL) {
      if (!current->Key) {
	IvClient save;
	for (current = IvClients; current != NULL;) {
	  save = current->Next;
	  if (current->Client != client)
	    CloseClient(current->Client);
	  current = save;
	}
      }
      RemoveClient(client);
    }
  }
    break;
  }

  return NULL;
}
  
