/******************************************************************************\
 * File: logger.c                                                             *
 * Author: James Greensky                                                     *
 * Notes: Implement logging routines                                          *
\******************************************************************************/

#define _GNU_SOURCE

#include "logger.h"
#include "cbuffer.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <pthread.h>

struct LogRecord {
  char* Name;
  int FD, DirFD;
  off_t MaxSize;
  char* Path;
  pthread_mutex_t BufferMutex, FileMutex, Mutex;
  LogLevel Level;
  CBuffer Buffer;
};

inline int LogIsFatal(Log log) {
  assert(log != NULL);
  if (log->Level >= LogFatal)
    return (log->FD != -1);
  return 0;
}

inline int LogIsError(Log log) {
  assert(log != NULL);

  if (log->Level >= LogError)
    return (log->FD != -1);
  return 0;
}

inline int LogIsWarn(Log log) {
  assert(log != NULL);
  
  if (log->Level >= LogWarn)
    return (log->FD != -1);
  return 0;
}

inline int LogIsInfo(Log log) {
  assert(log != NULL);

  if (log->Level >= LogInfo)
    return (log->FD != -1);
  return 0;
}

inline int LogIsDebug(Log log) {
  assert(log != NULL);

  if (log->Level >= LogDebug)
    return (log->FD != -1);
  return 0;
}

Log LogCreate(const char* name) {
  assert(name != NULL);

  Log log;
  
  if ((log = malloc(sizeof(struct LogRecord))) == NULL)
    return log;
  if ((log->Name = strdup(name)) == NULL) {
    free(log);
    return NULL;
  }
  if ((log->Buffer = CBufferCreate(256, sizeof(char))) == NULL) {
    free(log->Name);
    free(log);
    return NULL;
  }
  pthread_mutex_init(&log->BufferMutex, NULL);
  pthread_mutex_init(&log->FileMutex, NULL);
  pthread_mutex_init(&log->Mutex, NULL);
  log->DirFD = -1;
  log->FD = -1;
  log->Level = LogWarn;
  log->MaxSize = 2097152;

  return log;
}

void LogSetLevel(Log log, LogLevel logLevel) {
  assert(log != NULL);

  log->Level = logLevel;
}

LogLevel LogGetLevel(Log log) {
  assert(log != NULL);

  return log->Level;
}

char* LogGetPath(Log log) {
  assert(log != NULL);
 
  return log->Path;
}

void LogSetMaxSize(Log log, off_t maxSize) {
  assert(log != NULL);

  log->MaxSize = maxSize;
}

off_t LogGetMaxSize(Log log) {
  assert(log != NULL);
  
  return log->MaxSize;
}

static void Close(Log log) {
  assert(log != NULL);

  for (errno = 0;
       close(log->DirFD) == -1 && errno == EINTR;
       errno = 0);
  log->DirFD = -1;
  if (log->Path)
    free(log->Path);
  log->Path = NULL;
  for (errno = 0; 
       close(log->FD) == -1 && errno == EINTR; 
       errno = 0);
  log->FD = -1;
}

void LogDispose(Log log) {
  assert(log != NULL);

  if (pthread_mutex_trylock(&log->Mutex))
    return;

  if (!log->Name) {
    pthread_mutex_unlock(&log->Mutex);
    return;
  }

  pthread_mutex_lock(&log->FileMutex);
  if (log->FD != -1)
    Close(log);
  pthread_mutex_unlock(&log->FileMutex);
  for (;pthread_mutex_destroy(&log->FileMutex) == EINTR;);
  
  pthread_mutex_lock(&log->BufferMutex);
  if (log->Buffer != NULL)
    CBufferDispose(log->Buffer);
  log->Buffer = NULL;
  pthread_mutex_unlock(&log->BufferMutex);
  for (;pthread_mutex_destroy(&log->BufferMutex) == EINTR;);
  
  free(log->Name);
  log->Name = NULL;
  
  pthread_mutex_unlock(&log->Mutex);
  for (;pthread_mutex_destroy(&log->Mutex) == EINTR;);

  free(log);
}

static int Open(Log log) {
  assert(log != NULL);

  int opts;
  if (log->FD != -1) {
    for (errno = 0; close(log->FD) == -1 && errno == EINTR; errno = 0);
    log->FD = -1;
  }

  if ((log->FD = open(log->Path, O_WRONLY | O_CREAT | O_TRUNC, 0600)) > -1) {
    opts = fcntl(log->FD, F_GETFL);
    if ((opts = fcntl(log->FD, F_GETFL)) == -1)
      Close(log);
    else {
      opts = (opts | O_NONBLOCK);
      if (fcntl(log->FD, F_SETFL, opts) == -1)
	Close(log);
    }
  }
  else
    Close(log);
  
  return log->FD;
}

int LogOpen(Log log, char* path) {
  assert(log != NULL);
  assert(path != NULL);
 
  char* slash;
  
  if (pthread_mutex_lock(&log->FileMutex))
    return -2;
  
  if (log->FD != -1)
    LogClose(log);

  if ((slash = strrchr(path, '/')) == NULL) {
    char* dir = get_current_dir_name();
    if (dir == NULL)
      log->DirFD = -1;
    else {
      log->DirFD = open(dir, O_RDONLY);
      free(dir);
    }
  }
  else {
    char* next = (char*)(slash + 1);
    char c = *next;
    *next = '\0';
    log->DirFD = open(path, O_RDONLY);
    *next = c;
  }
  log->Path = strndup(path, 256);

  Open(log);

  pthread_mutex_unlock(&log->FileMutex);
  
  return log->FD;
}

static int Flush(Log log) {
  assert(log != NULL);

  int total = CBufferSize(log->Buffer);
  if (total) {
    char* buffer;
    int remaining, count, retVal;
    fd_set fds;
    
    remaining = total;
    for (;remaining;) {
      FD_ZERO(&fds);
      FD_SET(log->FD, &fds);
      count = remaining;
      buffer = CBufferGetReadBlock(log->Buffer, &count);
      retVal = write(log->FD, buffer, count);
      if (retVal > -1) {
	remaining -= retVal;
	CBufferDecommitBlock(log->Buffer, retVal);
      }
      else {
	if (errno != EINTR) {
	  if (errno != EAGAIN) {
	    fsync(log->DirFD);
	    fsync(log->FD);
	    Close(log);
	  }
	  else
	    for (errno = 0; select(log->FD + 1, NULL, &fds, NULL, NULL) == -1 && errno == EINTR; errno = 0);
	}
      }
    }
  }

  return 0;
}

int LogFlush(Log log) {
  assert(log != NULL);

  int retVal = 0;

  if (pthread_mutex_lock(&log->FileMutex))
    return -2;

  if (log->FD > -1) {

    if (pthread_mutex_lock(&log->BufferMutex)) {
      pthread_mutex_unlock(&log->FileMutex);
      return -2;
    }

    retVal = Flush(log);
    
    pthread_mutex_unlock(&log->BufferMutex);

    fsync(log->DirFD);
    fsync(log->FD);
  }

  pthread_mutex_unlock(&log->FileMutex);

  return retVal;
}

static int RecordMessage(Log log) {
  assert(log != NULL);

  char* buffer;
  int remaining, count, total, retVal, ret = 0;
  struct stat info;

  total = remaining = CBufferSize(log->Buffer);
  while (remaining) {
    count = remaining;
    buffer = CBufferGetReadBlock(log->Buffer, &count);
    retVal = write(log->FD, buffer, count);
    if (retVal > -1) {
      ret += retVal;
      remaining -= retVal;
      CBufferDecommitBlock(log->Buffer, retVal);
    }
    else {
      if (errno != EINTR) {
	if (errno != EAGAIN)
	  Close(log);
	ret = -1;
	break;
      }
    }
  }
  
  if (log->Path && !fstat(log->FD, &info)) {
    if (info.st_size > log->MaxSize) {
      char* path;
      for (errno = 0; close(log->FD) == -1 && errno == EINTR; errno = 0);
      log->FD = -1;
      if ((path = malloc(sizeof(char) * (strlen(log->Path) + 3))) != NULL) {
	sprintf(path, "%s.0", log->Path);
	rename(log->Path, path);
	free(path);
      }
      Open(log);
    }
  }

  return ret;
}

void LogClose(Log log) {
  assert(log != NULL);

  if (pthread_mutex_lock(&log->FileMutex))
    return;

  if (log->FD != -1) {

    if (pthread_mutex_lock(&log->BufferMutex)) {
      pthread_mutex_unlock(&log->FileMutex);
      return;
    }

    if (CBufferSize(log->Buffer))
      Flush(log);

    pthread_mutex_unlock(&log->BufferMutex);

    Close(log);
  }

  pthread_mutex_unlock(&log->FileMutex);
}

static int LogWriteHeader(CBuffer buffer, char* name, LogLevel logLevel) {
  assert(buffer != NULL);
  assert(name != NULL);

  struct timeval tv;
  struct tm* ptm;
  char *output = CBufferGetWriteBlockAbs(buffer, 20);
  if (output == NULL)
    return -1;

  gettimeofday(&tv, NULL);
  ptm = localtime(&tv.tv_sec);
  strftime(output, 20, "%Y-%m-%d %H:%M:%S", ptm);
  CBufferCommitBlock(buffer, 19);
  output = CBufferGetWriteBlockAbs(buffer, 20);
  if (output == NULL) {
    CBufferRemoveAt(buffer, CBufferSize(buffer) - 19, 19);
    return -1;
  }
  sprintf(output, ".%06ld %-10.10s ", tv.tv_usec, name);
  CBufferCommitBlock(buffer, 19);
  output = CBufferGetWriteBlockAbs(buffer, 9);
  if (output == NULL) {
    CBufferRemoveAt(buffer, CBufferSize(buffer) - 39, 39);
    return -1;
  }
  switch(logLevel) {
  case(LogFatal): sprintf(output, "%s - ", "FATAL"); break;
  case(LogError): sprintf(output, "%s - ", "ERROR"); break;
  case(LogWarn): sprintf(output, "%s - ", "WARN "); break;
  case(LogInfo): sprintf(output, "%s - ", "INFO "); break;
  case(LogDebug): sprintf(output, "%s - ", "DEBUG"); break;
  default: sprintf(output, "%s", "?BUG? - ");
  }
  CBufferCommitBlock(buffer, 8);

  return 46;
}

int LogMessage(Log log, LogLevel logLevel, const char* format, ...) {
  assert(log != NULL);
  assert(format != NULL);

  int headerLength = 0, retVal = 0, cnt = 0;

  if (pthread_mutex_lock(&log->FileMutex))
    return -2;

  if (log->FD == -1) {
    pthread_mutex_unlock(&log->FileMutex);
    return -1;
  }

  if (logLevel <= log->Level) {
    va_list ap;
    int length;
    char* buffer;

    if (pthread_mutex_lock(&log->BufferMutex)) {
      pthread_mutex_unlock(&log->FileMutex);
      return -2;
    }

    headerLength = LogWriteHeader(log->Buffer, log->Name, logLevel);
    if (headerLength < 0) {
      char errorMsg[] = "Write to buffer failed.\n";
      cnt = CBufferWrite(log->Buffer, errorMsg, strlen(errorMsg));
      if (cnt < 0)
	fprintf(stderr, "Logger error.  %s", errorMsg);
      else
	RecordMessage(log);

      pthread_mutex_unlock(&log->BufferMutex);
      pthread_mutex_unlock(&log->FileMutex);

      return -3;
    }
    
    length = CBufferSize(log->Buffer);
    buffer = CBufferGetWriteBlock(log->Buffer, &cnt);
    if (buffer == NULL) {
      char errorMsg[] = "Write to buffer failed.\n";
      cnt = CBufferWrite(log->Buffer, errorMsg, strlen(errorMsg));
      if (cnt < 0)
	fprintf(stderr, "Logger error.  %s", errorMsg);
      else
	RecordMessage(log);

      pthread_mutex_unlock(&log->BufferMutex);
      pthread_mutex_unlock(&log->FileMutex);

      return -3;
    }

    va_start(ap, format);
    retVal = vsnprintf(buffer, cnt, format, ap);
    va_end(ap);
    if (++retVal >= cnt) {
      buffer = CBufferGetWriteBlockAbs(log->Buffer, retVal + 1);
      if (buffer == NULL) {
	char errorMsg[] = "Write to buffer failed.\n";
	cnt = CBufferWrite(log->Buffer, errorMsg, strlen(errorMsg));
	if (cnt < 0)
	  fprintf(stderr, "Logger error.  %s", errorMsg);
	else
	  RecordMessage(log);

	pthread_mutex_unlock(&log->BufferMutex);
	pthread_mutex_unlock(&log->FileMutex);

	return -3;
      }
      va_start(ap, format);
      vsprintf(buffer, format, ap);
      va_end(ap);
    }
    buffer[retVal] = '\n';
    CBufferCommitBlock(log->Buffer, retVal + 1);

    cnt = RecordMessage(log);

    pthread_mutex_unlock(&log->BufferMutex);
  }
  pthread_mutex_unlock(&log->FileMutex);

  if (cnt == -1)
    return -4;
  return headerLength + retVal;
}
