/******************************************************************************\
 * File: main.c                                                               *
 * Author: James Greensky                                                     *
 * Notes: Entry point                                                         *
\******************************************************************************/

#define _GNU_SOURCE

#include "logger.h"
#include "server.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>

enum MReturnCodes {
  M_OK = 0,
  M_MemErr = 1,
  M_LogErr = 2,
  M_ProcReadErr = 3,
  M_ArgInvalid = 4,
  M_ArgUnknown = 5,
  M_HostNameErr = 6,
  M_SockErr = 7,
  M_IFaceErr = 8,
  M_SigErr = 9,
  M_NBlockErr = 10,
  M_CloseExecErr = 11,
  M_ReuseErr = 12,
  M_DeferAccErr = 13,
  M_BindErr = 14,
  M_ListenErr = 15,
  M_IbSockErr = 16,
  M_IbNBlockErr = 17,
  M_IbCloseExecErr = 18,
  M_IbReuseErr = 19,
  M_IbDeferAccErr = 20,
  M_IbBindErr = 21,
  M_IbListerErr = 22,
  M_AddrErr = 23,
  M_PortErr = 24
};

static const struct option LongOptions[] = {
  { "address", 1, NULL, 'a' },
  { "help", 0, NULL, 'h' },
  { "debug", 0, NULL, 'd' },
  { "verbose", 0, NULL, 'v' },
  { "logLevel", 1, NULL, 'l' },
  { "module-dir", 1, NULL, 'm' },
  { "size", 1, NULL, 's' },
  { NULL, 0, NULL, 0 }
};

static const char* ShortOptions = "a:hdvl:m:s:";

static const char* const Usage = 
"Usage: %s [ options ] \n"
"  -a, --address ADDR          Bind to the local address(es).  Specify\n"
"                                addresses in complete address:port:domain\n"
"                                format, i.e. 192.168.1.1:9005:26.  Seperate\n"
"                                multiple addresses by a comma. (by default,\n"
"                                bind to all addresses (INADDR_ANY) on port\n"
"                                9005 in the pf_inet(2) domain.\n"
"  -h, --help                  Print this information.\n"
"  -d, --debug                 Set the level of the logger to DEBUG (by\n"
"                                default, logger is set to WARN).\n"
"  -v, --verbose               Set the level of the logger to INFO.\n"
"  -l, --logLevel LLEVEL       Level of messages printed by logger.\n"
"                                Valid entries include: OFF(0), FATAL(1),\n"
"                                ERROR(2), WARN(3), INFO(4), DEBUG(5), ALL(5).\n"
"  -m, --module-dir DIR        Load modules from specified directory\n"
"                                (by default, load from executable directory).\n"
"  -s, --size SIZE             Specify the maximum size of the log file, if\n"
"                                exceeded it saves the current log file to\n"
"                                [log filename].0 and creates a new log file\n"
"                                (by default, size is set to 2 MB).\n";

static void PrintUsage(int exitCode) {
  fprintf((exitCode != 0) ? stderr : stdout, Usage, program_invocation_short_name);
  exit(exitCode);
}

Log Logger;
char* ModuleDirectory = NULL;

void CleanUp() {
  free(ModuleDirectory);
  LogClose(Logger);
  LogDispose(Logger);
  Logger = NULL;
}

char* GetExecutableDirectory() {
  int retVal;
  char* lastSlash;
  size_t resultLength;
  char* result;
  char* linkTarget = malloc(sizeof(char) * 1024);
  if (linkTarget == NULL) {
    fprintf(stderr, "%s: malloc failed on linkTarget.\n", program_invocation_short_name);
    exit(M_MemErr);
  }

  retVal = readlink("/proc/self/exe", linkTarget, 1024);
  if (retVal == -1) {
    fprintf(stderr, "%s: error reading executable directory form /proc/self/exe.\n",
	    program_invocation_short_name);
    exit(M_ProcReadErr);
  }
  
  linkTarget[retVal] = '\0';
  lastSlash = strrchr(linkTarget, '/');
  if (lastSlash == NULL || lastSlash == linkTarget) {
    fprintf(stderr, "%s: error parsing executable directory from /proc/self/exe.\n", 
	    program_invocation_short_name);
    exit(M_ProcReadErr);
  }

  resultLength = lastSlash - linkTarget;
  result = (char *) malloc(resultLength + 1);
  if (result == NULL) {
    fprintf(stderr, "%s: malloc failed on result.\n", program_invocation_short_name);
    exit(M_MemErr);
  }
  strncpy(result, linkTarget, resultLength);
  result[resultLength] = '\0';
  free(linkTarget);
  return result;
}  

int main(int argc, char* const argv[]) {
  int nextOption, k;
  LcseAddress address = NULL, current;
  void* save;
  char *logFile, *hostName, *shortHostName, *firstDot;

  ModuleDirectory = GetExecutableDirectory();

  hostName = malloc(sizeof(char) * 100);
  if (hostName == NULL) {
    fprintf(stderr, "%s: malloc failed on hostName.\n", program_invocation_short_name);
    free(ModuleDirectory);
    exit(M_MemErr);
  }

  if (gethostname(hostName, 100) < 0) {
    int err = errno;
    fprintf(stderr, "%s: gethostname failed.  %s.\n", program_invocation_short_name, strerror(err));
    free(hostName);
    free(ModuleDirectory);
    exit(M_HostNameErr);
  }
  hostName[99] = '\0';
  if ((firstDot = strchr(hostName, '.')) != NULL) {
    int length = firstDot - hostName;
    shortHostName = malloc(length + 1);
    if (shortHostName == NULL) {
      fprintf(stderr, "%s: malloc failed on shortHostName.\n", program_invocation_short_name);
      free(hostName);
      free(ModuleDirectory);
      exit(M_MemErr);
    }
    strncpy(shortHostName, hostName, length);
    shortHostName[length] = '\0';
    free(hostName);
  }
  else
    shortHostName = hostName;

  logFile = malloc(sizeof(char) * (strlen(ModuleDirectory) + strlen(shortHostName) +
				   strlen(program_invocation_short_name) + 7));
  if (logFile == NULL) {
    fprintf(stderr, "%s: malloc failed on logFile.\n", program_invocation_short_name);
    free(shortHostName);
    free(ModuleDirectory);
    exit(M_MemErr);
  }
  
  sprintf(logFile, "%s/%s_%s.log", ModuleDirectory, program_invocation_short_name, shortHostName);
  Logger = LogCreate(program_invocation_short_name);
  if (Logger == NULL) {
    fprintf(stderr, "%s: create failed on Logger.\n", program_invocation_short_name);
    free(logFile);
    free(shortHostName);
    free(ModuleDirectory);
    exit(M_MemErr);
  }
  if (LogOpen(Logger, logFile) < 0)
    fprintf(stderr, "%s: open failed on Logger.\n", program_invocation_short_name);
  
  free(shortHostName);
  free(logFile);

  do {
    nextOption = getopt_long(argc, argv, ShortOptions, LongOptions, NULL);
    switch(nextOption) {
    case 'a': {
      LcseAddress addr;
      char delim[] = ",";
      char delim2[] = ":";
      char *token, *token2, *arg, *arg2, *end, *savePtr, *savePtr2;
      struct hostent* hostname;
      int k, value;

      if ((arg = strndup(optarg, 4096)) == NULL) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "strndup failed on optarg");
	fprintf(stderr, "%s: strndup failed on optarg.\n", program_invocation_short_name);
	for (current = address; current != NULL;) {
	  addr = current->Next;
	  free(current);
	  current = addr;
	}
	LogClose(Logger);
	Logger = NULL;
	free(ModuleDirectory);
	exit(M_MemErr);
      }

      for (token = strtok_r(arg, delim, &savePtr); 
	   token != NULL; 
	   token = strtok_r(NULL, delim, &savePtr)) {
	if ((addr = malloc(sizeof(struct LcseAddressRecord))) == NULL) {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "malloc failed on addr");
	  fprintf(stderr, "%s: malloc failed on addr.\n", program_invocation_short_name);
	  for (current = address; current != NULL;) {
	    addr = current->Next;
	    free(current);
	    current = addr;
	  }
	  free(arg);
	  LogClose(Logger);
	  Logger = NULL;
	  free(ModuleDirectory);
	  exit(M_MemErr);
	}
	addr->Address.s_addr = INADDR_ANY;
	addr->Port = htons(9005);
	addr->Domain = 2;
	addr->Next = NULL;

	if ((arg2 = strndup(token, 4096)) == NULL) {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "strndup failed on optarg");
	  fprintf(stderr, "%s: strndup failed on optarg.\n", program_invocation_short_name);
	  for (current = address; current != NULL;) {
	    addr = current->Next;
	    free(current);
	    current = addr;
	  }
	  free(arg);
	  LogClose(Logger);
	  Logger = NULL;
	  free(ModuleDirectory);
	  exit(M_MemErr);
	}

	for (token2 = strtok_r(arg2, delim2, &savePtr2), k = 0; 
	     token2 != NULL; 
	     token2 = strtok_r(NULL, delim2, &savePtr2), k++) {
	  switch(k) {
	  case 0:
	    if (strcasecmp(token2, "inaddr_any")) {
	      if (inet_aton(token2, &addr->Address) == 0) {
		if ((hostname = gethostbyname(token2)) == NULL) {
		  if (LogIsFatal(Logger))
		    LogMessage(Logger, LogFatal, "%s lookup error:  %s.", token, hstrerror(h_errno));
		  fprintf(stderr, "%s: %s lookup error:  %s.", program_invocation_short_name, 
			  token, hstrerror(h_errno));
		  free(arg2);
		  free(addr);
		  for (current = address; current != NULL;) {
		    addr = current->Next;
		    free(current);
		    current = addr;
		  }
		  free(arg);
		  CleanUp();
		  exit(M_ArgInvalid);
		}
		memcpy(&addr->Address, hostname->h_addr, sizeof(struct in_addr));
	      }
	    }
	    break;
	  case 1:
	    value = strtol(token2, &end, 10);
	    if (*end != '\0') {
	      if (LogIsFatal(Logger))
		LogMessage(Logger, LogFatal, "Invalid port sequence %s in %s.", token2, token);
	      free(arg2);
	      free(addr);
	      for (current = address; current != NULL;) {
		addr = current->Next;
		free(current);
		current = addr;
	      }
	      free(arg);
	      CleanUp();
	      fprintf(stderr, "%s: Invalid port sequence %s in %s.\n\n", program_invocation_short_name, token2, token);
	      PrintUsage(M_ArgInvalid);
	    }
	    else if (value < 1024 || value > 65536) {
	      if (LogIsFatal(Logger))
		LogMessage(Logger, LogFatal, "Port %ld in %s is outside the valid range.", value, token);
	      free(arg2);
	      free(addr);
	      for (current = address; current != NULL;) {
		addr = current->Next;
		free(current);
		current = addr;
	      }
	      free(arg);
	      CleanUp();
	      fprintf(stderr, "%s: Port %d in %s is outside the valid range.\n\n", 
		      program_invocation_short_name, value, token);
	      PrintUsage(M_ArgInvalid);
	    }
	    addr->Port = (uint16_t)htons(value);
	    break;
	  case 2:
	    addr->Domain = strtol(token2, &end, 10);
	    if (*end != '\0') {
	      if (LogIsFatal(Logger))
		LogMessage(Logger, LogFatal, "Invalid domain number sequence %s in %s.", token2, token);
	      free(arg2);
	      free(addr);
	      for (current = address; current != NULL;) {
		addr = current->Next;
		free(current);
		current = addr;
	      }
	      free(arg);
	      CleanUp();
	      fprintf(stderr, "%s: Invalid domain number sequence %s in %s.\n\n", program_invocation_short_name, token2, token);
	      PrintUsage(M_ArgInvalid);
	    }
	    break;
	  }
	}
	for (current = address; current != NULL && current->Next != NULL; current = current->Next);
	if (current == NULL)
	  address = addr;
	else
	  current->Next = addr;
	free(arg2);
      }
      free(arg);
    }
      break;
    case 'h':
      for (current = address; current != NULL;) {
	save = current->Next;
	free(current);
	current = save;
      }
      CleanUp();
      PrintUsage(0);
      break;
    case 'm': {
      struct stat directoryInfo;

      if (access(optarg, F_OK) != 0) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "Module directory %s does not exist.", optarg);
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
	CleanUp();
	fprintf(stderr, "%s: Module directory %s does not exist.\n\n", 
		program_invocation_short_name, optarg);
	PrintUsage(M_ArgInvalid);
      }
      else if (access(optarg, R_OK | X_OK) != 0) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "Module directory %s is not accessible.", optarg);
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
	CleanUp();
	fprintf(stderr, "%s: Module directory %s is not accessible.\n\n", 
		program_invocation_short_name, optarg);
	PrintUsage(M_ArgInvalid);
      }
      else if (stat(optarg, &directoryInfo) != 0 || !S_ISDIR(directoryInfo.st_mode)) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "Module directory %s is not a directory.", optarg);
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
	CleanUp();
	fprintf(stderr, "%s: Module directory %s is not a directory.\n\n", 
		program_invocation_short_name, optarg);
	PrintUsage(M_ArgInvalid);
      }
      else {
	free(ModuleDirectory);
	ModuleDirectory = strdup(optarg);
      }
      break;
    }
    case 'd':
      LogSetLevel(Logger, LogDebug);
      break;
    case 'v':
      LogSetLevel(Logger, LogInfo);
      break;
    case 's': {
      off_t value;
      char* end;

      if ((value = strtoll(optarg, &end, 10)) < 1) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "size must be positive.");
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
	CleanUp();
	fprintf(stderr, "%s: size must be positive.\n", program_invocation_short_name);
	PrintUsage(M_ArgInvalid);
      }
      if (*end != '\0') {
	if (strcasecmp(end, "GB") == 0 || strcasecmp(end, "G") == 0)
	  value *= 1073741824;
	else if (strcasecmp(end, "MB") == 0 || strcasecmp(end, "M") == 0)
	  value *= 1048576;
	else if (strcasecmp(end, "KB") == 0 || strcasecmp(end, "K") == 0)
	  value *= 1024;
	else if (strcasecmp(end, "B") == 0);
	else {
	  if (LogIsFatal(Logger))
	    LogMessage(Logger, LogFatal, "size unit %s not known.", end);
	  for (current = address; current != NULL;) {
	    save = current->Next;
	    free(current);
	    current = save;
	  }
	  CleanUp();
	  fprintf(stderr, "%s: size unit %s not known.\n", program_invocation_short_name, end);
	  PrintUsage(M_ArgInvalid);
	}
      }
      LogSetMaxSize(Logger, value);
    }
      break;
    case 'l': {
      long value;
      char* end;
      
      value = strtol(optarg, &end, 10);
      if (*end != '\0') {
	if (strcasecmp(optarg, "off") == 0)
	  value = LogOff;
	else if (strcasecmp(optarg, "fatal") == 0)
	  value = LogFatal;
	else if (strcasecmp(optarg, "error") == 0)
	  value = LogError;
	else if (strcasecmp(optarg, "warn") == 0)
	  value = LogWarn;
	else if (strcasecmp(optarg, "info") == 0)
	  value = LogInfo;
	else if (strcasecmp(optarg, "debug") == 0)
	  value = LogDebug;
	else if (strcasecmp(optarg, "all") == 0)
	  value = LogAll;
	else
	  value = -1;
      }
      if (value == -1) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "LogLevel %s is not valid.", optarg);
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
        CleanUp();
	fprintf(stderr, "%s: LogLevel %s is not valid.\n\n", 
		program_invocation_short_name, optarg);
	PrintUsage(M_ArgInvalid);
      }
      else if (value < LogOff || value > LogAll) {
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogFatal, "Loglevel %ld is outside the valid range.", value);
	for (current = address; current != NULL;) {
	  save = current->Next;
	  free(current);
	  current = save;
	}
	CleanUp();
	fprintf(stderr, "%s: LogLevel %ld is outside the valid range.\n\n", 
		program_invocation_short_name, value);
	PrintUsage(M_ArgInvalid);
      }
      else
	LogSetLevel(Logger, (LogLevel)value);
      break;
    }
    case '?':
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "Invalid option specified.  aborting.");
      for (current = address; current != NULL;) {
	save = current->Next;
	free(current);
	current = save;
      }
      CleanUp();
      fprintf(stderr, "%s: Invalid option specified.\n\n", program_invocation_short_name);
      PrintUsage(M_ArgUnknown);
      break;
    case -1:
      break;
    }
  } while (nextOption != -1);

  if (!address) {
    if ((address = malloc(sizeof(struct LcseAddressRecord))) == NULL) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "malloc failed on address");
      fprintf(stderr, "%s: malloc failed on address.\n", program_invocation_short_name);
      LogClose(Logger);
      Logger = NULL;
      free(ModuleDirectory);
      exit(M_MemErr);
    }
    address->Address.s_addr = INADDR_ANY;
    address->Port = htonl(9005);
    address->Domain = 2;
    address->Next = NULL;
  }

  k = ServerRun(address);

  for (current = address; current != NULL;) {
    save = current->Next;
    free(current);
    current = save;
  }
  CleanUp();

  if (k) {
    switch(k) {
    case 1: fprintf(stderr, "%s: memory error.\n\n", program_invocation_short_name); break;
    case 2: ;
    case 10: fprintf(stderr, "%s: socket error.\n\n", program_invocation_short_name); break;
    case 3: fprintf(stderr, "%s: signal error.\n\n", program_invocation_short_name); break;
    case 4: ;
    case 11: fprintf(stderr, "%s: non-blocking error.\n\n", program_invocation_short_name); break;
    case 5: ;
    case 12:  fprintf(stderr, "%s: close on exec error.\n\n", program_invocation_short_name); break;
    case 6: ;
    case 13: fprintf(stderr, "%s: port reuse error.\n\n", program_invocation_short_name); break;
    case 7: ; 
    case 14: fprintf(stderr, "%s: defer accept error.\n\n", program_invocation_short_name); break;
    case 8: fprintf(stderr, "%s: bind error.\n\n", program_invocation_short_name); break;
    case 9: fprintf(stderr, "%s: listen error.\n\n", program_invocation_short_name); break;
    }
    if (k < 3) return k;
    return 8 + k - 2;
  }
  return 0;
}

