/******************************************************************************\
 * File: server.c                                                             *
 * Author: James Greensky                                                     *
 * Notes: Implement server routines                                           *
\******************************************************************************/

#define _GNU_SOURCE

#include "server.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <dlfcn.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <limits.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

enum ServerReturnCodes {
  S_OK = 0,
  S_MemErr = 1,
  S_SockErr = 2,
  S_SigErr = 3,
  S_NBlockErr = 4,
  S_CloseExecErr = 5,
  S_ReuseErr = 6,
  S_DeferAccErr = 7,
  S_BindErr = 8,
  S_ListenErr = 9,
  S_IbSockErr = 10,
  S_IbNBlockErr = 11,
  S_IbCloseExecErr = 12,
  S_IbReuseErr = 13,
  S_IbDeferAccErr = 14,
  S_IbBindErr = 15,
  S_IbListenErr = 16,
  S_AddrErr = 17,
  S_PortErr = 18
};

typedef enum ServerResponses {
  SR_OK = 0,
  SR_SYMNFND = -1,
  SR_LIBNFND = -2,
  SR_ILLFORM = -3,
  SR_MEMFAIL = -4,
  SR_IOFAIL = -5
} ServerResponse;

typedef struct ServerExecutableRecord {
  LcseExecutable Executable;
  struct ServerExecutableRecord* Next;
}* ServerExecutable;

typedef struct ServerFDRecord {
  int FD;
  struct ServerFDRecord* Next;
}* ServerFD;

typedef struct ServerModuleRecord {
  void* Handle;
  char* Name;
  unsigned int Count;
  void (*Initialize) (void* handle);
  struct timeval* (*ClientHandler) (const LcseClient client, LcseClientEvent event);
  struct timeval* (*ExecutableHandler) (const LcseExecutable executable, LcseExecutableEvent event);
  struct timeval* (*IOHandler) (int fd, LcseIOEvent event); 
  struct timeval* (*TimeoutHandler) ();
  void (*Dispose) ();
  ServerExecutable Executables;
  ServerFD ReadFDs;
  ServerFD WriteFDs;
  struct ServerModuleRecord* Next;
}* ServerModule;

typedef struct ServerClientRecord {
  LcseClient Client;
  ServerModule Module;
  int Active;
  struct ServerClientRecord* Next;
}* ServerClient;

typedef struct ServerSocketRecord {
  int Socket;
  struct sockaddr_in Address;
  struct ServerSocketRecord* Next;
}* ServerSocket;

fd_set MasterRead, MasterWrite;
int FDMax = -1, ServerExit = 0;
ServerClient ServerClients = NULL;
ServerModule ServerModules = NULL;
ServerExecutable ServerExecutables = NULL;
ServerSocket ServerSockets = NULL;
char* DiscardBuffer = NULL;

static void SetFDMax() { 
  ServerClient client = NULL;
  ServerModule module = NULL;
  ServerExecutable executable = NULL;
  ServerSocket serverSocket = NULL;
  ServerFD fd = NULL;

  for (FDMax = -1, serverSocket = ServerSockets;
       serverSocket != NULL; serverSocket = serverSocket->Next) {
    if (serverSocket->Socket > FDMax)
      FDMax = serverSocket->Socket;
  }
  for (client = ServerClients; client != NULL; client = client->Next) {
    if (client->Client->Socket > FDMax)
      FDMax = client->Client->Socket;
  }

  for (module = ServerModules; module != NULL; module = module->Next) {
    for (executable = module->Executables; executable != NULL; executable = executable->Next) {
      if (executable->Executable->StdOut > FDMax)
	FDMax = executable->Executable->StdOut;
      if (FD_ISSET(executable->Executable->StdIn, &MasterWrite)) {
	if (executable->Executable->StdIn > FDMax)
	  FDMax = executable->Executable->StdIn;
      }
    }
    for (fd = module->ReadFDs; fd != NULL; fd = fd->Next) {
      if (fd->FD > FDMax)
	FDMax = fd->FD;
    }
    for (fd = module->WriteFDs; fd != NULL; fd = fd->Next) {
      if (fd->FD > FDMax)
	FDMax = fd->FD;
    }
  }
}

static void RemoveModule(ServerModule module) {
  assert(module != NULL);
  
  void* save;
  ServerExecutable executable = NULL;
  ServerModule current = ServerModules;
  ServerFD fd = NULL;
  int maxFD = -1;

  if (current == module)
    ServerModules = current->Next;
  else {
    for (; current != NULL && current->Next != module; module = module->Next);
    current->Next = module->Next;
  }

  (*module->Dispose)();

  for (executable = module->Executables; executable != NULL;) {
    save = executable->Next;
    executable->Next = ServerExecutables;
    ServerExecutables = executable;
    module->Executables = save;
    kill(executable->Executable->ProcessID, SIGTERM);
    executable = save;
  }
  
  for (fd = module->ReadFDs; fd != NULL;) {
    save = fd->Next;
    if (fd->FD > maxFD)
      maxFD = fd->FD;
    FD_CLR(fd->FD, &MasterRead);
    free(fd);
    fd = save;
  }
  for (fd = module->WriteFDs; fd != NULL;) {
    save = fd->Next;
    if (fd->FD > maxFD)
      maxFD = fd->FD;
    FD_CLR(fd->FD, &MasterWrite);
    free(fd);
    fd = save;
  }
  if (maxFD == FDMax)
    SetFDMax();

  if (dlclose(module->Handle) != 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "dlclose failed for %s/%s.  %s.", module->Name, 
		 ModuleDirectory, dlerror());
  }

  free(module->Name);
  free(module);
} 

static void RemoveClient(ServerClient client) {
  assert(client != NULL);

  ServerClient current = ServerClients;
  if (current == client)
    ServerClients = current->Next;
  else {
    for (;current != NULL && current->Next != client; current = current->Next);
    current->Next = client->Next;
  }

  if (client->Module != NULL) {
    (*client->Module->ClientHandler)(client->Client, ClientDisconnected);
    if (!--client->Module->Count)
      RemoveModule(client->Module);
  }

  CBufferDispose(client->Client->InBuffer);
  CBufferDispose(client->Client->OutBuffer);
  
  FD_CLR(client->Client->Socket, &MasterWrite);
  FD_CLR(client->Client->Socket, &MasterRead);
  if (client->Client->Socket == FDMax)
    SetFDMax();
  
  Cclose(client->Client->Socket);

  free(client->Client);
  free(client);
}

static int ShutdownClient(ServerClient client) {
  assert(client != NULL);

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Shutting down client.");

  if (CBufferSize(client->Client->OutBuffer))
    client->Active = 0;
  else if (shutdown(client->Client->Socket, SHUT_WR) < 0) {
    int err = errno;
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "shutdown failed on client.  %s.", strerror(err));
    RemoveClient(client);
    return -1;
  }
  return 0;
}
  

static int ClientSend(ServerClient client) {
  assert(client != NULL);

  int retVal = 1, err = EINTR, count = 0, flags = 0, total = 0;
  char* buffer = NULL;

  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "MSG_NOSIGNAL = %d, MSG_MORE = %d.", MSG_NOSIGNAL, MSG_MORE);

  while (CBufferSize(client->Client->OutBuffer) && retVal > -1) {
    count = 0;
    buffer = CBufferGetReadBlock(client->Client->OutBuffer, &count);
    flags = MSG_NOSIGNAL | ((CBufferSize(client->Client->OutBuffer) - count > 0 || 
			     !client->Active) ? MSG_MORE : 0);
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "bufferSize = %d, count = %d, client->Active = %d, flags = %d.", 
	CBufferSize(client->Client->OutBuffer), count, client->Active, flags);
    retVal = send(client->Client->Socket, buffer, count, flags);
    if (retVal < 0) {
      err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else {
      if (LogIsDebug(Logger)) {
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(client->Client->Socket, (struct sockaddr *) &addr, &len) != -1)
	  LogMessage(Logger, LogDebug, "Sent %d bytes to %s:%u.", retVal,
		     inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	else
	  LogMessage(Logger, LogDebug, "Sent %d bytes.", retVal);
      }
      CBufferDecommitBlock(client->Client->OutBuffer, retVal);
      total += retVal;
    }
  }

  if (LogIsDebug(Logger)) {
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);
    if (getpeername(client->Client->Socket, (struct sockaddr *) &addr, &len) != -1)
      LogMessage(Logger, LogDebug, "Sent %d bytes total to (%s:%u).", total,
		 inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
    else
      LogMessage(Logger, LogDebug, "Sent %d bytes total.", total);
  }

  if (retVal < 0) {
    if (err != EAGAIN) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "send failed to client.  %s.", strerror(err));
      RemoveClient(client);
      return -1;
    }
  }
  if (!client->Active)
    return ShutdownClient(client);

  return 0;
}

static void AddModule(ServerClient client, char* moduleName) {
  assert(client != NULL);
  assert(moduleName != NULL);

  ServerModule module = NULL;
  for (module = ServerModules; module != NULL && strcmp(module->Name, moduleName) != 0; 
       module = module->Next);

  if (module == NULL) {
    void* handle = NULL;
    char* modulePath = malloc(sizeof(char) * (strlen(ModuleDirectory) + strlen(moduleName) + 5));
    if (modulePath == NULL) {
      int response = htonl(SR_MEMFAIL);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed for modulePath.");
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      return;
    }
    sprintf(modulePath, "%s/%s", ModuleDirectory, moduleName);

    handle = dlopen(modulePath, RTLD_NOW);
    if (handle == NULL) {
      int response = htonl(SR_LIBNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlopen failed for %s.  %s.", modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      free(modulePath);
      ShutdownClient(client);
      return;
    }

    free(modulePath);
    module = malloc(sizeof(struct ServerModuleRecord));
    if (module == NULL) {
      int response = htonl(SR_MEMFAIL);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed for modulePath.");
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      return;
    }
    
    module->Initialize = dlsym(handle, "Initialize");
    if (module->Initialize == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return Initialize from %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->ClientHandler = dlsym(handle, "ClientHandler");
    if (module->ClientHandler == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return  ClientHandler from %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->ExecutableHandler = dlsym(handle, "ExecutableHandler");
    if (module->ExecutableHandler == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return ExecutableHandler from %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->IOHandler = dlsym(handle, "IOHandler");
    if (module->IOHandler == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return IOHandler form %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->TimeoutHandler = dlsym(handle, "TimeoutHandler");
    if (module->TimeoutHandler == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return TimeoutHandler from %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->Dispose = dlsym(handle, "Dispose");
    if (module->Dispose == NULL) {
      int response = htonl(SR_SYMNFND);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "dlsym failed to return Dispose from %s.  %s.",
		   modulePath, dlerror());
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->Handle = handle;
    module->Name = strdup(moduleName);
    if (module->Name == NULL) {
      int response = htonl(SR_MEMFAIL);
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "strdup failed on module->Neme.");
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
      free(module);
      return;
    }

    module->Count = 0;
    module->Executables = NULL;
    module->ReadFDs = NULL;
    module->WriteFDs = NULL;
    module->Next = ServerModules;
    ServerModules = module;

    (*module->Initialize)(handle);
  }
  client->Module = module;
  module->Count++;
  
  (*module->ClientHandler)(client->Client, ClientAccepted);
}

static void ClientReceive(ServerClient client) {
  assert(client != NULL);

  int retVal = 1, err = EINTR, count = 0, total = 0;
  char* buffer = NULL;

  while (retVal > 0) {
    count = 0;
    if (client->Active) {
      buffer = CBufferGetWriteBlock(client->Client->InBuffer, &count);
      if (buffer == NULL) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "getwriteblock failed on in buffer.");
	if (ShutdownClient(client) < 0)
	  return;
	buffer = DiscardBuffer;
	count = 128;
      }
    }
    else {
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "Receiving to DiscardBuffer.");
      buffer = DiscardBuffer;
      count = 128;
    }
    retVal = recv(client->Client->Socket, buffer, count, 0);
    if (retVal > 0) {
      total += retVal;
      if (LogIsDebug(Logger)) {
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	if (getpeername(client->Client->Socket, (struct sockaddr *) &addr, &len) != -1)
	  LogMessage(Logger, LogDebug, "Received %d bytes from %s:%u.", retVal,
		     inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
	else
	  LogMessage(Logger, LogDebug, "Received %d bytes.", retVal);
      }
    }
    if (retVal < 0) {
      err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else if (buffer != DiscardBuffer)
      CBufferCommitBlock(client->Client->InBuffer, retVal); 
  }

  if (LogIsDebug(Logger)) {
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);

    /*  char* buffer = malloc(CBufferSize(client->Client->InBuffer));
      char* strbuffer = malloc((CBufferSize(client->Client->InBuffer) * 4) + 
      (((CBufferSize(client->Client->InBuffer) / 16) + 1) * 8) + 13); 
      int m, n; */


    if (getpeername(client->Client->Socket, (struct sockaddr *) &addr, &len) != -1)
      LogMessage(Logger, LogDebug, "Received %d total bytes from %s:%u", total,
		 inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
    else
      LogMessage(Logger, LogDebug, "Received %d total bytes.", total);

    /*CBufferPeek(client->Client->InBuffer, buffer, CBufferSize(client->Client->InBuffer));
    sprintf(strbuffer, "InBuffer =");
    for (m = 0, n = 10; m < CBufferSize(client->Client->InBuffer); m++) {
      if (m == 0 || ((m % 16) == 0)) {
	sprintf(&strbuffer[n], "\n%07d", m);
	n+=8;
      }
      sprintf(&strbuffer[n], "%4u,", (unsigned char)buffer[m]);
      n+=4;
    }
    free(buffer);
    sprintf(&strbuffer[n], ".");
    strbuffer[++n] = '\0';
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "%s", strbuffer);
      free(strbuffer);*/
  }


  if (retVal == 0) {
    if (LogIsInfo(Logger))
      LogMessage(Logger, LogInfo, "client connection closed.");
    RemoveClient(client);
  }
  else if (err != EAGAIN) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "recv failed from client.  %s.", strerror(err));
    RemoveClient(client);
  }
  else if (client->Module == NULL) {
    char* ptr = NULL;
    count = CBufferSize(client->Client->InBuffer);
    if (count > 256) count = 256;
    buffer = CBufferGetReadBlock(client->Client->InBuffer, &count);
    ptr = memrchr(buffer, '\0', count);
    if (ptr != NULL) {
      AddModule(client, buffer);
      CBufferDecommitBlock(client->Client->InBuffer, ptr - buffer + 1);
    }
    else if (count == 256) {
      int response = htonl(SR_ILLFORM);
      if (CBufferWrite(client->Client->OutBuffer, &response, 4) < 0) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "write failed on out buffer.");
      }
      ShutdownClient(client);
    }
  }
  else if (CBufferSize(client->Client->InBuffer))
    (*client->Module->ClientHandler)(client->Client, ClientDataReceived);
}

static void SocketSend(int socket, void* buffer, int length) {
  assert(socket > -1);

  int total = 0, retVal = 1;
  while (total < length && retVal > -1) {
    retVal = send(socket, buffer, length - total, MSG_NOSIGNAL);
    if (retVal < 0) {
      int err = errno;
      if (err == EINTR)
	retVal = 1;
    }
    else
      total += retVal;
  }
}

static void AddClient(int clientSocket) {
  assert(clientSocket > -1);
  
  int response = 0;
  ServerClient client = malloc(sizeof(struct ServerClientRecord));
  if (client == NULL) {
    response = htonl(SR_MEMFAIL);
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on client.");
    SocketSend(clientSocket, &response, 4);
    Cclose(clientSocket);
    return;
  }
     
  client->Client = malloc(sizeof(struct LcseClientRecord));
  if (client->Client == NULL) {
    response = htonl(SR_MEMFAIL);
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on client->Client.");
    SocketSend(clientSocket, &response, 4);
    free(client);
    Cclose(clientSocket);
    return;
  }

  client->Client->InBuffer = CBufferCreate(256, sizeof(char));
  if (client->Client->InBuffer == NULL) {
    response = htonl(SR_MEMFAIL);
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create in buffer failed.");
    SocketSend(clientSocket, &response, 4);
    free(client->Client);
    free(client);
    Cclose(clientSocket);
    return;
  }
  client->Client->OutBuffer = CBufferCreate(256, sizeof(char));
  if (client->Client->OutBuffer == NULL) {
    response = htonl(SR_MEMFAIL);
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "create out buffer failed.");
    SocketSend(clientSocket, &response, 4);
    CBufferDispose(client->Client->InBuffer);
    free(client->Client);
    free(client);
    Cclose(clientSocket);
    return;
  }
  
  client->Client->Socket = clientSocket;
  client->Module = NULL;
  client->Active = 1;
  client->Next = ServerClients;
  ServerClients = client;
  FD_SET(clientSocket, &MasterRead);
  if (clientSocket > FDMax)
    FDMax = clientSocket;
  
  ClientReceive(client);
}

static void ServerAccept(int serverSocket) {
  struct sockaddr_in clientAddress;
  socklen_t addressLength = 0;
  int retVal = 1, on = 1, size = BUFFER_SIZE;

  while (retVal > -1) {
    addressLength = sizeof(clientAddress);
    retVal = accept(serverSocket, (struct sockaddr*)&clientAddress,
		    &addressLength);
    
    if (retVal > -1) {
      if (LogIsInfo(Logger))
	LogMessage(Logger, LogInfo, "Connection accepted from %s:%u.", 
		   inet_ntoa(clientAddress.sin_addr), (int)ntohs(clientAddress.sin_port));
      
      if (SetNonBlockingIO(retVal) < 0) {
	int response = htonl(SR_IOFAIL);
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "Set non-blocking failed for client.  Aborting connection.");
	SocketSend(retVal, &response, 4);
	Cclose(retVal);
	return;
      }
      if (SetCloseOnExecIO(retVal) < 0) {
	int response = htonl(SR_IOFAIL);
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "Set close-on-exec failed for client.  Aborting connection.");
	SocketSend(retVal, &response, 4);
	Cclose(retVal);
	return;
      }
      if (setsockopt(retVal, SOL_TCP, TCP_NODELAY, &on, sizeof(int)) < 0) {
	int response = htonl(SR_IOFAIL);
	int err = errno;
	if (LogIsFatal(Logger))
	  LogMessage(Logger, LogError, "setsockopt failed for client.  Aborting connection.  %s.", 
		     strerror(err));
	SocketSend(retVal, &response, 4);
	Cclose(retVal);
	return;
      }
      size = BUFFER_SIZE;
      if (setsockopt(retVal, SOL_SOCKET, SO_SNDBUF, &size, sizeof(int)) < 0) {
	int response = htonl(SR_IOFAIL);
	int err = errno;
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "setsockopt failed for client.  Aborting connection.  %s.",
		     strerror(err));
	SocketSend(retVal, &response, 4);
	Cclose(retVal);
	return;
      }
      size = BUFFER_SIZE;
      if (setsockopt(retVal, SOL_SOCKET, SO_RCVBUF, &size, sizeof(int)) < 0) {
	int response = htonl(SR_IOFAIL);
	int err = errno;
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "setsockopt failed for client. Aborting connection.  %s.",
		     strerror(err));
	SocketSend(retVal, &response, 4);
	Cclose(retVal);
	return;
      }
      AddClient(retVal);
    }
    else {
      int err = errno;
      if (err == EINTR)
	retVal = 1;
      else {
	if (err != EAGAIN) {
	  if (LogIsError(Logger))
	    LogMessage(Logger, LogError, "accept failed on serverSocket.  %s.", strerror(err));
	}
      }
    }
  }
}

static int Listen(LcseAddress localAddress) {
  int on = 1, sock;
  LcseAddress current;
  ServerSocket currentSocket, serverSocket;
  struct sockaddr_in socketAddress;

  for (current = localAddress; current != NULL; current = current->Next) {
    if ((sock = socket(current->Domain, SOCK_STREAM, 0))== -1) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "Could not create server listening socket.  %s.", strerror(errno));
      return S_SockErr;
    }

    if (SetNonBlockingIO(sock) < 0) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "SetNonBlockingIO failed on sock.");
      Cclose(sock);
      return S_NBlockErr;
    }
    
    if (SetCloseOnExecIO(sock) < 0) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "SetCloseOnExecIO failed on sock.");
      Cclose(sock);
      return S_CloseExecErr;
    }
    
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int)) != 0) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "setsockopt failed on sock.  %s.", 
		   strerror(errno));
      Cclose(sock);
      return S_ReuseErr;
    }
    
    if (setsockopt(sock, SOL_TCP, TCP_DEFER_ACCEPT, &on, sizeof(int)) != 0) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "setsockopt failed on ServerSocket. %s.",
		   strerror(errno));
      Cclose(sock);
      return S_DeferAccErr;
    }

    if ((serverSocket = malloc(sizeof(struct ServerSocketRecord))) == NULL) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "malloc failed on serverSocket.");
      Cclose(sock);
      return S_MemErr;
    }
    
    memset(&serverSocket->Address, 0, sizeof(struct sockaddr_in));
    serverSocket->Socket = sock;
    serverSocket->Address.sin_family = current->Domain;
    serverSocket->Address.sin_port = current->Port;
    serverSocket->Address.sin_addr = current->Address;
    serverSocket->Next = NULL;
    if (bind(sock, (struct sockaddr*)&serverSocket->Address, sizeof(struct sockaddr_in))) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "Could not bind sock to %s:%d:%d.  %s.", inet_ntoa(current->Address), 
		   ntohs(current->Port), current->Domain, strerror(errno));
      free(serverSocket);
      Cclose(sock);
      return S_BindErr;
    }
    
    if (listen(sock, SOMAXCONN)) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "Could not listen on sock.  %s.", strerror(errno));
      free(serverSocket);
      Cclose(sock);
      return S_ListenErr;
    }

    for (currentSocket = ServerSockets; 
	 currentSocket != NULL && currentSocket->Next != NULL; 
	 currentSocket = currentSocket->Next);
    if (currentSocket == NULL)
      ServerSockets = serverSocket;
    else
      currentSocket->Next = serverSocket;
    
    if (LogIsInfo(Logger)) {
      socklen_t addressLength = sizeof(socketAddress);
      if (getsockname(sock, (struct sockaddr*)&socketAddress, &addressLength) == 0)
	LogMessage(Logger, LogInfo, "Server listening on %s:%d:%d.", inet_ntoa(socketAddress.sin_addr),
		   ntohs(socketAddress.sin_port), current->Domain);
      else
	LogMessage(Logger, LogError, "getsockname failed on sock.  %s.", strerror(errno));
    }
  }
  
  return 0;
}

static void DisposeExecutable(ServerExecutable executable) {
  assert(executable != NULL);

  int k = 0;

  FD_CLR(executable->Executable->StdIn, &MasterWrite);
  FD_CLR(executable->Executable->StdOut, &MasterRead);
  if (executable->Executable->StdOut == FDMax)
    SetFDMax();

  Cclose(executable->Executable->StdOut);
  Cclose(executable->Executable->StdIn);

  for (k = 0; executable->Executable->Argv[k] != NULL; k++)
    free(executable->Executable->Argv[k]);
  free(executable->Executable->Argv);
  
  free(executable->Executable);
  free(executable);
}

inline static void CheckReceive(ServerClient client, fd_set* readFDs) {
  assert(client != NULL);
  
  if (FD_ISSET(client->Client->Socket, readFDs))
    ClientReceive(client);
}

inline static void CheckSend(ServerClient client) {
  assert(client != NULL);
  
  if (CBufferSize(client->Client->OutBuffer)) {
    if (ClientSend(client) > -1) {
      if (CBufferSize(client->Client->OutBuffer))
	FD_SET(client->Client->Socket, &MasterWrite);
      else
	FD_CLR(client->Client->Socket, &MasterWrite);
    }
  }
}

static void SIGTERM_Handler(int signal) {
  ServerExit = 1;
}

static void SIGCHLD_Handler(int signal) {
  /* Just want to wake up from select if a child signal occurs. */
}

int ServerRun(LcseAddress localAddress) {
  assert(localAddress != NULL);
  
  int retVal = 0;
  fd_set readFDs;
  fd_set writeFDs;
  ServerClient client = NULL;
  ServerModule module = NULL;
  ServerExecutable executable = NULL, previous = NULL;
  ServerSocket serverSocket = NULL;
  ServerFD fd = NULL;
  struct sigaction sa;
  void* savePtr = NULL;
  pid_t processID = -1;
  int status = 0, found = 0;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = &SIGTERM_Handler;
  while (sigaction(SIGTERM, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGTERM.");
      return S_SigErr;
    }
  }  
  while (sigaction(SIGINT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGINT.");
      return S_SigErr;
    }
  }  
  while (sigaction(SIGQUIT, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGQUIT.");
      return S_SigErr;
    }
  }  
  while (sigaction(SIGHUP, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGHUP.");
      return S_SigErr;
    }
  }
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_IGN;
  while (sigaction(SIGPIPE, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGPIPE.");
      return S_SigErr;
    }
  }
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = &SIGCHLD_Handler;
  while (sigaction(SIGCHLD, &sa, NULL) == -1) {
    if (errno != EINTR) {
      if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "sigaction failed for SIGCHLD.");
      return S_SigErr;
    }
  }

  if ((retVal = Listen(localAddress))) {
    for (serverSocket = ServerSockets; serverSocket != NULL;) {
      savePtr = serverSocket->Next;
      Cclose(serverSocket->Socket);
      free(serverSocket);
      serverSocket = savePtr;
    }
    return retVal;
  }

  FD_ZERO(&MasterWrite);
  FD_ZERO(&MasterRead);
  for (serverSocket = ServerSockets; serverSocket != NULL; serverSocket = serverSocket->Next) {
    FD_SET(serverSocket->Socket, &MasterRead);
    if (serverSocket->Socket > FDMax)
      FDMax = serverSocket->Socket;
  }

  if ((DiscardBuffer = malloc(sizeof(char) * 128)) == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on DiscardBuffer.\n");
    for (serverSocket = ServerSockets; serverSocket != NULL;) {
      savePtr = serverSocket->Next;
      Cclose(serverSocket->Socket);
      free(serverSocket);
      serverSocket = savePtr;
    }
    return S_MemErr;
  }

  for(;;) {

    if (ServerExit) {
      for (client = ServerClients; client != NULL;) {
	savePtr = client->Next;
	RemoveClient(client);
	client = savePtr;
      }
      for (executable = ServerExecutables; executable != NULL;) {
	savePtr = executable->Next;
	DisposeExecutable(executable);
	executable = savePtr;
      }
      break;
    }

    for (;;) {
      do {
	errno = 0;
	processID = waitpid(-1, &status, WNOHANG);
      } while (processID <=0 && errno == EINTR);

      if (processID <=0) 
	break;

      if (LogIsDebug(Logger)) {
	if (WIFEXITED(status))
	  LogMessage(Logger, LogDebug, "executable %d exited.  %d.", processID, WEXITSTATUS(status));
	else if (WIFSIGNALED(status))
	  LogMessage(Logger, LogDebug, "executable %d exited abnormally.  %d.", processID, 
		     WTERMSIG(status));
      }

      found = 0;
      for (module = ServerModules; module != NULL && !found; module = module->Next) {
	for (previous = NULL, executable = module->Executables; executable != NULL && !found;) {
	  savePtr = executable->Next;
	  if (executable->Executable->ProcessID == processID) {
	    if (previous == NULL)
	      module->Executables = savePtr;
	    else
	      previous->Next = savePtr;
	    
	    if (WIFEXITED(status))
	      executable->Executable->Status = WEXITSTATUS(status);
	    else if (WIFSIGNALED(status))
	      executable->Executable->Status = 128 + WTERMSIG(status);
	    else {
	      if (LogIsWarn(Logger))
		LogMessage(Logger, LogWarn, "executable %d exited unknown.  %d.", processID, status);
	      executable->Executable->Status = status;
	    }

	    (*module->ExecutableHandler)(executable->Executable, ExecutableExited);
	    DisposeExecutable(executable);
	    executable = NULL;
	    found = 1;
	  }
	  previous = executable;
	  executable = savePtr;
	}
      }
      for (previous = NULL, executable = ServerExecutables; executable != NULL && !found;) {
	savePtr = executable->Next;
	if (executable->Executable->ProcessID == processID) {
	  if (previous == NULL)
	    ServerExecutables = savePtr;
	  else
	    previous->Next = savePtr;
	  executable->Executable->Status = status;
	  DisposeExecutable(executable);
	  executable = NULL;
	  found = 1;
	}
	previous = executable;
	executable = savePtr;
      }
    }
    
    writeFDs = MasterWrite;
    readFDs = MasterRead;
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Waiting for more work.");
    retVal = select(FDMax + 1, &readFDs, &writeFDs, NULL, NULL);
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "Ok, more work.");
    if (retVal < 0) {
      int err = errno;
      if (err != EINTR) {
	if (LogIsFatal(Logger))
	LogMessage(Logger, LogFatal, "select failed.  %s.", strerror(err));
	ServerExit = 1;
      }
      else if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "select failed with EINTR.");
      continue;
    }
    else if (retVal == 0) {
      if (LogIsDebug(Logger))
	LogMessage(Logger, LogDebug, "select returned empty.");
      continue;
    }
    else {
      if (LogIsDebug(Logger)) {	
	struct sockaddr_in socketAddress;
	socklen_t addressLength = sizeof(socketAddress);

	for (serverSocket = ServerSockets; serverSocket != NULL; serverSocket = serverSocket->Next) {
	  if (FD_ISSET(serverSocket->Socket, &readFDs))
	    LogMessage(Logger, LogDebug, "Incoming client on %s:%d:%d.",
		       inet_ntoa(serverSocket->Address.sin_addr), ntohs(serverSocket->Address.sin_port),
		       serverSocket->Address.sin_family);
	}

	for (module = ServerModules; module != NULL; module = module->Next) {
	  for (executable = module->Executables; executable != NULL; executable = executable->Next) {
	    if (FD_ISSET(executable->Executable->StdOut, &readFDs))
	      LogMessage(Logger, LogDebug, "Executable %d stdout.", executable->Executable->ProcessID);
	    if (FD_ISSET(executable->Executable->StdIn, &writeFDs))
	      LogMessage(Logger, LogDebug, "Executable %d stdin.", executable->Executable->ProcessID);
	  }
	  for (fd = module->ReadFDs; fd != NULL; fd = fd->Next) {
	    if (FD_ISSET(fd->FD, &readFDs))
	      LogMessage(Logger, LogDebug, "Module %s readFD %d.", module->Name, fd);
	  }
	  for (fd = module->WriteFDs; fd != NULL; fd = fd->Next) {
	    if (FD_ISSET(fd->FD, &writeFDs))
	      LogMessage(Logger, LogDebug, "Module %s writeFD %d.", module->Name, fd);
	  }
	}

	for (executable = ServerExecutables;  executable != NULL; executable = executable->Next) {
	  if (FD_ISSET(executable->Executable->StdOut, &readFDs))
	    LogMessage(Logger, LogDebug, "ServerExecutable %d stdout.", 
		       executable->Executable->ProcessID);
	  if (FD_ISSET(executable->Executable->StdIn, &writeFDs))
	    LogMessage(Logger, LogDebug, "ServerExecutable %d stdin.", 
		       executable->Executable->ProcessID);
	}
	
	for (client = ServerClients; client != NULL; client = client->Next) {
	  if (FD_ISSET(client->Client->Socket, &readFDs)) {
	    addressLength = sizeof(struct sockaddr_in);
	    retVal = getpeername(client->Client->Socket, (struct sockaddr*)&socketAddress, 
				 &addressLength);
	    if (retVal == 0)
	      LogMessage(Logger, LogDebug, "Client %s:%u read.", inet_ntoa(socketAddress.sin_addr),
			 ntohs(socketAddress.sin_port));
	    else
	      LogMessage(Logger, LogDebug, "Client %d read.", client->Client->Socket);
	  }
	  if (FD_ISSET(client->Client->Socket, &writeFDs)) {
	    addressLength = sizeof(struct sockaddr_in);
	    retVal = getpeername(client->Client->Socket, (struct sockaddr*)&socketAddress, 
				 &addressLength);
	    if (retVal == 0)
	      LogMessage(Logger, LogInfo, "Client %s:%u write.", inet_ntoa(socketAddress.sin_addr),
			 ntohs(socketAddress.sin_port));
	    else
	      LogMessage(Logger, LogInfo, "Client %d write.", client->Client->Socket);
	  }
	}
      }
      
      for (serverSocket = ServerSockets; serverSocket != NULL; serverSocket = serverSocket->Next) {
	if (FD_ISSET(serverSocket->Socket, &readFDs))
	  ServerAccept(serverSocket->Socket);
      }

      for (module = ServerModules; module != NULL; module = module->Next) {
	for (executable = module->Executables; executable != NULL; executable = executable->Next) {
	  if (FD_ISSET(executable->Executable->StdOut, &readFDs))
	    (*module->ExecutableHandler)(executable->Executable, ExecutableDataReady);
	  if (FD_ISSET(executable->Executable->StdIn, &writeFDs))
	    (*module->ExecutableHandler)(executable->Executable, ExecutableWriteReady);
	}
	for (fd = module->ReadFDs; fd != NULL;) {
	  savePtr = fd->Next;
	  if (FD_ISSET(fd->FD, &readFDs))
	    (*module->IOHandler)(fd->FD, IOReadReady);
	  fd = savePtr;
	}
	for (fd = module->WriteFDs; fd != NULL;) {
	  savePtr = fd->Next;
	  if (FD_ISSET(fd->FD, &writeFDs))
	    (*module->IOHandler)(fd->FD, IOWriteReady);
	  fd = savePtr;
	}
      }

      for (executable = ServerExecutables; executable != NULL; executable = executable->Next) {
	if (FD_ISSET(executable->Executable->StdOut, &readFDs))
	  while (read(executable->Executable->StdOut, DiscardBuffer, 128) > 0);
	if (FD_ISSET(executable->Executable->StdIn, &writeFDs)) {
	  FD_CLR(executable->Executable->StdIn, &writeFDs);
	  if (executable->Executable->StdIn > FDMax)
	    SetFDMax();
	}
      }

      for (client = ServerClients; client != NULL;) {
	savePtr = client->Next;
	CheckReceive(client, &readFDs);
	client = savePtr;
      }

      for (client = ServerClients; client != NULL;) {
	savePtr = client->Next;
	CheckSend(client);
	client = savePtr;
      }
    }
  }
  
  for (serverSocket = ServerSockets; serverSocket != NULL;) {
    savePtr = serverSocket->Next;
    Cclose(serverSocket->Socket);
    free(serverSocket);
    serverSocket = savePtr;
  }
  free(DiscardBuffer);
  return 0;
}

void AddExecutableStdIn(LcseExecutable executable) {
  assert(executable != NULL);

  FD_SET(executable->StdIn, &MasterWrite);
  if (executable->StdIn > FDMax)
    FDMax = executable->StdIn;
}

void RemoveExecutableStdIn(LcseExecutable executable) {
  assert(executable != NULL);

  FD_CLR(executable->StdIn, &MasterWrite);
  if (executable->StdIn == FDMax)
    SetFDMax();
}

int AddSelectIO(void* handle, int fd, LcseSelectType type) {
  assert(handle != NULL);
  assert(fd > -1);

  ServerModule module = NULL;
  ServerFD newFD = malloc(sizeof(struct ServerFDRecord));
  if (newFD == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on newFD.\n");
    return -1;
  }
  newFD->FD = fd;

  if (type == SelectRead) {
    if (LogIsDebug(Logger)) {
      struct sockaddr_in addr;
      socklen_t len = sizeof(addr);
      if (getpeername(fd, (struct sockaddr *) &addr, &len) != -1)
	LogMessage(Logger, LogDebug, "Adding socket(%s:%u) to MasterRead.",
		   inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
      else
	LogMessage(Logger, LogDebug, "Adding fd to MasterRead.");
    }
    FD_SET(fd, &MasterRead);
  }
  else {
    if (LogIsDebug(Logger)) {
      struct sockaddr_in addr;
      socklen_t len = sizeof(addr);
      if (getpeername(fd, (struct sockaddr *) &addr, &len) != -1)
	LogMessage(Logger, LogDebug, "Adding socket(%s:%u) to MasterWrite.",
		   inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
      else
	LogMessage(Logger, LogDebug, "Adding fd to MasterWrite.");
    }
    FD_SET(fd, &MasterWrite);
  }
  if (fd > FDMax) {
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "FDMax = %d, fd = %d, updating.", FDMax, fd);
    FDMax = fd;
  }

  for (module = ServerModules; module != NULL && module->Handle != handle; module = module->Next);
  if (type == SelectRead) {
    newFD->Next = module->ReadFDs;
    module->ReadFDs = newFD;
  }
  else {
    newFD->Next = module->WriteFDs;
    module->WriteFDs = newFD;
  }

  return 0;
}

int IsSelectIOSet(int fd, LcseSelectType type) {
  assert(fd > -1);

  if (type == SelectRead)
    return FD_ISSET(fd, &MasterRead);
  else
    return FD_ISSET(fd, &MasterWrite);
}

void RemoveSelectIO(void* handle, int fd, LcseSelectType type) {
  assert(handle != NULL);
  assert(fd > -1);
  
  ServerModule module = NULL;
  ServerFD current = NULL, previous = NULL;

  if (type == SelectRead) {
    if (LogIsDebug(Logger)) {
      struct sockaddr_in addr;
      socklen_t len = sizeof(addr);
      if (getpeername(fd, (struct sockaddr *) &addr, &len) != -1)
	LogMessage(Logger, LogDebug, "Removing socket(%s:%u) from MasterRead.",
		   inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
      else
	LogMessage(Logger, LogDebug, "Removing fd from MasterRead.");
    }
    FD_CLR(fd, &MasterRead);
  }
  else if (type == SelectWrite) {
    if (LogIsDebug(Logger)) {
      struct sockaddr_in addr;
      socklen_t len = sizeof(addr);
      if (getpeername(fd, (struct sockaddr *) &addr, &len) != -1)
	LogMessage(Logger, LogDebug, "Removing socket(%s:%u) from MasterWrite.",
		   inet_ntoa(addr.sin_addr), (int)ntohs(addr.sin_port));
      else
	LogMessage(Logger, LogDebug, "Removing fd from MasterWrite.");
    }
    FD_CLR(fd, &MasterWrite);
  }
 
  for (module = ServerModules; module != NULL && module->Handle != handle; module = module->Next);
  for (previous = NULL, current = (type == SelectRead) ? module->ReadFDs : module->WriteFDs; 
       current != NULL;) {
    if (current->FD == fd) {
      if (previous == NULL) {
	if (type == SelectRead) 
	  module->ReadFDs = current->Next;
	else
	  module->WriteFDs = current->Next;
      }
      else
	previous->Next = current->Next;
      free(current);
      current = NULL;
    }
    else {
      previous = current;
      current = current->Next;
    }
  }

  if (fd == FDMax) {
    int fdMax = 0;
    for (current = (type == SelectRead) ? module->WriteFDs : module->ReadFDs; current != NULL; 
	 current = current->Next) {
      if (current->FD > fdMax)
	fdMax = current->FD;
    }
    if (LogIsDebug(Logger))
      LogMessage(Logger, LogDebug, "fdMax = %d, fd = %d, FDMax = %d", fdMax, fd, FDMax);
    if (fdMax != fd)
      SetFDMax();
  }
}
 
int SetNonBlockingIO(int fd) {
  assert(fd > -1);

  int opts = 0;
  while ((opts = fcntl(fd, F_GETFL)) < 0) {
    int err = errno;
    if (err != EINTR) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError,
		   "fcntl failed when trying to read option from fd.  %s.", strerror(err));
      return opts;
    }
  }

  if ((opts & O_NONBLOCK) != O_NONBLOCK) {
    int retVal = 0;
    opts = (opts | O_NONBLOCK);    
    while ((retVal = fcntl(fd, F_SETFL, opts)) < 0) {
      int err = errno;
      if (err != EINTR) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError,
		     "fcntl failed when trying to set non-blocking option on fd.  %s.", strerror(err));
	return retVal;
      }
    }
  }

  return opts;
}

int SetCloseOnExecIO(int fd) {
  assert(fd > -1);

  int flags = 0;
  while ((flags = fcntl(fd, F_GETFD)) < 0) {
    int err = errno;
    if (err != EINTR) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError,
		   "fcntl failed when trying to read flags from fd.  %s.", strerror(err));
      return flags;
    }
  }

  if ((flags & FD_CLOEXEC) != FD_CLOEXEC) {
    int retVal = 0;
    flags = (flags | FD_CLOEXEC);
    while ((retVal = fcntl(fd, F_SETFD, flags)) < 0) {
      int err = errno;
      if (err != EINTR) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError,
		     "fcntl failed when trying to set close-on-exec flag on fd.  %s.", strerror(err));
	return retVal;
      }
    }
  }
  
  return flags;
}


void Cclose(int fd) {
  assert(fd > -1);

  while(close(fd) < 0) {
    int err = errno;
    if (err != EINTR) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "close failed on file descriptor.  %s.", strerror(err));
      return;
    }
  }
}

void CloseClient(LcseClient client) {
  assert(client != NULL);
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Closing client from handler.");
  
  ServerClient current = NULL;
  
  for (current = ServerClients; current != NULL && current->Client != client; current = current->Next);
  
  ShutdownClient(current);
}

pid_t SpawnExecutable(void* handle, char** argv) {
  assert(handle != NULL);
  if (LogIsDebug(Logger))
    LogMessage(Logger, LogDebug, "Spawning new executable.");
  
  pid_t processID = -1;
  int fds[4], k = 0, err = 0;
  ServerExecutable executable = NULL;
  ServerModule module = NULL;
  
  executable = malloc(sizeof(struct ServerExecutableRecord));
  if (executable == NULL) {
    if (LogIsError(Logger)) 
      LogMessage(Logger, LogError, "malloc failed on executable.");
    return -2;
  }
  executable->Executable = malloc(sizeof(struct LcseExecutableRecord));
  if (executable->Executable == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on executable->Executble.");
    free(executable);
    return -2;
  }
  for (k = 0; argv[k] != NULL; k++);
  executable->Executable->Argv = malloc(sizeof(char*) * (k + 1));
  if (executable->Executable->Argv == NULL) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "malloc failed on executable->Executable->Argv.");
    free(executable->Executable);
    free(executable);
    return -2;
  }
  for (k = 0; argv[k] != NULL; k++) {
    executable->Executable->Argv[k] = strdup(argv[k]);
    if (executable->Executable->Argv[k] == NULL) {
      if (LogIsError(Logger))
	LogMessage(Logger, LogError, "malloc failed on executable->Executable->Argv[%d].", k);
      for (err = 0; err < k; err++)
	free(executable->Executable->Argv[err]);
      free(executable->Executable->Argv);
      free(executable->Executable);
      free(executable);
      return -2;
    }
  }
  executable->Executable->Argv[k] = NULL;
  
  if (pipe(fds) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "pipe failed on fds[0].");
    for (k = 0; executable->Executable->Argv[k] != NULL; k++)
      free(executable->Executable->Argv[k]);
    free(executable->Executable->Argv);
    free(executable->Executable);
    free(executable);
    return -3;
  }
  if (pipe(&fds[2]) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "pipe failed on fds[2].");
    Cclose(fds[0]);
    Cclose(fds[1]);
    for (k = 0; executable->Executable->Argv[k] != NULL; k++)
      free(executable->Executable->Argv[k]);
    free(executable->Executable->Argv);
    free(executable->Executable);
    free(executable);
    return -3;
  }

  executable->Executable->StdOut = fds[0];
  executable->Executable->StdIn = fds[3];
  FD_SET(executable->Executable->StdOut, &MasterRead);
  if (executable->Executable->StdOut > FDMax)
    FDMax = executable->Executable->StdOut;

  if (SetNonBlockingIO(executable->Executable->StdOut) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "SetNonBlockingIO failed on StdOut.");
    DisposeExecutable(executable);
    Cclose(fds[1]);
    Cclose(fds[2]);
    return -4;
  }
  if (SetCloseOnExecIO(executable->Executable->StdOut) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "SetCloseOnExecIO failed on StdOut.");
    DisposeExecutable(executable);
    Cclose(fds[1]);
    Cclose(fds[2]);
    return -4;
  }

  if (SetNonBlockingIO(executable->Executable->StdIn) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "SetNonBlockingIO failed on StdIn.");
    DisposeExecutable(executable);
    Cclose(fds[1]);
    Cclose(fds[2]);
    return -4;
  }
  if (SetCloseOnExecIO(executable->Executable->StdIn) < 0) {
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "SetCloseOnExecIO failed on StdIn.");
    DisposeExecutable(executable);
    Cclose(fds[1]);
    Cclose(fds[2]);
    return -4;
  }

  executable->Executable->ProcessID = executable->Executable->Status = 0;
  for (module = ServerModules; module != NULL && module->Handle != handle; module = module->Next);
  executable->Next = module->Executables;
  module->Executables = executable;
  
  processID = fork();

  if (processID < 0) {
    err = errno;
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "fork failed.  %s.", strerror(err));
    module->Executables = executable->Next;
    DisposeExecutable(executable);
    Cclose(fds[1]);
    Cclose(fds[2]);
  }
  else if (processID == 0) {
    char byte = '\0';

    while (dup2(fds[1], STDOUT_FILENO) < 0) {
      err = errno;
      if (err != EINTR) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "dup2 failed on STDOUT_FILENO.  %s.", strerror(err));
	exit(1);
      }
    }
    Cclose(fds[1]);

    while (dup2(fds[2], STDIN_FILENO) < 0) {
      err = errno;
      if (err != EINTR) {
	if (LogIsError(Logger))
	  LogMessage(Logger, LogError, "dup2 failed on STDIN_FILENO.  %s.", strerror(err));
	exit(1);
      }
    }
    Cclose(fds[2]);
    
    write(STDOUT_FILENO, &byte, 1);
    execvp(argv[0], argv);
    
    err = errno;
    if (LogIsError(Logger))
      LogMessage(Logger, LogError, "execvp failed on %s.  %s.", argv[0], strerror(err));
    exit(127);
  }
  else {
    executable->Executable->ProcessID = processID;
    Cclose(fds[1]);
    Cclose(fds[2]);
  }
  return processID;
}
